/****************************************************************************
 *									    *
 *			  COPYRIGHT (c) 1990 - 2010			    *
 *			   This Software Provided			    *
 *				     By					    *
 *			  Robin's Nest Software Inc.			    *
 *									    *
 * Permission to use, copy, modify, distribute and sell this software and   *
 * its documentation for any purpose and without fee is hereby granted,	    *
 * provided that the above copyright notice appear in all copies and that   *
 * both that copyright notice and this permission notice appear in the	    *
 * supporting documentation, and that the name of the author not be used    *
 * in advertising or publicity pertaining to distribution of the software   *
 * without specific, written prior permission.				    *
 *									    *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, 	    *
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN	    *
 * NO EVENT SHALL HE BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL   *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR    *
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  *
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF   *
 * THIS SOFTWARE.							    *
 *									    *
 ****************************************************************************/
/*
 * Module:	dt.c
 * Author:	Robin T. Miller
 *
 * Description:
 *	Main line code for generic data test program 'dt'.
 */
#include "dt.h"
#include <ctype.h>
#include <fcntl.h>
#include <math.h>
#include <signal.h>
#if !defined(WIN32)
#  if !defined(_QNX_SOURCE) 
#    if !defined(sun)
#      include <sys/ioctl.h>
#    endif /* !defined(sun) */
#    include <sys/file.h>
#    include <sys/param.h>
#    if defined(sun) || defined(_OSF_SOURCE)
#      include <sys/mman.h>
#    endif /* defined(sun) || defined(_OSF_SOURCE) */
#  endif /* !defined(_QNX_SOURCE) */
#include <sys/wait.h>
#endif /* !defined(WIN32) */
#if defined(DEC)
#  include <sys/utsname.h>
#endif /* defined(DEC) */

#if defined(HP_UX)
#  include <sys/scsi.h>
#  if !defined(SCSI_MAX_Q_DEPTH)
#    define SCSI_MAX_Q_DEPTH 255
#  endif
#endif /* defined(HP_UX) */

/*
 * Modification History:
 *
 * June 2oth, 2012 by Robin T. Miller
 * 	Per request of SAN Automation, when receiving and unexpected exit status
 * from ExecuteTrigger(), call terminate(). This usually means the script path
 * cannot be found, but our usual exit status tells dt to terminate, so we will!
 * 
 * June 6th, 2012 by Robin T. Miller
 * 	Add support for multiple devices (you'll get one process per device).
 * 	Fix bug with enable=deleteperpass, correctly handle WARNING (dir exists)
 * from setup_directory_info(). Was breaking out of main loop in warning! :-(
 * 
 * January 5th, 2012 by Robin T. Miller
 *	Added prealloc flag for use with Windows when sparse files are disabled.
 * Please see the modification history in dtgen.c for a full description of why.
 *
 * December 19th, 2011 by Robin T. Miller
 * 	Moved init_vars() here from dtinfo.c and started added more per device
 * variables to the device information data structure (dinfo_t), in preparation
 * to (eventually) make dt multi-threaded. Alot of work, given it's original
 * single threaded design (too many globals, etc), but we expect to get there!
 * 
 * December 16th, 2011 by Robin T. Miller
 * 	Add iotune=file to allow IO delays to be overriden from file while running.
 * 	Allow keyword 'random' with *delay values to support random IO delays.
 *
 * November 11th, 2011 by Robin T. Miller
 * 	Remove setting the thread stack size, causing issues with being too small,
 * and we don't need this until dt is reworked for doing threaded I/O anyways.
 * 
 * October 13th, 2011 by Robin T. Miller
 * 	Added enable=reread option for use with read-after-write (raw) option.
 * This allows a full read pass to help determine if data was incorrectly changed
 * after the initial write/read/verify pass. Adds an extra debug data point. :-)
 * 
 * October 6th, 2011 by Robin T. Miller
 * 	When "deleteperpass" option is enabled, do NOT delete the files when the
 * error limit is reached, otherwise the files cannot be reread for analysis.
 * The terminate function will honor the dispose={delete|keep|keeponerror} param.
 * 
 * August 18th, 2011 by Robin T. Miller
 * 	Update StartMonitorThread() to ensure the thread stack isn't too large!
 * 
 * August 16th, 2011 by Robin T. Miller
 * 	Added stop=filename option, to watch for file existance and stop after
 * a pass (therefore, a complete pass is considered a write/read/verify).
 * 
 * August 14th, 2011 by Robin T. Miller
 * 	Enhanced the no-progress message to include the lba and file offset to
 * help with debugging slow/hung I/O (easier to track in traces).
 * 
 * August 5th, 2011 by Robin T. Miller
 * 	Added enable=fsincr option to support file size incrementing. This only
 * has meaning with multiple files, and basically sets the record count to the
 * file number. So, bs=512 files=3 enable=fsincr writes: 512, 1024, 1536 bytes.
 *
 * July 28th, 2011 by Robin T Miller
 * 	When the parent dt receives a signal and starts terminating, added a kill
 * delay to give children a chance to terminate from same process group signal.
 * If parent sends another signal immediately, child may terminate before cleanup
 * and reporting its' statistics. For file systems, files were not getting removed.
 * Note: The default kill delay is 1 second, settable via: kill_delay=value option.
 * 
 * July 27th, 2011 by Robin T. Miller
 * 	Update thread cancel logic when terminating, so set asynchronous aborts,
 * to avoid a race with I/O thread and cancelation points (it may continue to run).
 * Async cancel should terminate the thread immediately, avoiding false errors.
 * Also added a "cancel_delay=value" option to permit an option delay after cancel.
 * 
 * May 6th, 2011 by Robin T. Miller
 * 	Fix bug when step option and slices are speicifed, set the end position.
 * When this is *not* set, then we set a premature EOF/EOM with both options.
 * 
 * March 24th, 2011 by Robin T. Miller
 * 	Adding bufmodes={buffered,unbuffered,cachereads,cachewrites} option.
 * This allows testing different buffering modes on each pass. One or all
 * buffering modes can be specified, and we'll round-robin on each pass.
 * 
 * March 3rd, 2011 by Robin T. Miller
 * 	Allocate the global message buffer much earlier, since report_error()
 * uses this to log messages to syslog(). (gulp, we were seg faulting! :-().
 * 
 * February 16th, 2011 by Robin T. Miller
 * 	In report_error(), ensure we have a valid dip (early errors do not!).
 * Otherwise we seg fault dereferencing a NULL pointer (grrr).
 * 
 * January 15th, 2011 by Robin T. Miller
 *	Added separate options for controlling read/write caching. Well,
 * what this really provides is control of direct I/O for write pass and
 * read pass, to bypass the buffer cache (actually separate DIO control).
 *
 * November 16th, 2010 by Robin T. Miller
 *	Added enable=deleteperpass option to delete files between passes.
 *	This only applies to regular files (of course) on file systems.
 *
 * January 23rd, 2010 by Robin T. Miller
 *	Add support for maxdata= and maxfiles= options to limit the amount of
 * data and files when creating multiple directories and files.
 *
 * December 4th, 2009 by Robin T. Miller
 *	When exiting due to a signal, do NOT exit with the signal number, but
 * detect and map the signal to FAILURE status, easier for scripts to handle.
 * Note: Expected signals, such as SIGALRM, still get handled properly.
 *
 * December 2nd, 2009 by Robin T. Miller
 *	When executing multiple processes, when the parent receives a signal,
 * don't send a signal to child processes, since they automatically receive the
 * signal since they are part of the process group. If the parent sends another
 * signal, the child process will terminate immediately and won't get a chance
 * to perform its' cleanup (reporting stats and deleting files).
 *
 * December 1st, 2009 by Robin T. Miller
 *	Added support for file systems: dirs=, sdirs=, depth=, and dirprefix=.
 *
 * November 25th, 2009 by Robin T. Miller
 *	Added support for a directory path option (dir=dirpath). The dirpath
 * is made unique for multiple processes, so each proc uses its' own directory.
 * The directory is created is it doesn't exist, and automatically removed,
 * assuming all files are deleted when exiting.
 *
 * November 21st, 2009 by Robin T. Miller
 *	Added support for multiple (regular) files via files= option. This
 * option was previous only implemented for multiple tape files, now it is
 * honored for multiple disk files. This is the first step towards improved
 * file system testing (with multiple procs, lots of metedata activity).
 *
 * October 29th, 2009 by Robin T. Miller
 *	Honor the runtime, even when the runtime is greater than the alarm time.
 *	Add support for a keepalive time, so keepalive messages can be used in
 * conjunction with short alarm values doing I/O monitoring.
 *
 * September 2nd, 2009 by Robin T. Miller.
 *	Add nice=value option to allow dt's priority to be altered.
 *
 * February 4th, 2009 by Robin T. Miller.
 *	Change default for eof_status flag to FALSE to map EOF to Successful
 * exit status.  This can be reverted via "enable=eof", but most folks expect
 * Success (0) when tests complete, and checking for EOF status (254) causes
 * more problems than it's worth right now. Legacy scripts may want EOF status.
 *
 * November 14th, 2008 by Robin T. Miller.
 *	When using the step option in the forward direction to regular files,
 * added an end position to be used by the I/O function to stop properly.
 *
 * October 4th, 2008 by Robin T Miller.
 *	Two major changes for native Windows:
 *	- use beginthreadex() instead of CreateThread() for C RTL
 *	- do NOT use ThreadSuspect/ThreadResume, since it may deadlock
 *	  with the multi-threaded LIBCMT and MSVCRT C run-time libraries
 *
 * September 30th, 2008 by Robin T. Miller.
 *	When trigger script exits with TERMINATE or ABORT status code, then
 * ensure the exit status gets set to FAILURE.  This ensures failure is used
 * as our exit status, and with multiple procs, this failed status forced
 * other processes to be aborted whtn oncerr=abort is specified.
 *	In terminate() for Windows, do not continue waiting for processes
 * after calling abort_procs(), since we do not need to reap their status.
 * This can cause the parent dt to hang after being aborted via Ctrl/C.
 *
 * September 9th, 2008 by Robin T. Miller.
 *	Read ThreadSuspend/ThreadResume around calls to keepalive_alarm(),
 * to avoid race conditions (and avoid adding locks).  This function handles
 * no-progress, keepalive messages, and runtime (overloaded).
 *
 * September 8th, 2008 by Robin T. Miller.
 *	Move signal arming earlier, so it's in affect for multiple processes.
 *
 * July 23rd, 2008 by Robin T. Miller.
 *	Add support for a history mechanism.  This saves the specified number
 * of I/O request in a circular history buffer, which is dumped when there's
 * a data corruptiom or an I/O error. New options are history= & hdsize=value.
 * Added dispose=keeponerror option, so file is kept on errors, else deleted.
 *
 * July 15th, 2008 by Robin T. Miller.
 *	Map O_APPEND to FILE_APPEND_DATA for native Windows.
 *
 * June 24th, 2008 by Robin T. Miller.
 *	Add support for new notime=optype to control no-progress operation types.
 *	For parsing errors, log failures as logLevelError instead of logLevelCrit.
 *
 * June 7th, 2008 by Robin T. Miller.
 *	Implement no-progress/keepalive timers for Windows.
 *	Removed PRT_NOIDENT when logging keepalive stats so the process ID is
 * printed with multiple processes. Needed to identify proc!
 *
 * March 11th, 2008 by Robin T. Miller.
 *	Add enable/disable=prefill option to control the automatic pre-filling
 * of the read buffers, which add to the compute (cpu) intensity (step 1).
 * Note: This prefill was needed for simple 4-byte internal data patterns,
 * but is *not* required for IOT and most pattern files since they're unique.
 *
 * December 18th, 2007 by Robin T. Miller.
 *	Add trigargs flag to disable trigger arguments appended to trigger
 * command/script.  This allows a simple trigger command, without the need
 * to generate a script to invoke.
 *
 * December 13th, 2007 by Robin T. Miller.
 *	Add support for logging errors to syslog on Unix based systems.
 *
 * October 24th, 2007 by Robin T. Miller.
 *	When a user pattern is specified, disable the unqiue pattern flag so
 * multiple slices use the users' pattern rather than a unique pattern per proc.
 *
 * October 23rd, 2007 by Robin T. Miller.
 *	When exiting with no-progress trigger script, ensure the exit status
 * is set to the SIGALRM, to ensure a non-zero exit status. Otherwise, SIGALRM
 * is mapped to SUCCESS, which is desired when a runtime=secs expires.
 *
 * September 26th, 2007 by Robin T. Miller.
 *	When IOT pattern is chosen, enable fsalign_flag so variable I/O sizes
 * are aligned to the device size instead of the pattern size. The latter gets
 * set to the max size, thus preventing variable IOT blocks with a regular file.
 * Note: This SNAFU affected variable I/O sizes and random I/O offsets w/IOT!
 *
 * June 22nd, 2007 by Robin T. Miller.
 *	Added "disable=sighup" option to control handling of SIGHUP. Ignoring
 * SIGHUP is desirable when backgrounding 'dt' from shell scripts, so it keeps
 * running when folks log off (most shells send a SIGHUP).  Yea, Unix 'nohup'
 * could be used. Also altered the no-progress message to include device name
 * and "mm:hh:ss" format when the number of seconds is over a minute.
 *
 * October 27th, 2006 by Robin T. Miller.
 *      Fixed logic error where we were reopen'ing the file in preparation
 * for the next write pass, when we've reached the error limit.  This was
 * a problem if oflags=trunc and dispose=keep were used since the resulting
 * output file would be zero length.
 *
 * January 6th, 2007 by Robin T. Miller.
 *	Updated report_record(), used by debug code, since LBA's are now
 * treated as 64-bit values to support capacities over 2TB!
 *
 * October 16th, 2006 by Robin T. Miller.
 *      Added support for "enable=timestamp" option to write a timestamp
 * at the beginning of each data block to help troubleshooting corruptions.
 *
 * October 4th, 2006 by Robin Miller.
 *      If regular file is not open when terminating, do not try to delete.
 *      Added check for non-modulo block size patterns, when random I/O is
 * enabled, to avoid false data corruption when blocks are overwritten.
 *
 * May 16th, 2006 by Robin Miller.
 *	Added options "noprog" and "noprogt" to be used in conjunction with
 * alarm option to detect no I/O progress being made.
 *
 * December 7th, 2005 by Robin Miller.
 *	Added __linux__ conditional for hz initialization, so compiles will
 * work on Red Hat RHEL4 (patch from Alan Brunelle, thanks!).
 *
 * November 22nd, 2005 by Robin Miller.
 *      For Hazard, allow empty prefix string so users can override default.
 *
 * January 4th, 2005 by Robin Miller.
 *	Fix bug in terminate() where signal code was getting overwritten when
 * a timer was active.  Now, only set code with exit_status on SIGALRM.
 *
 * December 23rd, 2004 by Robin Miller.
 *      Added "stats={brief,full,none}" option to provide finer control
 * over pass and total statistics.  This finer control was previously
 * provided by the "verbose" flag, but that was inadequate.
 *
 * December 21st, 2004 by Robin Miller.
 *      Added per pass and total keepalive format control messages.
 *
 * December 20th, 2004 by Robin Miller.
 *      Added support for alarm/keepalive options. The alarm=time specifies
 * the alarm time (same format as runtime), and keepalive=string specifies
 * the alive message with special format control characters to display when
 * the alarm expires.
 *
 * November 22nd, 2004 by Robin Miller.
 *      Change the dump data limit from 64 byte to one block (512 bytes).
 * If read-after-write (raw) is enabled, don't enable FS align logic.
 *
 * November 11th, 2004 by Robin Miller.
 *      For Solaris, since O_DIRECT isn't supported, set dio_flag so
 * directio() API can be used after the file is open'ed.
 *
 * October 21st, 2004 by Robin Miller.
 *      Enable file system alignments by default for random I/O,
 * since variable sizes and offsets cause false data corruptions.
 * When logic is added to track previously written blocks, this
 * can be reverted.
 *
 * July 7th, 2004 by Robin Miller.
 *      For HP-UX, add option qdepth=N" to set the queue depth.
 *
 * June 24th, 2004 by Robin Miller.
 *      Handle errors from ctime_r() - two flavors are around!
 *      Added "enable/disable=fsalign" option to control behaviour
 * of aligning FS offsets and variable record sizes to device size.
 * This option is off by default so random non-device sized values
 * will now be generated (useful for file system testing).  This
 * option can be enabled to return to previous behaviour for FS's.
 *
 * June 22nd, 2004 by Robin Miller.
 *      Added support for triggers on corruption.
 *
 * February 27th, 2004 by Robin Miller.
 *       Switch ctime() to ctime_r() in preparation for threads.
 *
 * February 24th, 2004 by Robin Miller.
 *      Allow parsing of aios=0, to ease supporting of OS's which
 * don't support POSIX AIO.  Makes script writing a little simpler.
 *
 * December 6th, 2003 by Robin Miller.
 *      Conditionalize to exclude tty code.
 *
 * November 17th, 2003 by Robin Miller.
 *	Breakup output to stdout or stderr, rather than writing
 * all output to stderr.  If output file is stdout ('-') or a log
 * file is specified, then all output reverts to stderr.
 *	Parse AIO options, but then emit warning if not supported.
 * This allows portable scripts to be written without parameterizing
 * exlusion of AIO if a platform (OS) does not support it.
 *
 * October 31st, 2003 by Robin Miller.
 *	In terminate(), don't attempt flush if device or file is
 * not open when we receive the alarm signal.
 *
 * September 27th, 2003 by Robin Miller.
 *	Added support for AIX.
 *
 * June 9th, 2003 by Robin Miller.
 *	When terminating as result of alarm, flush output data if we
 * were writing a file.  Thanks to Hank Jakiela for reporting this.
 *
 * March 20th, 2003 by Robin Miller.
 *	Fix bug or'ing O_LARGEFILE into open flags (thanks Kris Corwin!).
 *
 * March 15th, 2003 by Robin Miller.
 *	Added "prefix=" option to support data pattern prefix string.
 *
 * March 14th, 2003 by Robin Miller.
 *	Added "slice=" option to support testing an individual slice.
 *
 * February 23rd, 2002 by Robin Miller.
 *      Make porting changes for HP-UX IA64.
 *
 * May 31st, 2001 by Robin Miller.
 *	Don't allow different data patterns with multiple processes,
 * unless writing to a regular file (each process has own filename).
 *
 * April 13th, 2001 by Robin Miller.
 *	Added "capacity=max" for Tru64 Unix, this will force using the
 * driver returned media capacity.  Prevents lseek/read algorithm when
 * using random I/O too!  Note:  Will inhibit disk driver EOM testing.
 *
 * February 24th, 2001 by Robin Miller.
 *	Add conditionalization for QNX RTP (Neutrino).
 *
 * February 21st, 2001 by Robin Miller.
 *	Disable EOF/EOM exit status by default.  This causes trouble for
 * normal testers.  Use "enable=eof" to reenable previous behaviour.
 *
 * February 6th, 2001 by Robin Miller.
 *	If input file and multiple slices, allow cycling through data
 * patterns, since the file is expected to have been written with unique
 * patterns in each slice region.  Minor update to deleting files too.
 *
 * February 3rd, 2001 by Robin Miller.
 *	Lift restriction of bs= or min= sizes being smaller than
 * sizeof(u_int32) for IOT pattern or lbdata option.  Although this
 * isn't *really* correct, enforcing this breaks numerious scripts.
 * Although smaller works, it's only by luck and the fact we have
 * pad bytes at EOB, which prvented data corruption in the past!
 *
 * January 28th, 2001 by Robin Miller.
 *	Added "slices=value" and "enable/disable=unique" options.
 * The slices option carves up a disk with each process exercising a
 * differnent range of blocks.  The unique pattern option sets up a
 * unique pattern for each process started for slices & regular files.
 *
 * January 26th, 2001 by Robin Miller.
 *	Added report_record() to the report record information.
 *	Added "iodir={forward,reverse}" option to all reverse I/O
 * to rotating media.  Add validation checks both before and after
 * a device is open (mostly random/reverse I/O checks).
 *
 * January 24th, 2001 by Robin Miller.
 *	Added "dsize=value" to set the device block size.
 *	Added "incr=var" option for variable request sizes.
 *	Removed some dead code (#if 0 stuff), to cleanup a bit.
 *	Updates to allow the IOT pattern to use non-modulo 512 byte
 * sizes.  The logic is: 1) user defined "lbs=value", 2) device block
 * size (Tru64 Unix), or 3) default to 512 byte block (original default).
 * This change was initiated by Windows/NT IOT disc using 2KB sector size.
 *
 * January 18th, 2001 by Robin Miller.
 *	When requesting multiple volumes, ensure the tape device gets
 * closed prior to retries, if the rewind operation fails.  Otherwise,
 * the next tape open fails with EBUSY (exclusive open device).
 *
 * January 14th, 2001 by Robin Miller.
 *	Added support for multiple volumes option.
 *
 * December 30th, 2000 by Robin Miller.
 *	Make changes to build using MKS/NuTCracker product.
 *
 * November 19th, 2000 by Robin Miller.
 *	Add missing enable/disable=fsync goto label statements.
 *
 * November 10th, 2000 by Robin Miller.
 *	Added sanity check to warn user about unpredictable results
 * when writing to a disk with multiple procs and multiple passes.
 *
 * November 8th, 2000 by Robin Miller.
 *	Added "disable=cerrors" to disable device close errors.
 * [ Note: This is really a workaround for the Linux tape driver. ]
 *	In HandleMultiVolume(), if closing the device fails, then
 * return that failure to abort the test.  With tapes, this means
 * flushing the buffered data or writing filemarks has failed.
 *
 * October 2nd, 2000 by Robin Miller.
 *	Enhanced report_error() to display errno value (same as Scu).
 *
 * August 22nd, 2000 by Robin Miller.
 *	Added boolean flags to track user set min, max, & incr values.
 *
 * July 14th, 2000 by Robin Miller.
 *	Added fsync_flag to control sync'ing data to disk files.
 *
 * May 8th, 2000 by Robin Miller.
 *	Adding parsing of "version" option, which is used to only
 * display the version string, dtversion() in file dtusage.c
 *
 * May 5th, 2000 by Robin Miller.
 *	Set proper exit code in terminate(), when exiting due to
 * an alarm, i.e., runtime= option.  Previously, the end of file
 * exit status was being ignored and we exited with failure status.
 *
 * April 25th, 2000 by Robin Miller.
 *	Fix problem of not breaking out of test loop if the error
 * limit has been reached on previous pass.
 *
 * March 28th, 2000 by Robin Miller.
 *	Modify "position=value" to accept a large numeric value,
 * otherwise we were limited to 32 bit value on non-64bit systems.
 * Do the same thing for "step=value" for 64-bit file offsets.
 * When using the "runtime=" option, when the alarm expires and
 * we call terminate() to exit, look at the exit_status variable
 * instead of error_count for errors, so we don't miss errors!
 *
 * March 27th, 2000 by Robin Miller.
 *   o	Fixed parsing of a couple flags, including "flags=direct".
 *   o	Added "capacity=value" option to set the drive capacity.
 *
 * March 20th, 2000 by Robin Miller.
 *	Don't allow a pass limit of zero.  This also avoids a
 * core dump when preparing statistics, since the active device
 * never got setup.
 *
 * March 2nd, 2000 by Robin Miller.
 *	In HandleMultiVolume(), reset the exit status to SUCCESS,
 * since this got set to END_OF_FILE earlier.  Failure to do this
 * means we exit with an END_OF_FILE (254) status (no good :-).
 *
 * February 17th, 2000 by Robin Miller.
 *	Adding better support for multi-volume tape testing.
 *
 * January 17th, 2000 by Robin Miller.
 *	In copy mode, is input file is stdin, don't do verify.
 *	Enable random I/O, for ralign, rlimit, or rseed options.
 *
 * January 14th, 2000 by Robin Miller.
 *	Don't delete the output file unless we're in test mode.
 *
 * January 6th, 2000 by Robin Miller.
 *	Added support for multi-volume media (lot's of changes).
 *
 * January 1st, 2000 by Robin Miller.
 *	Added read after write support, "enable=raw" option.
 *
 * December 31st, 1999 by Robin Miller.
 *	Added "rseed=" option, so user can specify the random seed.
 *	Modify reopen_file logic to use O_RDWR if skip_count exists.
 *
 * November 11th, 1999 by Robin Miller.
 *	Added logging of diagnostic information to event logger.
 *
 * November 10th. 1999 by Robin Miller.
 *	Fixed parsing of "dispose=delete".
 *
 * November 9th, 1999 by Robin Miller.
 *	Modify logic associated with making stderr stream buffered.
 *
 * August 26th, 1999 by Robin Miller.
 *	If during the write pass no data was transferred, don't do
 * the read pass since obviously this will result in a compare error.
 *
 * August 7th, 1999 by Robin Miller.
 *	Allow enabling open w/O_LARGEFILE via "flags=large", incase
 * _FILE_OFFSET_BITS=64 does _not_ enable this by default.
 *
 * July 22nd, 1999 by Robin Miller.
 *	Added support for IOT (DJ's) test pattern.
 * 
 * July 19th, 1999 by Robin Miller.
 *	Add flag to control log file header normally displayed.
 *
 * June 28, 1999 by Robin Miller.
 *	For 32-bit systems, change count and limit variables from
 * u_long to double, since u_long limits us to 4GB (too small).
 *
 * May 27, 1999 by Robin Miller.
 *	Merge in George Bittner's changes for micro-second delays.
 *	Conditionalized all common/output file open flags.
 *
 * April 8, 1999 by Robin Miller.
 *	Merge in Jeff Detjen's changes for table()/sysinfo timing.
 *
 * April 7, 1999 by Robin Miller.
 *	On DEC systems, obtain tick/second via sysconf(_SC_CLK_TCK).
 *
 * December 21, 1998 by Robin Miller.
 *	Add pasing of "enable=resets", which enables tape repositioning
 * logic when bus/device resets are detected by DUNIX EEI status codes.
 *
 * December 19, 1998 by Robin Miller.
 *	- Disable setbuf() for stderr... screws up log files.
 *	- Write cmd line and version to log file for reference.
 *
 * December 16, 1998 by Robin Miller.
 *	Merge in changes made by George Bittner:
 *	- set random_seed using times() instead of time(0).
 *	- remove DLM_VALB, per Dan Christians' recommendation.
 *
 * October 29, 1998 by Robin Miller.
 *	Implement a random I/O data limit, instead of using the normal
 * data limit variable (not good to dual purpose this value).
 *
 * October 22, 1998 by Robin Miller. (Happy Birthday Mom!)
 *	When specifying a runtime, ensure we break out of the
 * passes loop if we exceed the error limit.  Previously, we'd
 * loop (possibly with error) for the duration of runtime :-)
 *
 * June 16, 1998 by Robin Miller.
 *	Add common open "flags={excl,ndelay,nonblock,rsync}".
 *	Add output (write) open "oflags={defer,dsync,trunc}".
 *
 * April 29, 1998 by Robin Miller.
 *	Add support for an alternate device directory.
 *
 * April 7, 1998 by Robin Miller.
 *	Setup the pattern as a pattern string, so non-modulo
 *	sizeof(pattern) read counts will data compare properly.
 *
 * March 20, 1998 by Robin Miller.
 *	When terminating, call close function before reporting stats,
 *	since in the case of AIO the I/O rundown updates statistics
 *	with outstanding requests, when we're aborted by a signal.
 *
 * May 14, 1997 by Robin Miller.
 *	In report_error(), set exit value to FAILURE instead of errno
 *	since this makes detecting errors easier in scripts.
 *
 * May 13, 1997 by Robin Miller.
 *	If a "skip=" count is specified with an output file "of=",
 *	open the file for R/W access since skips are accompished via
 *	read()'s of record size.
 *
 * March 27, 1997 by Ali Eghlima.
 *      Added cluster support, so more than one process or one system
 *      can access a resource. dlm are being used to synchronize all
 *      access.
 * 
 * March 7, 1997 by Robin Miller.
 *	If a copy or verify operation is selected, and a data limit
 *	is specified, then double the data limit to account for double
 *	the I/O's.  Previously, only half of the desired data limit was
 *	copied and/or verified.
 *
 * February 3, 1997 by Robin Miller.
 *	Check status from closing file descriptor, and if it's not
 *	SUCCESS, use this for the exit status.  For tapes, deferred
 *	writes and failure to write file marks occur at close time.
 *
 * March 30, 1996 by Robin Miller.
 *	When writing, now default to an infinite record limit, so
 *	writes stop when EOF is reached (disks and tapes only).
 *
 * February 29, 1996 by Robin Miller.
 *	Added callouts to setup data limit for disk random I/O.
 *
 * February 28, 1996 by Robin Miller.
 *	Added support for copying and verifying device/files.
 *
 * February 17, 1996 by Robin Miller.
 *	Add "enable=spad" option to control checking of pad bytes
 *	when long reads of short records is performed, since some
 *	controllers corrupt data bytes following the short record.
 *	[ NOTE:  At DEC, this action was deemed acceptable, since
 *	  the requester stated the driver could trash a larger buf. ]
 *
 * December 6, 1995 by Robin Miller.
 *	Add "ttymin=value" option to give user control over tty VMIN.
 *
 * November 19, 1995 by Robin Miller.
 *	When the device type is a disk, default min_size & incr_count to
 *	BLOCK_SIZE (512) instead of 1 (which is non-modulo for disks).
 *
 * November 18, 1995 by Robin Miller.
 *	Removing unaligned data access (ade) test code (code cleanup).
 *
 * July 24, 1995 by Robin Miller.
 *	Changed O_NDELAY to O_NONBLOCK which is defined by POSIX.  They are
 *	supposed to do the same thing, but they _do_ have different values.
 *	[ NOTE: This is only used for serial line modem testing. ]
 *
 * July 22, 1995 by Robin Miller.
 *	Add additional checking on "pattern=string" to avoid ambiguity
 *	with 4-byte hex data pattern strings... which is misleading).
 *
 * July 15, 1995 by Robin Miller.
 *	Add "oncerr=action" option to control child process
 *	error processing.
 *
 * July 7, 1995 by Robin Miller.
 *	Add "dlimit=value" option to override default dump limit.
 *
 * July 6, 1995 by Robin Miller.
 *	Enable dumping of data bytes on compare errors by default,
 *	since this program is used mainly as a diagnostic tool, and
 *	we now limit the number of bytes dumped.
 *
 * November 5, 1994 by Robin Miller.
 *	Don't set SIGCHLD signal to SIG_IGN (set to SIG_DFL) or else
 *	waitpid() won't detect any child processes (OSF R1.3 and QNX).
 *
 * September 23, 1994 by Robin Miller.
 *      Make changes necessary to build on QNX 4.21 release.  This
 *      required changing O_DSYNCH to O_DSYNC, and O_FSYNCH to O_SYNC.
 *
 * November 11, 1993 by Robin Miller.
 *	Removed code which was inadvertantly disabling data compares
 *	when there was no output file and verify was disabled.  This
 *	caused data comparisions to be disabled for terminal devices,
 *	now that the code has been rewritten/restructured (shit!!!).
 *
 * October 31, 1993 by Robin Miller.
 *	Enhance device type setup and honor user specified device type.
 *
 * October 28, 1993 by Robin Miller.
 *	Correct problem during loopback testing, where the device info
 *	structure fd's were not marked closed due to two places where
 *	close() was called directly (oops, missed a couple changes).
 *	After forking, display which process is the reader and writer.
 *
 * October 11, 1993 by Robin Miller.
 *	Conditionalize SunOS code (4.1.2 used) to avoid using strerror()
 *	function for obtaining error messages (return to perror() method).
 *
 * October 7, 1993 by Robin Miller.
 *	Add appropriate casting to block size checking, use ssize_t to
 *	prevent negative block sizes which don't work with read & write
 *	system calls (even though the count arg is declared as size_t).
 *
 * September 16, 1993 by Robin Miller.
 *	Properly report write pass count when read verify is disabled.
 *	Only update data pattern when writing files or loopback enabled
 *	when multiple passes specified (pattern cannot change for reads).
 *	Support "enable=loop" to enable loopback operation, rather than
 *	requiring both the same input & output devices (need only one).
 *	Added "oflags={append,sync}" for DEC OSF/1 systems (write opts).
 *
 * September 15, 1993 by Robin Miller.
 *	Added additional check for FIFO's prior to opening the device
 *	since O_NONBLOCK flag must be set to keep open() from blocking.
 *
 * September 8, 1993 by Robin Miller.
 *	Moved generic test functions into file dtgen.c.
 *
 * September 7, 1993 by Robin Miller.
 *	Moved tty specific code into dttty.c.  This code is dispatched
 *	to automatically after being setup in the device info structure.
 *
 * September 4, 1993 by Robin Miller.
 *	Lots of restructuring & code cleanup.  Nearly all testing is
 *	now appropriately parameterized (helps speed & new tests).
 *
 * September 3, 1993 by Robin Miller.
 *	Dispatch to test functions via function lookup table in
 *	preparation for device specific tests.
 *
 * September 1, 1993 by Robin Miller.
 *	Added "min=value", "max=value", and "incr=value" options for
 *	testing variable length records (mainly for tape devices).
 *
 * August 31, 1993 by Robin Miller.
 *	Added "align=rotate" option to rotate the starting data buffer
 *	address through sizeof(ptr) to force unaligned buffer accesses.
 *	This feature tests special handling of unaligned DMA addresses.
 *	Added "pattern=incr" option to use incrementing data pattern.
 *
 * August 27, 1993 by Robin MIller.
 *	Added support for DEC OSF/1 POSIX Asynchronous I/O (AIO).
 *
 * August 20, 1993 by Robin Miller.
 *	Added handling of reading & writing multiple tape files.
 *
 * August 18, 1993 by Robin Miller.
 *	Added "step=" option to specify a step offset for seeks.
 *	This is the number of bytes stepped after an I/O request.
 *
 * August 17, 1993 by Robin Miller.
 *	Added reporting of start & end times, time of errors, and
 *	added "runtime=" option to specify how long to run.
 *
 * August 13, 1993 by Robin Miller.
 *	Added alternate data patterns to use with multiple passes.
 *
 * August 5, 1993 by Robin Miller.
 *	Added "files=value" option for processing tape files.
 *
 * August 3, 1993 by Robin Miller.
 *	Added "procs=value" option to create multiple processes.
 *
 * September 19, 1992 by Robin Miller.
 *	Initialize flush_flag to TRUE so flushing of tty I/O queues
 *	gets done by default to discard any left over junk.
 * 
 * September 18, 1992 by Robin Miller.
 *	Added calls to save & restore terminal characteristics.
 *
 * September 17, 1992 by Robin Miller.
 *	Added parsing for "pattern=string" to allow ASCII string.
 *
 * September 12, 1992 by Robin Miller.
 *	Ensure the stderr stream is fully buffered, so the total time
 *	is not affected by flushing previous statistics.
 *
 * September 11, 1992 by Robin Miller.
 *	Added "oflags={cache,dsync,fsync,temp}" options to allow testing
 *	the affect of these open flags on QNX systems.
 *
 * September 10, 1992 by Robin Miller.
 *	Added parsing of MARK & SPACE parity for QNX Operating System.
 *
 * September 9, 1992 by Robin Miller.
 *	Allow loopback to same tty port (avoid exclusive open).
 *
 * September 5, 1992 by Robin Miller.
 *	Initial port to QNX 4.1 Operating System.
 *
 * September 3, 1992 by Robin Miller.
 *	Make changes necessary to remove BSD dependencies.
 *
 * August 19, 1992 by Robin Miller.
 *	Added support for testing tty modem control.
 *
 * May 25, 1992 by Robin Miller.
 *	Don't set exclusive open mode for terminal device if debug is
 *	enabled, so "stty -a" can be done on terminal under test.
 *
 * May 22, 1992 by Robin Miller.
 *	Control / force kernel address data exception via flag.
 *	Added option "enable=ade" (address data exception).
 *
 * April 24, 1992 by Robin Miller.
 *	Inform user that either a record count or a data limit must be
 *	specified.  Added flags to determine if we are reading stdin or
 *	writing to stdout streams.
 *
 * March 11, 1992 by Robin Miller.
 *	Changes necessary for port to 64-bit Alpha architecture.
 *
 * October 16, 1990 by Robin Miller.
 *	Added "align=offset" option to align memory buffers at a specific
 *	offset.  The default is page aligned.
 *
 * October 9, 1990 by Robin Miller.
 *	Added "hz=ticks" option to allow the ticks per second value to
 *	be specified.  On VAX systems, this is consistent (100), but on
 *	MIPS systems it various and is setup at boot time in the machine
 *	dependent code.
 *
 * August 21, 1990 by Robin Miller.
 *	Changed exit status so scripts can detect and handle errors
 *	based on the exit code.  If not success, fatal error, or end
 *	of file/tape, the exit code is the error number (errno).
 *
 * August 21, 1990 by Robin Miller.
 *	Added "position=offset" option to position to specified offset
 *	before starting test.
 *
 * August 16, 1990 by Robin Miller.
 *	Added "enable=coredump" option so if exiting with error status,
 *	we'll generate a core dump for analysis purposes.
 *
 * August 8, 1990 by Robin Miller.
 *	Changed malloc() to valloc() to align buffer on page boundry.
 *	On some archetectures, this results on better performance to
 *	raw devices since the DMA is done directly to the users' buffer.
 *
 * August 2, 1990 by Robin Miller.
 *	Added "pf=file" to specify a pattern file to use.  In this
 *	case, the contents of the specified file are used for the
 *	data pattern.
 *
 * April 11, 1990 by Robin Miller.
 *	Added "enable=mmap" option to enable memory mapped I/O.
 *
 * March 22, 1990 by Robin Miller.
 *	Added "dispose={delete|keep}" option to control deleting the
 *	output file at the end of testing.  "delete" is the default
 *	action, to avoid leaving test files around.
 *
 * Novemeber 9, 1989 by Robin Miller.
 *	Add additional delay commands (cdelay=, rdelay=, wdelay=) to
 *	help with testing and debugging various devices.  These aren't
 *	normally enabled and don't appear in the help section.
 *
 * Novemeber 6, 1989 by Robin Miller.
 *	Allow start delay (sdelay=) and end delay (edelay=) to be used
 *	with any device, not just terminals.  The parallel output device
 *	needs an end delay to allow the parallel input device to reopen
 *	during multiple passes (can't disable cycle request).
 *
 * November 3, 1989 by Robin Miller.
 *	Change open mode for output files from read/write (O_RDWR) to
 *	write only (O_WRONLY) so write only devices can be tested.
 *
 * October 16, 1989 by Robin Miller.
 *	Made flushing of input queue before testing optional via
 *	the "enable=flush" command.  This caused a timing problem
 *	where occasionally the writer started before the reader.
 *
 * October 4, 1989 by Robin Miller.
 *	Added fsync() system call to force in-core buffers to disk
 *	before reporting statistics after writing the file.
 *
 * September 27, 1989 by Robin Miller.
 *	Display statistics at end of each read/write pass by default.
 *	Added enable/disable=pstats to control end of pass statistics.
 *	Added enable/disable=compare to control data comparisions.
 *
 * September 18, 1989 by Robin Miller.
 *	Allow both an input and an output device to be specified.
 *	This was done mainly for terminal devices so a child could
 *	be forked to allow proper synchronization.
 *
 * July 25, 1989 by Robin Miller.
 *	Added support for testing terminal ports.  New options are:
 *		speed=, timeout=, parity=, flow=, sdelay=, edelay=
 *
 * January 25, 1989 by Robin Miller.
 *	Added reporting of pass statistics if debugging is enabled
 *	so write/read elapsed times can be determined.
 *
 * December 8, 1988 by Robin Miller.
 *	Added catching of software terminate signal (SIGTERM) so the
 *	final statistics will be displayed before exiting.
 */

#if defined(ultrix) || defined(sun)
extern int sys_nerr;
extern char *sys_errlist[];
#endif /* defined(sun) */

/*
 * Forward References:
 */
#if defined(WIN32)

DWORD resume_thread(HANDLE thread_handle);
DWORD suspend_thread(HANDLE thread_handle);
void TerminateParentThread(void);
void winAlarm(u_int sec);
void ReadStdin(void);

/*
 * Local Storage:
 */
DWORD ParentThreadId;			/* Parents' thread ID.		*/
HANDLE ParentThread;			/* Detect parent during exit.	*/
HANDLE exitMutex;			/* Synchronize terminate exit.	*/

#elif defined(THREADS)

pthread_attr_t	thread_attrs;
pthread_attr_t	*tattrp = &thread_attrs;
pthread_t ParentThreadId;               /* The parents' thread ID.      */
pthread_t MonitorThreadId;              /* The monitoring thread ID.    */

void *MonitorThread(void *arg);
int StartMonitorThread(int interval);
int SetThreadCancelType(int cancel_type);
char *decodeCancelType(int cancel_type);

#endif /* defined(WIN32) */

static int match(char *s);
static u_long number(int base);
static large_t large_number(int base);
void parse_args(int argc, char **argv);
static time_t time_value(void);
void SignalHandler(int signal_number);

/*
 * Variable Declarations:
 */
#if !defined(AIO)
int	aio_bufs = 0; /* Normally declared in dtaio.c, but for parsing. */
#endif /* !defined(AIO) */

#if defined(MUNSA)
bool	munsa_flag = FALSE;		/* if TRUE enable MUNSA features*/
dlm_lock_mode_t  munsa_lock_type = DLM_NLMODE; /* default munsa lock type*/
dlm_lock_mode_t  input_munsa_lock_type = DLM_PRMODE; /* lock for input file*/
dlm_lock_mode_t  output_munsa_lock_type = DLM_PWMODE; /* lock for output file*/
char	     *resnam;
int	     resnlen, i;
dlm_lkid_t   lkid;
dlm_status_t l_stat;
dlm_nsp_t    nsp;
#endif /* defined(MUNSA) */

u_int32 pattern = DEFAULT_PATTERN;	/* Default data pattern.	*/
bool	user_pattern = FALSE;		/* Flags user specified pattern	*/
bool	unique_pattern = TRUE;		/* Unique pattern per process.	*/
char	*pattern_string;		/* Pointer to pattern string.	*/
char	*prefix_string;			/* User defined prefix string.	*/
/* TODO:
 * Temporary until all functions accept dip and resulting fields.
 */
char	*fprefix_string;		/* Formatted prefix string.	*/
int	fprefix_size;			/* The formatted prefix size.	*/
/* XXX */
bool	aio_flag = FALSE;		/* Asynchronous I/O (AIO) flag.	*/
int	align_offset = 0;		/* Align buffer at this offset.	*/

bool	dumpall_flag = FALSE;		/* Controls dumping all blocks.	*/
u_int	max_bad_blocks = MAXBADBLOCKS;	/* Maximum bad blocks to dump.	*/
bfmt_t	boff_format = HEX_FMT;		/* Buffer offset data format.	*/
dfmt_t	data_format = NONE_FMT;		/* Data display format.		*/
bufmodes_t buffer_modes[NUM_BUFMODES];	/* The I/O buffer modes.	*/
int	bufmode_index = 0;		/* Current buffer mode index.	*/
int	bufmode_count = 0;		/* The user specified bufmodes.	*/

#if defined(TTY)
speed_t	baud_rate;			/* The user selected baud rate.	*/
#endif /* defined(TTY) */
size_t	block_size = BLOCK_SIZE;	/* Default block size to use.	*/
size_t	data_size;			/* Data buffer size + pad bytes	*/
size_t	dump_limit = DEFAULT_DUMP_LIMIT;/* The dump buffer data limit.	*/
u_int32	device_size = 0;		/* Default device block size.	*/
pid_t	child_pid;			/* For the child process ID.	*/
pid_t	process_id;			/* Our current process ID.	*/
bool	bypass_flag = FALSE;		/* Bypass (some) sanity checks.	*/
bool	cerrors_flag = TRUE;		/* Report device close errors.	*/
bool	child_flag = FALSE;		/* This is a child process.	*/
bool	compare_flag = TRUE;		/* Controls data comparisions.	*/
bool	core_dump = FALSE;		/* Generate core dump on errors	*/
bool	debug_flag = FALSE;		/* Enable debug output flag.	*/
bool	Debug_flag = FALSE;		/* Verbose debug output flag.	*/
bool	eDebugFlag = FALSE;		/* End of file debug flag.	*/
bool	fDebugFlag = FALSE;		/* File operations debug flag.	*/
bool	pDebugFlag = FALSE;		/* Process related debug flag.	*/
bool	rDebugFlag = FALSE;		/* Random (seek) debug flag.	*/
bool	tDebugFlag = FALSE;		/* Timer (alarm) debug flag.	*/
bool	delete_per_pass = FALSE;	/* Delete files per pass flag.	*/
v_bool	dio_flag = FALSE;               /* Solaris/Win32 Direct I/O.    */
/* File system caching is enabled by default, these control each. */
v_bool	read_cache_flag = TRUE;		/* FS Read cache control flag.	*/
v_bool	write_cache_flag = TRUE;	/* FS Write cache control flag.	*/
bool	dump_flag = TRUE;		/* Dump data buffer on errors.	*/
#if defined(EEI)
bool	eei_flag = TRUE;		/* Extended Error Information.	*/
bool	eei_resets = FALSE;		/* Handle device reset errors.	*/
#endif /* defined(EEI) */
v_bool	end_of_file = FALSE;		/* End of file detected.	*/
bool	eof_status = FALSE;		/* Controls EOF exit status.	*/
vu_long	error_count;			/* Number of errors detected.	*/
u_long	error_limit = 1;		/* Number of errors tolerated.	*/
int	exit_status = SUCCESS;		/* Normal success exit status.	*/
u_int	dir_limit;			/* The number of directories.	*/
u_int	file_limit;			/* Number of files/directory.	*/
u_int	subdir_depth;			/* The subdirectory depth.	*/
u_int	subdir_limit;			/* Number of subdirectories.	*/
large_t max_data;	 		/* Maximum data for all files.	*/
u_int	max_files;			/* Maximum files (all dirs).	*/
bool	forked_flag = FALSE;		/* Forked child process flag.	*/
bool	fsincr_flag = FALSE;		/* File size increment flag.	*/
bool	fsync_flag = UNINITIALIZED;	/* fsync() after writes flag.	*/
u_int	fsync_frequency;		/* The fsync frequency.		*/
bool    fsalign_flag = FALSE;           /* Align FS offsets and sizes.  */
bool	multiple_devs = FALSE;		/* Multiple devices specified.	*/
bool	multiple_dirs = FALSE;		/* Multiple directories flag.	*/
bool	multiple_files = FALSE;		/* Multiple files flag.		*/
OFF_T	end_position;			/* End position for stepping.	*/
OFF_T	file_position;			/* File position to lseek to.	*/
OFF_T	last_position;			/* Last position lseeked to.	*/
OFF_T	step_offset;			/* Step offset for disk seeks.	*/
bool	flush_flag = TRUE;		/* Flush tty input/output queue	*/
bool	keep_existing = TRUE;		/* Don't delete existing files.	*/
bool	header_flag = TRUE;		/* The log file header flag.	*/
bool	hazard_flag = FALSE;		/* Emit Hazard RPCLOGn: prefix.	*/
int	history_data_size = DEFAULT_HISTORY_DATA_SIZE;
int	history_size = 0;		/* Size of the history array.	*/
bool	history_dump = FALSE;		/* Dump history when exiting.	*/
bool	history_timing = FALSE;		/* History timing control flag.	*/
bool	noprog_flag = FALSE;		/* Check for no I/O progress.	*/
bool	prefill_flag = TRUE;		/* I/O buffer pre-fill flag.	*/
bool	trigargs_flag = TRUE;		/* Trigger arguments flag.	*/
bool	unique_log = FALSE;		/* Make the log file unique.	*/
bool	unique_file = FALSE;		/* Make output file unqiue.	*/
bool	user_incr = FALSE;		/* User specified incr count.	*/
bool	user_min = FALSE;		/* User specified min size.	*/
bool	user_max = FALSE;		/* User specified max size.	*/
bool    user_ralign = FALSE;            /* User specified random align. */
size_t	incr_count;			/* Record increment byte count.	*/
size_t	min_size;			/* The minimum record size.	*/
size_t	max_size;			/* The maximum record size.	*/
bool	lbdata_flag = FALSE;		/* Logical block data flag.	*/
static u_int32 lbdata_addr = 0;		/* Starting logical block addr.	*/
size_t lbdata_size = 0;			/* Logical block data size.	*/
bool	user_lbdata = FALSE;		/* User specified starting lba.	*/
bool	user_lbsize = FALSE;		/* User specified lbdata size.	*/
bool	user_position = FALSE;		/* User specified file position.*/
bool	incr_pattern = FALSE;		/* Incrementing data pattern.	*/
u_int32	iot_seed = IOT_SEED;		/* The default IOT seed value.	*/
static u_int32 iot_seed_per_pass;	/* The per pass IOT seed value.	*/
static bool iot_pattern = FALSE;	/* IOT test pattern selected.	*/
bool	iotune_alarm = FALSE;		/* IO tune alarm enabled.	*/
bool	logappend_flag = TRUE;		/* Append to log file flag.	*/
bool	logdiag_flag = FALSE;		/* Log diagnostic messages.	*/
bool	logpid_flag = FALSE;		/* Log process ID w/messages.	*/
bool	syslog_flag = FALSE;		/* Log errors to syslog.	*/
bool	loopback = FALSE;		/* Loopback to the same device.	*/
bool	loop_on_error = FALSE;		/* The loop on error flag.	*/
bool	micro_flag = FALSE;		/* Controls micro-second delay.	*/
bool	mmap_flag = FALSE;		/* Do memory mapped file I/O.	*/
#if defined(TTY)
bool	modem_flag = FALSE;		/* Testing tty modem control.	*/
#endif /* defined(TTY) */
bool	media_changed = FALSE;		/* Shows when media changed.	*/
bool	multi_flag = FALSE;		/* Multi-volume media flag.	*/
v_int	multi_volume = 1;		/* Multi-volume media count.	*/
int	nice_value = 0;			/* Apply nice value when <> 0.	*/
int	open_flags = 0;			/* Common file open flags.	*/
int	wopen_flags = 0;		/* Additional write open flags.	*/
#if defined(WIN32)
int	ropen_mode = GENERIC_READ ;	/* The read open mode to use.	*/
int	wopen_mode = GENERIC_WRITE;	/* The write open mode to use.	*/
#else /* !defined(WIN32) */
int	ropen_mode = O_RDONLY;		/* The read open mode to use.	*/
int	wopen_mode = O_WRONLY;		/* The write open mode to use.	*/
#endif /* defined(WIN32) */
bool	pad_check = TRUE;		/* Check data buffer pad bytes.	*/
bool	spad_check = FALSE;		/* Check short record pad bytes.*/
u_long	pass_count;			/* Number of passes completed.	*/
u_long	pass_limit = 1UL;		/* Default number of passes.	*/
u_long	skip_count;			/* # of input record to skip.	*/
u_long	seek_count;			/* # of output records to seek.	*/
large_t	record_limit;			/* Max # of records to process.	*/
u_long  records_processed;		/* # of full records processed.	*/
u_long	partial_records;		/* # of partial records proc'ed	*/
large_t	data_limit;			/* Total data limit per pass.	*/
u_long	random_align = 0UL;		/* Random I/O offset alignment.	*/
large_t	rdata_limit = 0;		/* The random I/O data limit.	*/
large_t	total_bytes;			/* Total bytes transferred.	*/
large_t total_bytes_read;		/* Total bytes read.		*/
large_t total_bytes_written;		/* Total bytes written.		*/
vu_long	total_errors;			/* Total errors (all passes).	*/
large_t	total_files;			/* Total files (all passes).	*/
large_t total_files_read;		/* Total files read.		*/
large_t total_files_written;		/* Total files written.		*/
large_t	total_records;			/* Total records (all passes).	*/
large_t total_records_read;             /* Total records read (test).   */
large_t total_records_written;          /* Total records written (test) */
u_long	total_partial;			/* Total partial records.	*/
u_long	total_partial_reads;   		/* Total partial record reads.  */
u_long	total_partial_writes;		/* Total partial record writes. */
u_long	warning_errors;			/* Total non-fatal error count.	*/
int	term_wait_retries = 0;		/* Termination wait retries.	*/
int	term_wait_timeout = 0;		/* Termination wait timeout.	*/
bool	pstats_flag = TRUE;		/* Display per pass statistics.	*/
bool	raw_flag = FALSE;		/* The read after write flag.	*/
bool	reread_flag = FALSE;		/* Force a re-read after raw.	*/
bool	rotate_flag = FALSE;		/* Force data buffer rotating.	*/
int	rotate_offset = 0;		/* Current rotate buffer offset	*/
bool	sighup_flag = TRUE;		/* Hangup signal control.	*/
bool	sparse_flag = TRUE;		/* Sparse file attribute (win).	*/
bool	prealloc_flag = TRUE;		/* Preallocate file blocks (win)*/
bool	stats_flag = TRUE;		/* Display total statistics.	*/
bool	stdin_flag = FALSE;		/* Presume not reading stdin.	*/
bool	stdout_flag = FALSE;		/* Presume not writing stdout.	*/
volatile bool terminating_flag = FALSE;	/* Program terminating flag.	*/
bool    timestamp_flag = FALSE;         /* Timestamp each data block.   */
bool	ttyport_flag = FALSE;		/* Input/output is a terminal.	*/
bool	verbose_flag = TRUE;		/* Verbose messages output.	*/
bool	verify_flag = TRUE;		/* Verify the read/write data.	*/
bool	verify_only = FALSE;		/* Verify of copied data flag.	*/
char	*dtcmd;				/* Original dt command line.	*/
char	*cmd_line;			/* Copy of our command line.	*/
char	*log_file;			/* Pointer to log file name.	*/
char	*log_format;			/* Log file with format strings	*/
char	*log_buffer;			/* Pointer to log file buffer.	*/
char	*log_bufptr;			/* Pointer into log buffer.	*/
char	*msg_buffer;			/* Diagnostic message buffer.	*/
char	*dirpath;			/* The base directory path.	*/
char	*dirprefix = DIR_PREFIX;	/* The directory prefix.	*/
char	*iotune_file;			/* Pointer to IO tune file.	*/
char	*input_file;			/* Pointer to input file name.	*/
char	*output_file;			/* Pointer to output file name.	*/
char	*pattern_file;			/* Pointer to pattern file name	*/
char	*stop_on_file;			/* Stop on file existance.	*/
u_char	*pattern_buffer;		/* Pointer to pattern buffer.	*/
u_char	*pattern_bufptr;		/* Pointer into pattern buffer.	*/
u_char	*pattern_bufend;		/* Pointer to end of pat buffer	*/
u_char	*base_buffer;			/* Base address of data buffer.	*/
u_char	*data_buffer;			/* Pointer to data buffer.	*/
u_char	*mmap_buffer;			/* Pointer to mmapped buffer.	*/
u_char	*mmap_bufptr;			/* Pointer into mmapped buffer.	*/
u_char	*verify_buffer;			/* The data verification buffer.*/
size_t	patbuf_size;			/* The pattern buffer size.	*/
int	pattern_size;			/* User specified pattern size.	*/
int	prefix_size;			/* User defined prefix size.	*/
int	page_size = 0;			/* Define number of bytes/page.	*/
#if defined(HP_UX)
u_int   qdepth = 0xFFFFFFFF;            /* Value to set queue depth to. */
#endif
u_int	cancel_delay = 0;		/* Time to delay after cancel.	*/
u_int	kill_delay = 3;			/* Delay before children kill.	*/
u_int	cdelay_count = 0;		/* Delay before closing file.	*/
u_int	edelay_count = 0;		/* Delay between multiple passes*/
u_int	rdelay_count = 0;		/* Delay before reading record.	*/
u_int	sdelay_count = 0;		/* Delay before starting test.	*/
u_int	wdelay_count = 0;		/* Delay before writing record.	*/
u_int	tdelay_count = 1;		/* Child terminate delay count.	*/
u_int	random_seed = 0;		/* Seed for random # generator.	*/
u_int	retry_delay = RETRY_DELAY;	/* The default retry delay.	*/
bool	retryDC_flag = TRUE;		/* Retry data corruptions.	*/
bool	retryIO_flag = FALSE;		/* Retry I/O failures.		*/
bool	user_rseed = FALSE;		/* Flags user specified rseed.	*/
bool	max_capacity = FALSE;		/* Use max capacity from IOCTL.	*/
large_t	user_capacity;			/* The user set drive capacity.	*/
bool	variable_flag = FALSE;		/* Variable block size flag.	*/
bool	volumes_flag = FALSE;		/* Flags the volumes option.	*/
int	volume_limit = 0;		/* Number of volumes to process.*/
vu_long	volume_records = 1;		/* The last volume record limit.*/

#if defined(TTY)
enum opt softcar_opt = OPT_NONE;	/* Leave tty soft carrier alone.*/
enum flow flow_type = XON_XOFF;		/* Terminal flow type to use.	*/
#endif /* defined(TTY) */
enum iodir  io_dir  = FORWARD;		/* Default is forward I/O.	*/
enum iomode io_mode = TEST_MODE;	/* Default to testing mode.	*/
enum iotype io_type = SEQUENTIAL_IO;	/* Default to sequential I/O.	*/
enum dispose dispose_mode = DELETE_FILE;/* Output file dispose mode.	*/
enum onerrors oncerr_action = CONTINUE;	/* The child error action.	*/
					/* Error limit controls tests.	*/
enum statslevel stats_level = STATS_FULL; /* Type of statistics to report.*/
enum trigger_type trigger = TRIGGER_NONE; /* Trigger for corruptions.   */
char *trigger_cmd = NULL;               /* The users' trigger command.  */
char *trigger_args = NULL;		/* Extra user trigger arguments	*/
#if defined(SOLARIS) || defined(OSFMK) || defined(__QNXNTO__) || defined(AIX) || defined(__linux__)
clock_t hz;
#else
clock_t	hz = HZ;			/* Default clock ticks / second	*/
#endif

#if defined(TTY)
/*
 * Values for Terminal (serial) Line Testing:
 */
u_short	tty_timeout = 3*10;		/* Default tty timeout (3 sec).	*/
					/* VTIME = 0.10 second interval	*/
bool	tty_minflag = FALSE;		/* User specified VMIN value.	*/
u_short	tty_minimum = 0;		/* The tty minimum (VMIN) value	*/
#endif /* defined(TTY) */

FILE	*efp;				/* Default error data stream.	*/
FILE	*ofp;				/* Default output data stream.	*/
HANDLE	pfd = NoFd;			/* Pattern file descriptor.	*/
char	*cmdname;			/* Pointer to our program name.	*/
char	*dtpath;			/* Path to our dt executable.	*/
char	*string;			/* Pointer to argument string.	*/

/*
 * Pointers to various device information.
 */
struct dinfo *active_dinfo;		/* Active device information.	*/
struct dinfo *input_dinfo;		/* Input device information.	*/
struct dinfo *output_dinfo;		/* Output device information.	*/
struct dtype *input_dtype;		/* The input device type info.	*/
struct dtype *output_dtype;		/* The output device type info.	*/

/*
 * System time information.
 */
clock_t start_time, end_time, pass_time; /* Per pass elapsed time.	*/
struct tms stimes, ptimes, etimes;	/* For user / system times.	*/
#if defined(DEC)
bool table_flag = FALSE;		/* Table control flag.		*/
struct tbl_sysinfo s_table, p_table, e_table;	/* Table information.	*/
#endif /* defined(DEC) */
#if defined(_BSD)
union wait child_status;		/* For child exit status.	*/
#else /* !defined(_BSD) */
int	child_status;			/* For child exit status.	*/
#endif /* defined(_BSD) */

/*
 * Program run time information.
 */
time_t  alarmtime;                      /* The alarm interval time.     */
time_t	keepalive_time;			/* The keepalive time.		*/
time_t	iotune_frequency = 1;		/* The default IO tune freuency */
time_t	noprogtime;			/* The no progress time (secs).	*/
time_t	noprogttime;			/* The no prog trigger time.	*/
time_t	runtime;			/* The program run time.	*/
time_t  runtime_end;                    /* The ending future runtime.   */
time_t	elapsed_time;			/* Amount of time program ran.	*/
time_t	program_start, program_end;	/* Program start & end times,	*/
time_t	error_time;			/* Time last error occurred.	*/
bool	TimerActive;			/* Set after timer activated.	*/
bool	TimerExpired;			/* Set after timer has expired.	*/

bool	StdinIsAtty;			/* Standard input isatty flag.  */
bool	StdoutIsAtty;			/* Standard output isatty flag. */
bool	StderrIsAtty;			/* Standard error isatty flag.  */

/*
 * Default alarm message is per pass statistics, user can override.
 */
char    *keepalive;
char    *keepalive0 = "%d Stats: mode %i, blocks %l, %m Mbytes, pass %p/%P,"
                      " elapsed %t";
char    *keepalive1 = "%d Stats: mode %i, blocks %L, %M Mbytes, pass %p/%P,"
                      " elapsed %T";
                                        /* Default keepalive messages.  */
/*
 * When stats is set to brief, these message strings get used:
 */
char    *pkeepalive;
char    *pass_msg = "pass %p/%P, %l blocks, %m Mbytes, %c records,"
                    " errors %e/%E, elapsed %t";
                                        /* Per pass keepalive message.  */
char    *tkeepalive;
char    *totals_msg = "%d Totals: %L blocks, %M Mbytes,"
                      " errors %e/%E, passes %p/%P, elapsed %T";
                                        /* Totals keepalive message.    */ 
bool    user_keepalive = FALSE;         /* User specified keepalive flag*/
bool    user_pkeepalive = FALSE;        /*        ... ditto ...         */
bool    user_tkeepalive = FALSE;        /*        ... ditto ...         */
char	*user_runtime;			/* User specific runtime string	*/

#if defined(NETAPP)
    char *ontap_cserver;		/* The cluster server name.	*/
    char *ontap_nodes;			/* The cluster node name(s).	*/
    char *ontap_username;		/* The user name (optional).	*/
    char *ontap_password;		/* The password (optional).	*/
    char *ontapi_path;			/* The 'ontapi' path (optional)	*/
#endif /* defined(NETAPP) */

/*
 * Data patterns used for multiple passes.
 */
u_int32 data_patterns[] = {
	DEFAULT_PATTERN,
	0x00ff00ffU,
	0x0f0f0f0fU,
	0xc6dec6deU,
	0x6db6db6dU,
	0x55555555U,
	0xaaaaaaaaU,	/* Complement of previous data pattern.		 */
	0x33333333U,	/* Continuous worst case pattern (media defects) */
	0x26673333U,	/* Frequency burst worst case pattern #1.	 */
	0x66673326U,	/* Frequency burst worst case pattern #2.	 */
	0x71c7c71cU,	/* Dibit worst case data pattern.		 */
	0x00000000U,
	0xffffffffU,
};
int npatterns = sizeof(data_patterns) / sizeof(u_int32);

/*
 * This table is indexed by the operation type (enum optype):
 */
optiming_t optiming_table[] =
{
  /*  Operation Type  Control Flag      Name       */
    { NONE_OP,		FALSE,		NULL       },
    { OPEN_OP,		TRUE,		"open"     },
    { CLOSE_OP,		TRUE,		"close"    },
    { READ_OP,		TRUE,		"read"     },
    { WRITE_OP,		TRUE,		"write"    },
    { IOCTL_OP,		TRUE,		"ioctl"    },
    { FSYNC_OP,		TRUE,		"fsync"    },
    { MSYNC_OP,		TRUE,		"msync"    },
    { AIOWAIT_OP,	TRUE,		"aiowait"  }
};

/*
 * The mainline sets this up, need common for logging anywhere!
 */
void
log_header(void)
{
    /*
     * Write the command line for future reference.
     */
    Lprintf("Command Line:\n\n    %c ", getuid() ? '%' : '#');
    Lprintf("%s ", cmd_line);
    Lprintf("\n\t--> %s <--\n", version_str);
    if (debug_flag || Debug_flag || rDebugFlag | tDebugFlag) {
	Lprintf("\n");
    }
    Lflush();
    return;
}

/*
 * main() - Start of data transfer program.
 */
int
main (int argc, char **argv)
{
	struct dinfo *dip;
	struct dtfuncs *dtf;
	char *tmp, *devs;
	int status;
	struct stat sb;
#if defined(WIN32)
        SYSTEM_INFO pgsize;
#endif /* defined(WIN32) */

	efp = stderr;			/* Initialize our error stream.	*/
	ofp = stdout;			/* Initialize our output stream.*/

	/*
	 * Allocate buffer for pre-formatting messages.
	 *
	 * Note: Not safe for threads, so must go with time!
	 */
	msg_buffer = (char *)Malloc(LOG_BUFSIZE);

	dtpath = argv[0];
#if defined(_BSD)
	tmp = rindex (argv[0], '/');
#else /* !defined(_BSD) */
	tmp = strrchr (argv[0], '/');
#endif /* defined(_BSD) */
	cmdname = tmp ? &(tmp[1]) : argv[0];
	process_id = getpid();
#if defined(_QNX_SOURCE) && !defined(_QNX_32BIT)
	page_size = 0;		/* Presume page size doesn't matter. */
#elif defined(_QNX_SOURCE) && defined(_QNX_32BIT)
	page_size = 4096;	/* Not sure how to query for this... */
#elif defined(DEC) || defined(SOLARIS) || defined(__linux__) || defined(SCO) || defined(__NUTC__) || defined(HP_UX) || defined(AIX)
	hz = sysconf(_SC_CLK_TCK);
	page_size = sysconf(_SC_PAGESIZE);
#elif defined(WIN32)
	GetSystemInfo(&pgsize);
	process_id = GetCurrentProcessId();
	page_size = pgsize.dwPageSize;
	ParentThreadId = GetCurrentThreadId();
	/*
	 * Convert the pseudo handle into a real handle by duplicating!
	 */
	if (!DuplicateHandle(GetCurrentProcess(), GetCurrentThread(),
		     GetCurrentProcess(), &ParentThread, 0, TRUE, DUPLICATE_SAME_ACCESS)) {
	    Fprintf("DuplicateHandle() failed\n");
	}
	/*
	 * Create a mutex to synchronize multiple threads when exiting.
	 */
	exitMutex = CreateMutex(NULL,	/* Default security attributes. */
				FALSE,	/* Initially not owned. */
				NULL);	/* Unnamed mutex. */
	if (exitMutex == NULL) {
	    report_error("CreateMutex() failed", FALSE);
	    terminate(FAILURE);
	}
#else /* !defined(_QNX_SOURCE) && !defined(_QNX_32BIT) */
	page_size = getpagesize();
#endif /* defined(_QNX_SOURCE) && !defined(_QNX_32BIT) */
#if defined(OSFMK) || defined(__QNXNTO__) || defined(WIN32)
	hz = CLK_TCK;			/* Actually a libc function. */
#endif /* defined(OSFMK) || defined(__QNX_NTO__) defined(WIN32) */

#if defined(LOG_DIAG_INFO)
	/*
	 * Allow DT_LOG_DIAG to enable logging diagnostic information.
	 */
	if (tmp = getenv ("DT_LOG_DIAG")) {
	    logdiag_flag = TRUE;
	}
#endif /* defined(LOG_DIAG_INFO) */

	/* That's right, one day we'll have an interactive mode! */
        StdinIsAtty = isatty (fileno(stdin));
        StdoutIsAtty = isatty (fileno(stdout));
        StderrIsAtty = isatty (fileno(stderr));

	data_limit = INFINITY;		/* Set to maximum limit. */

	parse_args (argc, argv);

#if defined(NETAPP)
	if ( (trigger == TRIGGER_ZAPI_PANIC) && !ontap_cserver) {
	    LogMsg(efp, logLevelCrit, 0,
		   "Please specify the ONTAP cserver for ZAPI calls!\n");
	    exit (FATAL_ERROR);
	}
#endif /* defined(NETAPP) */

	if (!input_file && !output_file) {
	    LogMsg (efp, logLevelCrit, 0,
		    "You must specify an input file, an output file, or both.\n");
	    exit (FATAL_ERROR);
	}
	devs = (input_file) ? input_file : output_file;
	if ( EQS(devs, ",") ) {
	    multiple_devs = TRUE;
	}

	/*
	 * Catch some of the pattern option conflicts, misleading!
	 */
	if ((iot_pattern || incr_pattern || pattern_size) && pattern_file) {
	    LogMsg (efp, logLevelCrit, 0,
		    "Multiple pattern options selected, please choose only one!\n");
	    exit (FATAL_ERROR);
	}

	/*
	 * Set the default display data format, if not specified by the user.
	 */
	if (data_format == NONE_FMT) {
	    data_format = BYTE_FMT;
	}
#if defined(WIN32)
	/*
	 * TODO - Not sure why this was added, India did the initial port.
	 */
	if (hazard_flag) {
	    ReadStdin();
	}
#endif /* defined(WIN32) */

	/*
	 * Options parsed, validate options, do initialization, and open
	 * input & output files to be tested.
	 */
	program_start = time((time_t) 0);	/* Record our start time. */

	/*
	 * If a log file was specified, redirect stderr to that file.
	 */
	if (log_file) {
	    ofp = efp;			/* Send all output to stderr. */
            /*
             * Note: The child flag is enabled for children processes.
             */
	    if (unique_log && child_flag) {
		char *logp = NULL;	/* Threaded dt will cleanup! */
		/* Add the prcoess ID if not part of the log format. */
                if ( strstr(log_file, "%pid") == (char *) 0 ) {
		    logp = log_file = make_unique_file(log_file);
                }
		/* Translate any other format strings. */
                log_file = FmtLogFile(log_file);
		if (logp) free(logp);
	    } else if (strstr (log_file, "%")) {
		log_format = strdup(log_file);
                log_file = FmtLogFile(log_file);
            }
	    if (freopen (log_file, (logappend_flag) ? "a" : "w", efp) == NULL) {
	        Perror ("freopen() of %s failed, exiting...", log_file);
	        exit (FAILURE);
	    }
	}

	/*
	 * Make stderr buffered, so timing is not affected by output.
	 */
	log_buffer = (char *)Malloc(LOG_BUFSIZE);

	/*
	 * The concept here is simple, set stderr buffered so multiple processes
	 * don't have their output intermixed.  This piece of code has been very
	 * problematic, so it you have problems with garbled output, remove it.
	 */
	log_bufptr = log_buffer;

	/*
	 * Since stderr is normally unbuffered, we make it buffered here.
	 * The reason for this is to keep text from getting fragnmented.
	 */
	if ( StderrIsAtty ) {
	    char *stderr_buffer = (char *)Malloc(LOG_BUFSIZE);
	    /*
	     * Can't use log buffer, or we get undesirable results :-)
	     */
	    if (setvbuf(efp, stderr_buffer, _IOFBF, LOG_BUFSIZE) < 0) {
		Perror ("setvbuf() failed, exiting...");
		exit (FAILURE);
	    }
	}

	/*
	 * Save the command line for logging later (log file or syslog).
	 */
	{
	    int arg;
	    char *buffer, *bp;

	    bp = buffer = (char *)Malloc(LOG_BUFSIZE);
	    for (arg = 0; arg < argc; arg++) {
		char *opt = argv[arg];
		if ( EQS(opt, "keepalive=") ||
		     EQS(opt, "prefix=")    ||
		     EQS(opt, "trigger=cmd:") ) {
		    /* Quote the entire option! */
		    bp += Sprintf(bp, "\"%s\" ", argv[arg]);
		} else {
		    bp += Sprintf(bp, "%s ", argv[arg]);
		}
	    }
	    sprintf(bp, "\n");
	    cmd_line = Malloc(strlen(buffer)+1);
	    strcpy(cmd_line, buffer);
	    free(buffer);
	}

	/*
	 * Write the command line to the log file, if one exists.
	 */
	if (log_file && header_flag) log_header();

	/*
	 * Users can also use "enable=logpid" to log this with messages.
	 */
	if (debug_flag) {
	    Printf("Parent process ID is %d\n", process_id);
	}

	/*
	 * Since Windows does NOT have a fork() API, we will doing
	 * a CreateProcess() and need the original command line,
	 * without options like procs= (of course).
	 * 
	 * Note: We also need this for *nix dt for multiple devices.
	 */
	if (num_procs || num_slices || multiple_devs) {
	    int arg;
	    char *bp;
	    bp = dtcmd = (char *)Malloc(LOG_BUFSIZE);
	    for (arg = 0; arg < argc; arg++) {
		char *opt = argv[arg];
		/* Options to skip. */
		if ( !multiple_devs && EQL(opt, "procs=", 6) ) {
		    continue;
		} else if ( multiple_devs &&
			    (EQL(opt, "if=", 3) && EQS(opt, ",")) ||
			    (EQL(opt, "of=", 3) && EQS(opt, ",")) ) {
		    continue; /* Skip input or output devices. */
		} else if ( EQS(opt, "keepalive=") ||
			    EQS(opt, "prefix=")    ||
			    EQS(opt, "trigger=cmd:") ) {
		    /* Quote the entire option, hope this works. */
		    bp += sprintf(bp, "\"%s\" ", argv[arg]);
		} else {
		    bp += sprintf(bp, "%s ", argv[arg]);
		}
	    }
	}

	/* Multiple Device Support! */
	if ( multiple_devs ) { /* Multiple devices? */
	    if ( start_devs() ) {
		await_procs();	/* Parent waits. */
		if (debug_flag || pDebugFlag) {
		    Printf ("Parent exiting with status %d...\n", exit_status);
		}
		exit (exit_status);
	    }
	}


	/*
	 * Multiple Directory Sanity Checks:
	 */
	multiple_dirs = (dirpath || dir_limit || subdir_limit || subdir_depth);
	multiple_files = (dirpath || dir_limit || subdir_limit || subdir_depth || file_limit);

	if ( dirpath 				          &&
	     ((input_file && strrchr(input_file, DIRSEP)) ||
	      (output_file && strrchr(output_file, DIRSEP))) ) {
	    LogMsg (efp, logLevelCrit, 0,
		    "Please do not specify an if/of= path with dir= option!\n");
	    exit (FATAL_ERROR);
	}

	/*
	 * If we're doing multiple files, set the dirpath from the file path.
	 */
	if ( !dirpath && multiple_files ) {
	    char *p;
	    if (input_file && (p = strrchr(input_file, DIRSEP)) ) {
		*p = '\0';
		dirpath = input_file;
		input_file = ++p;
		if (debug_flag || fDebugFlag) {
		    Printf("Directory Path: %s, Input File: %s\n",
			   dirpath, input_file);
		}
	    }
	    if (output_file && (p = strrchr(output_file, DIRSEP)) ) {
		*p = '\0';
		dirpath = output_file;
		output_file = ++p;
		if (debug_flag || fDebugFlag) {
		    Printf("Directory Path: %s, Output File: %s\n",
			   dirpath, output_file);
		}
	    }
	}

	/*
	 * With multiple files, we'll create a top level directory "d0",
	 * by default, to place files. This allows multiple procs with
	 * just a mount point to work properly (help out automation).
	 */
	if (multiple_dirs) {
	    char topdir[PATH_BUFFER_SIZE];
	    if (dirpath) {
		(void)sprintf(topdir, "%s%c%s0", dirpath, DIRSEP, dirprefix);
	    } else {
		(void)sprintf(topdir, "%s0", dirprefix);
	    }
	    dirpath = strdup(topdir);
	}

	/*
	 * Disallow both seek type options, to simplify test loops.
	 */
	if ( (io_dir == REVERSE) && (io_type == RANDOM_IO) ) {
	    LogMsg (efp, logLevelCrit, 0,
		    "Please specify one of iodir=reverse or iotype=random, not both!\n");
	    exit (FATAL_ERROR);
	}

	if (slice_num) {
	    if (!num_slices) {
		LogMsg (efp, logLevelCrit, 0,
			"Please specify number of slices with slice option!\n");
		exit (FATAL_ERROR);
	    } else if (slice_num > num_slices) {
		LogMsg (efp, logLevelCrit, 0,
			"Please specify slice (%d) <= max slices (%d)\n",
						slice_num, num_slices);
		exit (FATAL_ERROR);
	    }
	}

#if defined(MUNSA)
	if (munsa_flag) {

	    if (input_file && !output_file) {
	        input_munsa_lock_type =  munsa_lock_type;
		if (debug_flag) {
		    Printf ("input_munsa_lock_type = %d\n",
					input_munsa_lock_type);
		}
	    }

	    if (output_file && !input_file) {
		if ((munsa_lock_type == DLM_PWMODE) ||
		    (munsa_lock_type == DLM_EXMODE)) {
		    output_munsa_lock_type =  munsa_lock_type;
		    if (debug_flag) {
			Printf ("output_munsa_lock_type = %d\n",
					output_munsa_lock_type);
		    }
		} else {
		    LogMsg (efp, logLevelCrit, 0,
			    "invalid write lock type it should be pw,ex\n");
		    exit(FATAL_ERROR);
		}
	    }

	    if (input_file && output_file) {
		input_munsa_lock_type =  DLM_PRMODE;
		output_munsa_lock_type =  DLM_PWMODE;
		if (debug_flag) {
		    Printf ("input_munsa_lock_type = %d\n",
					input_munsa_lock_type);
		    Printf ("output_munsa_lock_type = %d\n",
					output_munsa_lock_type);
		}
	    }
	}  /*   end if(munsa_flag) ....    */
#endif /* defined(MUNSA) */

	if ( (!input_file || !output_file) &&
	      ((io_mode == COPY_MODE) || (io_mode == VERIFY_MODE)) ) {
	    LogMsg (efp, logLevelCrit, 0,
		    "Copy/verify modes require both input and output devices.\n");
	    exit (FATAL_ERROR);
	}

	/*
	 * When reading multiple tape files, don't require data or record
	 * limits (these will vary).  But when writing multiple tape files,
	 * we need to know how many records or bytes to be written.
	 */
	if (input_file && !output_file && file_limit && !record_limit) {
	    record_limit = INFINITY;
	}

	/*
	 * Check the variable record size parameters.
	 */
	if (min_size && !max_size) max_size = block_size;
	if (block_size < max_size) block_size = max_size;
	/* NOTE: Other checks are done below now... */

	/*
	 * Calculate the data limit if it wasn't specified by the user.
	 */
	if ( (data_limit == INFINITY) &&
	     ( (record_limit != 0L) && (record_limit != INFINITY) ) ) {
	    data_limit = (block_size * record_limit);
	}

	/*
	 * Process the pattern file (if one was specified).
	 */
	if (pattern_file) {
#if defined(WIN32)
	    process_pfile (&pfd, pattern_file, GENERIC_READ);
#else /* !defined(WIN32) */
	    process_pfile (&pfd, pattern_file, O_RDONLY);
#endif /* defined(WIN32) */
	}

	if (min_size && (max_size <= min_size)) {
	    LogMsg (efp, logLevelCrit, 0,
		    "Please specify max count > min count for record sizes.\n");
	    exit (FATAL_ERROR);
	}

	/*
	 * If the pattern is unique, disable prefilling of buffer.
	 */
	if (iot_pattern || lbdata_flag) {
	    prefill_flag = FALSE;
	}
	/*
	 * Verify counts are large enough, to avoid false compare errors.
	 */
	if ( (iot_pattern || lbdata_flag || timestamp_flag) && (block_size < lbdata_size) ) {
	    LogMsg (efp, logLevelCrit, 0,
	"Please specify a block size >= %u (lbdata size) for iot, lbdata, or timestamp options!\n",
								lbdata_size);
	    exit (FATAL_ERROR);
	}

	if ( ((io_mode == COPY_MODE) || (io_mode == VERIFY_MODE)) &&
	     (iot_pattern || lbdata_flag || prefix_string || timestamp_flag) ) {
	    LogMsg (efp, logLevelCrit, 0,
		    "IOT, Lbdata, prefix, & timestamp options disallowed with Copy/Verify options!\n");
	    exit (FATAL_ERROR);
	}

        /*
         * Do special handling of IOT data pattern.
         */
	if (iot_pattern) {
	    size_t size = block_size;
	    u_char *buffer = (u_char *) myalloc(size, 0);
	    setup_pattern (buffer, size);
	    pattern_string = "IOT Pattern";
	    /*
	     * When variable I/O sizes are requested, align sizes to device size
	     * instead of pattern buffer size else max transfer size is used!
	     */
	    fsalign_flag = TRUE;
	    /* IOT takes precedence! */
	    lbdata_flag = FALSE;
	    user_lbdata = FALSE;
	}

	/*
	 * Setup the pattern as a pattern string, so non-modulo
	 * sizeof(u_int) read counts will compare data properly.
	 */
	if (!pattern_buffer) {
	    size_t size = sizeof(u_int32);
	    u_char *buffer = (u_char *)Malloc(size);
	    copy_pattern (pattern, buffer);
	    setup_pattern (buffer, size);
	}

#if defined(TTY)
	/*
	 * The following check was added for tty loopback to same port.
	 * [ A future version may extend this support to other devices. ]
	 */
	if (input_file && output_file) {
	    if (strcmp (input_file, output_file) == 0) {
		loopback = TRUE;	/* Loopback to the same file. */
	    }
	} else if (loopback) {
	    if (!input_file && output_file) input_file = output_file;
	    if (input_file && !output_file) output_file = input_file;
	}
#endif /* defined(TTY) */

	/****************************************************************
	 * Do NOT use the device information pointer (dip) before here! *
	 ****************************************************************/

	/*
	 * Setup the initial device information & validate options.
	 */
	if (input_file) {
	    dip = setup_device_info (input_file, input_dtype);
	    active_dinfo = input_dinfo = dip;
	    dip->di_mode = READ_MODE;
	    dip->di_ftype = INPUT_FILE;
	    dip->di_fsync_flag = FALSE;
	    status = (*dip->di_funcs->tf_validate_opts)(dip);
	    if (status == FAILURE) exit (FATAL_ERROR);
	}
	if (output_file) {
	    dip = setup_device_info (output_file, output_dtype);
	    active_dinfo = output_dinfo = dip;
	    dip->di_mode = WRITE_MODE;
	    dip->di_ftype = OUTPUT_FILE;
	    status = (*dip->di_funcs->tf_validate_opts)(dip);
	    if (status == FAILURE) exit (FATAL_ERROR);
	}

#if 1 /* #if def'ed out for debug only!!! */

        /*
         * When doing random I/O, enable file system alignments, to help
         * prevent false corruptions.  This only affects regular files,
         * and only when read-after-write (raw) is disabled.  When raw
         * is enabled, we don't have to worry about data overwrites,
         * unless POSIX Async I/O (AIO) is enabled.
         */
        active_dinfo = dip = (input_file) ? input_dinfo : output_dinfo;
	/*
	 * Note: The device type may not get setup until device is open'ed below!
	 */
        if ( dip->di_dtype &&
	     ((dip->di_dtype->dt_dtype == DT_DISK)  ||
	      (dip->di_dtype->dt_dtype == DT_BLOCK) ||
	      (dip->di_dtype->dt_dtype == DT_CHARACTER)) ) {
	    /* 
	     * IOT pattern changes with each pass, so disable this unqiueness
	     * if user selects multiple procs with multiple passes, otherwise
	     * false data corruptions may occur when procs on differnet passes.
	     */
	    if (iot_pattern && num_procs && unique_pattern && (runtime || (pass_limit > 1)) ) {
		Printf("Warning: Disabling unique IOT patterns with multiple procs!\n");
		Printf("Please consider using slices= option instead of procs= option.\n");
		dip->di_unique_pattern = unique_pattern = FALSE;
	    }
	}
	/*
	 * Special file system handling.
	 *
	 * Note: This mess needs cleaned up!
	 */
        if ( (dip->di_dtype && (dip->di_dtype->dt_dtype == DT_REGULAR)) &&
	     (io_type == RANDOM_IO) && !bypass_flag &&
	     (!raw_flag || (raw_flag && reread_flag) || aio_flag) ) {
            if (debug_flag || Debug_flag) {
                LogMsg(efp, logLevelWarn, 0,
                       "Enabling FS alignment for sizes and random offsets!\n");
            }
            fsalign_flag = TRUE;    /* Align FS sizes & random offsets. */
            /*
             * Sanity check the pattern size is modulo the device size,
             * otherwise this too will cause false failures w/random I/O.
             */
            if (patbuf_size && dip->di_dsize) {
                if ( ((patbuf_size > dip->di_dsize) && (patbuf_size % dip->di_dsize)) ||
                     ((patbuf_size < dip->di_dsize) && (dip->di_dsize % patbuf_size)) ) {
                    LogMsg(efp, logLevelCrit, 0,
                           "Please specify a pattern size (%u) modulo the device size (%u)!\n",
                            patbuf_size, dip->di_dsize);
                    exit (FATAL_ERROR);
                }
            }
        }
#endif /* #if def'ed out for debug only!!! */

#if defined(sun) && defined(TTY)
	/*
	 * Soft carrier existed on the Sun/386i (Roadrunner) system.
	 * Setting O_NDELAY was necessary to open the terminal line.
	 *
	 * Normally, O_NDELAY must NOT be set, otherwise the terminal
	 * driver returns EWOULDBLOCK on reads if no data is available.
	 *
	 * Note:  Without O_NDELAY, the open will hang on modem lines
	 *        (-CLOCAL) if carrier is not asserted.
	 */
	if (ttyport_flag) {
	    open_flags = O_NDELAY;	/* Incase no soft carrier.	*/
	}
#endif /* defined(sun) && defined(TTY) */

	/*
	 * Catch a couple signals to do elegant cleanup.
	 */
	(void) signal (SIGTERM, SignalHandler);
	(void) signal (SIGINT, SignalHandler);
#if !defined(WIN32)
	(void) signal (SIGHUP, (sighup_flag) ? SignalHandler : SIG_IGN);
	/*
	 * Windows Note: SIGINT is not supported for any Win32 application,
	 * including Windows 98/Me and Windows NT/2000/XP. When a CTRL+C
	 * interrupt occurs, Win32 operating systems generate a new thread
	 * to specifically handle that interrupt.  This can cause a
	 * single-thread application such as UNIX, to become multithreaded,
	 * resulting in unexpected behavior. (and indeed it does!)
	 *
	 * Note: Multiple threads are now handled by terminate().
	 */
	//(void) signal (SIGINT, SignalHandler);
	(void) signal (SIGPIPE, SignalHandler);
#endif /* !defined(WIN32) */
#if (!defined(__MSDOS__) && !defined(WIN32)) || defined(__NUTC__)
	(void) signal (SIGCHLD, SIG_DFL);
#endif

    /*
     * Do multiple slices processing.
     */
    if (num_slices) {
	if (input_file && output_file) {
	    LogMsg (efp, logLevelCrit, 0,
		    "Please specify only an input or output file, not both!\n");
	    exit (FATAL_ERROR);
	}

	/*
	 * Create multiple slices (if requested).
	 */
	active_dinfo = dip = (input_file) ? input_dinfo : output_dinfo;
	if (!dip->di_random_access) {
	    LogMsg (efp, logLevelCrit, 0,
		    "Multiple slices is only supported on random access devices!\n");
	    exit (FATAL_ERROR);
	}
	if ((status = FindCapacity (dip)) == FAILURE) {
	    exit (FATAL_ERROR);
	}
	/*
	 * The remaining work is done when starting the processes.
	 */
	if (slice_num) {
	    init_slice(active_dinfo, slice_num);
	    cur_proc = slice_num;
	    //num_slices = 0; /* Operate on a single slice. */
	} else {
	    if ( start_slices() ) {
		await_procs();
		if (debug_flag || pDebugFlag) {
		    Printf ("Parent exiting with status %d...\n", exit_status);
		}
		exit (exit_status);
	    }
	}
    } /* End of multiple slices processing. */

    /* Until we find the Windows equivalent! */
#if !defined(WIN32)
    if (nice_value) {
	errno = 0;
	if ( (nice(nice_value) == -1) && (errno != 0) ) {
	    Perror("nice() failed");
	    exit (FAILURE);
	}
    }
#endif /* !defined(WIN32) */

	/*
	 * Create multiple processes (if requested).
	 */
	if (num_procs && !loopback && !ttyport_flag) {
	    if ( start_procs() ) {
		await_procs();		/* Parent waits. */
		if (debug_flag || pDebugFlag) {
		    Printf ("Parent exiting with status %d...\n", exit_status);
		}
		exit (exit_status);
	    }
	}

	/*
	 * Open device / Setup system / device specific test information.
	 */
	if (input_file) {
	    int open_mode = (ropen_mode | open_flags);
	    dip = active_dinfo = input_dinfo;
	    if ( dirpath &&
		 ((dip->di_dtype->dt_dtype == DT_MMAP) ||
		  (dip->di_dtype->dt_dtype == DT_REGULAR)) ) {
		int status = setup_directory_info(dip, dirpath);
		if (status == FAILURE) terminate(status); /* For now! */
	    }
	    SetupBufferingMode(dip, &open_mode);
	    if ((*dip->di_funcs->tf_open)(dip, open_mode) == FAILURE) {
		exit (exit_status);
	    }
	    system_device_info (dip);
	    input_dtype = dip->di_dtype;

	    status = (*dip->di_funcs->tf_validate_opts)(dip);
	    if (status == FAILURE) exit (FATAL_ERROR);

	    /*
	     * The new logic for the IOT pattern is to make it unique during
	     * multiple passes. However, during reads it must remain constant.
	     */
	    if (dip->di_iot_pattern) {
		dip->di_unique_pattern = FALSE;	/* Re-reads during all pass use same seed! */
	    }
#if defined(MUNSA)
	    if (munsa_flag) {
		/* first we need to join a namespace */

		l_stat = dlm_nsjoin( getuid(), &nsp, DLM_USER);
		if ((l_stat != DLM_SUCCESS) && (l_stat != DLM_ATTACHED)) {
		    Fprintf ("Can't join namespace\n");
		    dlm_error(0, l_stat);
		}

		resnam = input_file;
		/* now let DLM know what signal to use for blocking routines */
		dlm_set_signal(SIGIO, &i);
		resnlen = strlen(resnam);
		if (debug_flag) {
		    Printf ("dlm_set_signal: i %d\n", i);
		    Printf ("resnam %s\n", resnam);
		    Printf ("grab a NL mode lock\n");
		}
		l_stat = dlm_lock(nsp, 
				  (uchar_t *)resnam,
				  resnlen,
				  0,
				  &lkid,
				  DLM_NLMODE,
				  NULL,
				  DLM_SYNCSTS,
				  0, 0, 0, 0);
		/* NL mode better be granted SYNC status */
		if (l_stat !=  DLM_SYNCH) {
		    if (debug_flag) {
			Fprintf ("dlm_lock failed\n");
		    }
		    dlm_error(&lkid, l_stat);
		}
	    }  /*  end if(munsa_flag)... */
#endif /* defined(MUNSA) */

	    /*
	     * If disk device and random I/O selected, attempt to get
	     * device / partition capacity to limit random I/O seeks.
	     */
	    if ( user_capacity ||
		( ((io_dir == REVERSE) || (io_type == RANDOM_IO)) &&
		  (dip->di_random_access && !num_slices) ) ) {
		if ((status = FindCapacity (dip)) == FAILURE) {
		    exit (exit_status);
		}
	    }
	    if (!record_limit) {
		record_limit = INFINITY;	/* Read until EOF on reads. */
	    }
	}

	/*
	 * Process the output device/file.
	 */
	if (output_file) {
	    int open_mode;
	    dip = active_dinfo = output_dinfo;
	    if (unique_file && (dip->di_dtype->dt_dtype == DT_REGULAR)) {
		/* Note: Must be dynamic for multiple dirs/files. */
		dip->di_dname = make_unique_file(output_file);
		dip->di_bname = strdup(dip->di_dname);
		output_file = strdup(dip->di_dname); /* Cleanup usage! */
		/* NOTE: Some flags may be wrong if file existed! */
	    }
	    if ( dirpath &&
		 ((dip->di_dtype->dt_dtype == DT_MMAP) ||
		  (dip->di_dtype->dt_dtype == DT_REGULAR)) ) {
		int status = setup_directory_info(dip, dirpath);
		if (status == FAILURE) terminate(status); /* For now! */
	    }
	    /*
	     * If a skip count was specified, open output file for R/W,
	     * since skips are accomplished via read()'s. (for pelle)
	     */
	    if (skip_count || raw_flag) {
#if defined(WIN32)
		open_mode = (GENERIC_READ | GENERIC_WRITE);
#else /* !defined(WIN32) */
		open_mode = (O_RDWR  | wopen_flags | open_flags);
#endif /* defined(WIN32) */
	    } else {
		open_mode = (wopen_mode | wopen_flags | open_flags);
	    }
	    /*
	     * Don't create files in the /dev directory (presume the file
	     * should exist instead of creating file & misleading user).
	     */
	    if ( (NEL (output_file, DEV_PREFIX, DEV_LEN)) &&
		 (NEL (output_file, ADEV_PREFIX, ADEV_LEN)) ) {
		open_mode |= O_CREAT;	/* For creating test files.	*/
	    }
	    /*
	     * If verify mode, the output device is open for reads.
	     */
	    if (io_mode == VERIFY_MODE) {
		open_mode = (ropen_mode | open_flags);
		dip->di_mode = READ_MODE;
	    }
	    SetupBufferingMode(dip, &open_mode);
	    if ((*dip->di_funcs->tf_open)(dip, open_mode) == FAILURE) {
		exit (exit_status);
	    }

#if defined(MUNSA)
	    if (munsa_flag) {
		/* first we need to join a namespace */
		l_stat = dlm_nsjoin( getuid(), &nsp, DLM_USER);
		if ((l_stat != DLM_SUCCESS) && (l_stat != DLM_ATTACHED)) {
		    Fprintf ("Can't join namespace\n");
		    dlm_error(0, l_stat);
		}

		resnam = output_file;
		/* now let DLM know what signal to use for blocking routines */
		dlm_set_signal(SIGIO, &i);  /* do we need this ????? */

		resnlen = strlen(resnam);
		if (debug_flag) {
		    Printf ("dlm_set_signal: i %d\n", i);
		    Printf ("resnam %s\n", resnam);
		    Printf ("grab a NL mode lock\n");
		}
		l_stat = dlm_lock(nsp, 
				  (uchar_t *)resnam, 
				  resnlen,
				  0,
				  &lkid,
				  DLM_NLMODE,
				  NULL,
				  DLM_SYNCSTS,
				  0, 0, 0, 0);
		/* NL mode better be granted SYNC status */
		if (l_stat !=  DLM_SYNCH) {
		    Fprintf ("dlm_lock failed\n");
		    dlm_error(&lkid, l_stat);
		}
	    }  /* end if (munsa_flag)... */
#endif /* defined(MUNSA) */
	    system_device_info (dip);
	    output_dtype = dip->di_dtype;
	    open_flags &= ~O_CREAT;	/* Only create on first open. */
	    status = (*dip->di_funcs->tf_validate_opts)(dip);
	    if (status == FAILURE) exit (FATAL_ERROR);

	    /*
	     * If disk device and random I/O selected, attempt to get
	     * device / partition capacity to limit random I/O seeks.
	     */
	    if ( user_capacity ||
		( ((io_dir == REVERSE) || (io_type == RANDOM_IO)) &&
		  (dip->di_random_access && !num_slices) ) ) {
		if ((status = FindCapacity (dip)) == FAILURE) {
		    exit (exit_status);
		}
	    }
	    /*
	     * For disks and tapes, default writing until EOF is reached.
	     */
	    if ( !record_limit &&
		 ((dip->di_dtype->dt_dtype == DT_DISK)  ||
		  (dip->di_dtype->dt_dtype == DT_BLOCK) ||
		  (dip->di_dtype->dt_dtype == DT_TAPE)) ) {
		record_limit = INFINITY;
	    }
	}

	/*
	 * Set the default lbdata size, if not setup by the system
	 * dependent functions above.  Delaying this check to this
	 * point allows the device sector size to be setup, instead
	 * of forcing it to 512 byte blocks.  At least this is true
	 * on Tru64 Unix, where this disk information is available.
	 */
	if (!lbdata_size) lbdata_size = BLOCK_SIZE;

	/*
	 * This is to catch me (dah!) as much as anyone else :-)
	 */
	if ( (rdata_limit || random_align) &&
	     ((io_dir != REVERSE) && (io_type != RANDOM_IO)) ) {
	    Printf ("Warning, random options have no effect without iotype=random!\n");
	}

	/*
	 * If random I/O was selected, and a data limit isn't available,
	 * inform the user we need one, and don't allow testing.
	 */
	if (rdata_limit == 0UL) rdata_limit = data_limit;
	if ( (rdata_limit == 0) && (io_type == RANDOM_IO) ) {
	    LogMsg (efp, logLevelCrit, 0,
		    "Please specify a record or data limit for random I/O.\n");
	    exit (FATAL_ERROR);
	}

	/*
	 * Sanity check the random I/O data limits.
	 */
	if ( (io_type == RANDOM_IO) &&
	     ((large_t)(file_position + block_size + random_align) > rdata_limit)) {
	    LogMsg (efp, logLevelCrit, 0,
		    "The max block size is too large for random data limits!\n");
	    if (Debug_flag) {
	        Printf ("file position " FUF ", bs=%ld, ralign=%ld, rlimit=" LUF "\n",
			file_position, block_size, random_align, rdata_limit);
	    }
	    exit (FATAL_ERROR);
	}

	/*
	 * Special handling for step option:
	 *
	 * For regular files, we must setup an end position (offset), since
	 * doing I/O to a file system will NOT encounter an EOF like raw disks.
	 * Slices also requires an ending position when step option is used.
	 */
	if (step_offset &&
	    ((dip->di_dtype->dt_dtype == DT_REGULAR) || num_slices) ) {
	    if (data_limit && (data_limit != INFINITY)) {
		end_position = (file_position + data_limit);
	    } else {
		end_position = (file_position + (record_limit * block_size));
	    }
	}
	/*
	 * Ensure either a data limit and/or a record count was specified.
	 */
	if (!record_limit) {
	    LogMsg (efp, logLevelCrit, 0,
		    "You must specify a data limit, a record count, or both.\n");
	    exit (FATAL_ERROR);
	}

	/*
	 * If both an input and an output files were specified, then
	 * fork and make child process the reader, parent the writer.
	 */
#if !defined(WIN32)
	if ( (io_mode == TEST_MODE) && input_file && output_file) {
	    if ( (child_pid = fork()) == (pid_t) -1) {
		report_error ("fork", TRUE);
		exit (exit_status);
	    }
	    forked_flag = TRUE;
	    if (child_pid) {			/* Parent = writer. */
		struct dinfo *dip = input_dinfo;
		(void) close_file (dip);
		input_file = NULL;
		if (debug_flag) {
		    Printf ("Parent PID (Writer) = %d, Child PID (Reader) = %d\n",
						process_id, child_pid);
		}
#if !defined(__MSDOS__) || defined(__NUTC__)
		signal (SIGCHLD, SignalHandler);
#endif /* !defined(__MSDOS__) || defined(__NUTC__) */
	    } else {				/* Child = reader. */
		struct dinfo *dip = output_dinfo;
		(void) close_file (dip);
		output_file = NULL;
	    }
	}
#endif /* !defined(WIN32) */
	/*
	 * Some drivers require the input device to open before we start
	 * writing.  For example, terminal devices must have speed, parity,
	 * and flow control setup before we start writing.  The parallel
	 * input device must open before we send the "readya" interrupt.
	 */
	if (output_file && (dip->di_start_delay != 0)) {
	    mySleep (dip->di_start_delay);		/* Allow reader to start. */
	}

	/*
	 * Calculate size necessary for the data buffer & the pad bytes.
	 */
	data_size = (block_size + PADBUFR_SIZE);
	if (rotate_flag) data_size += ROTATE_SIZE;

	/*
	 * For read-after-write (raw) option, we need a verification buffer.
	 * Note:  Other tests can benefit from this too, switch to later.
	 */
	if (raw_flag) {
	    verify_buffer = myalloc(data_size, align_offset);
	}

	/*
	 * Do the device / test specific initialization.
	 *
	 * This function is responsible for allocating the necessary
	 * data buffers and performing special device setup/checking.
	 */
	if (input_file) {
	    dip = input_dinfo;
	    status = (*dip->di_funcs->tf_initialize)(dip);
	    if (status == FAILURE) exit (FATAL_ERROR);
	}
	if (output_file) {
	    dip = output_dinfo;
	    status = (*dip->di_funcs->tf_initialize)(dip);
	    if (status == FAILURE) exit (FATAL_ERROR);
	}

	/*
	 * Re-adjust size of data buffer to avoid init'ing too much.
	 */
	if (rotate_flag) data_size -= ROTATE_SIZE;

	/*
	 * Finally format the prefix string (if any), after the device is
	 * setup and processes are forked so we can setup a unique prefix.
	 */
	if (prefix_string) {
	    status = FmtPrefix(dip, prefix_string, prefix_size);
	    if (status == FAILURE) exit (FATAL_ERROR);
	    fprefix_string = dip->di_prefix_string;
	    fprefix_size = dip->di_prefix_size;
            if ((size_t)fprefix_size > lbdata_size) {
                LogMsg (efp, logLevelCrit, 0,
                        "The prefix size (%d) is larger than lbdata size (%d)!\n",
                        fprefix_size, lbdata_size);
                exit (FATAL_ERROR);
            }
	}

	/*
	 * If an IO tune file was specified, but no alarm, set a default time.* 
	 */
	if (iotune_file && !alarmtime) {
	    alarmtime = iotune_frequency;
	    iotune_alarm = TRUE;
	}

	/*
	 * Start an alarm timer if the run time was specified.
	 */
	if (alarmtime || runtime) {
            if (runtime) {
                runtime_end = time((time_t *)NULL) + runtime;
            }
	    /*
	     * Adjust the alarm time, if it's greater than other times.
	     */
	    if (alarmtime) {
		if (noprogtime) alarmtime = MIN(alarmtime,noprogtime);
		if (keepalive_time) alarmtime = MIN(alarmtime,keepalive_time);
		if (runtime) alarmtime = MIN(alarmtime,runtime);
	    }

	    /*
	     * With monitor no I/O progress, user must specify a keepalive
	     * string, otherwise we won't log anything (just monitor).
	     */
	    if (noprog_flag && !noprogtime) { noprogtime = alarmtime; }
	    if (noprog_flag && !user_keepalive && !keepalive_time) {
		keepalive = "";
		user_keepalive = TRUE;
	    }
	    /*
	     * Only enable alarm for child processes.
	     */
            if (alarmtime && (child_pid == (pid_t)0) ) {
		if ( !iotune_alarm && !keepalive ) {
		    if ( !keepalive || (!user_keepalive && !strlen(keepalive)) ) {
			if (pstats_flag) {
			    keepalive = keepalive0;
			} else {
			    keepalive = keepalive1;
			}
		    }
                }
#if !defined(WIN32)
# if !defined(THREADS)
                (void) signal (SIGALRM, alarmtime);
# endif /* !defined(THREADS) */
#endif /* !defined(WIN32) */
#if defined(THREADS)
                (void) StartMonitorThread(alarmtime);
#else /* !defined(THREADS) */
                (void) alarm (alarmtime);
#endif /* defined(THTREADS) */
		TimerActive = TRUE;
            } else {
#if !defined(WIN32)
# if !defined(THREADS)
	        (void) signal (SIGALRM, SignalHandler);
# endif /* !defined(THREADS) */
#endif /* !defined(WIN32) */
#if defined(THREADS)
		/*
    		 * Even without monitoring, use a thread to avoid those
		 * nasty signals from being blocked, thus hanging dt!
		 */
                (void) StartMonitorThread(runtime);
#else /* !defined(THREADS) */
	        (void) alarm (runtime);
#endif /* defined(THTREADS) */
		TimerActive = TRUE;
            }
	}
        /*
         * When stats=brief, we allow a one line format for pass and totals.
         * If the user has not specified their own format, set our defaults.
         * Note:  They aren't really keepalive, just use the same logic. :-)
         */
        if ( !pkeepalive || !strlen(pkeepalive) ) {
            pkeepalive = pass_msg;
        }
        if ( !tkeepalive || !strlen(tkeepalive) ) {
            tkeepalive = totals_msg;
        }

	/*
	 * Start of main test loop.
	 */
	start_time = times (&stimes);
#if defined(LOG_DIAG_INFO)
	if (logdiag_flag) {
	    sprintf(msg_buffer, "Starting: %s", cmd_line);
	    LogDiagMsg(msg_buffer);
	}
#endif /* defined(LOG_DIAG_INFO) */
#if defined(SYSLOG)
	if (syslog_flag) {
	    syslog(LOG_INFO, "(%d) Starting: %s", process_id, cmd_line);
	}
#endif /* defined(SYSLOG) */
#if defined(DEC)
	if (table_flag) {
	    status = table(TBL_SYSINFO,0,(char *)&s_table,1,sizeof(struct tbl_sysinfo));
	    if (status == FAILURE) report_error ("table", FALSE);
	}
#endif /* defined(DEC) */

    /*
     * We support multiple modes of operation: copy, test, & verify
     */
    if ( (io_mode == COPY_MODE) || (io_mode == VERIFY_MODE) ) {
	/*
	 * This is ugly, but I'm using pattern buffer as verify buffer.
	 */
	if ( (io_mode == VERIFY_MODE) || verify_flag) {
	    pattern_buffer = myalloc (data_size, align_offset);
	    setup_pattern (pattern_buffer, data_size);
	}
	pass_time = times (&ptimes);	/* Start the pass timer	*/
#if defined(DEC)
	if (table_flag) {
	    status = table(TBL_SYSINFO,0,(char *)&p_table,1,sizeof(struct tbl_sysinfo));
	    if (status == FAILURE) report_error ("table", FALSE);
	}
#endif /* defined(DEC) */
	dip = active_dinfo = input_dinfo;
	dtf = dip->di_funcs;
	dip->di_mode = READ_MODE;
	(void) (*dtf->tf_start_test)(dip);
	(void) (*dtf->tf_read_file)(dip);
	(void) (*dtf->tf_end_test)(dip);
	gather_stats(output_dinfo);

	/*
	 * Now verify the data copied (if requested).
	 */
	if ( (io_mode == COPY_MODE) && verify_flag &&
	     !stdin_flag && (error_count < error_limit) ) {
	    struct dinfo *odip = output_dinfo;
	    int open_mode = (ropen_mode | open_flags);
	    report_pass (dip, COPY_STATS);	/* Report copy statistics. */
	    /*
	     * Verify Pass.
	     */
	    status = (*dtf->tf_reopen_file)(dip, open_mode);
	    if (status != SUCCESS) terminate(status);
	    odip->di_mode = READ_MODE;	/* Switch to read mode. */
	    status = (*odip->di_funcs->tf_reopen_file)(odip, open_mode);
	    if (status != SUCCESS) terminate(status);
	    pass_time = times (&ptimes); /* Time the verify. */
#if defined(DEC)
	    if (table_flag) {
		status = table(TBL_SYSINFO,0,(char *)&p_table,1,sizeof(struct tbl_sysinfo));
		if (status == FAILURE) report_error ("table", FALSE);
	    }
#endif /* defined(DEC) */
	    io_mode = VERIFY_MODE;
	    (void) (*dtf->tf_start_test)(dip);
	    (void) (*dtf->tf_read_file)(dip);
	    (void) (*dtf->tf_end_test)(dip);
	    pass_count++;			/* End of copy/verify pass. */
	    gather_stats(output_dinfo);		/* Gather device statistics. */
	    report_pass (dip, VERIFY_STATS);	/* Report the verify stats. */
	} else {
	    pass_count++;			/* End of copy pass. */
	}
    } else { /* not Copy or Verify modes, Test Mode! */
	while ( (total_errors < error_limit) &&
		((pass_count < pass_limit) || runtime) ) {
	    if (pattern_buffer) {
		pattern_bufptr = pattern_buffer;
	    }
	    /*
	     * Use a different data pattern for each pass.
	     */
	    if ( !user_pattern && !ttyport_flag &&
		 (loopback || output_file || stdin_flag ||
		  (input_file && num_slices)) ) {
		/*
		 * Logic:
		 * - If multiple slices, choose a different pattern
		 *   for each pass, factoring in the process number.
		 * - Else, each pass gets a different data pattern
		 *   unless multiple processes were selected.
		 */
		if ( dip->di_unique_pattern &&
		     ( (!num_procs && num_slices) ||
		       (num_procs && (dip->di_dtype->dt_dtype == DT_REGULAR)) ) ) {
		    dip->di_pattern_index = ((cur_proc - 1) + pass_count);
		} else if (!num_procs) {
		    dip->di_pattern_index = pass_count;
		}
		pattern = data_patterns[dip->di_pattern_index % npatterns];
		if (pattern_buffer) copy_pattern (pattern, pattern_buffer);
		if (debug_flag) {
		    Printf ("Using data pattern 0x%08x for pass %u\n",
						pattern, (pass_count + 1));
		}
	    } else if (dip->di_iot_pattern) {
		dip->di_iot_seed_per_pass = dip->di_iot_seed;
		if (dip->di_unique_pattern) {
		    /*
		     * For uniqueness, factor the pass count into the seed.
		     */
		    dip->di_iot_seed_per_pass *= (pass_count + 1);
		}
	    }
	    /*
	     * Use time for random # generator seed so different areas
	     * of disk get affected during multiple passes.  Seed set
	     * here, since it must be the same seed during read pass.
	     * NOTE: Pid is added so seed different for each process.
	     */
	    if ( (io_type == RANDOM_IO) || variable_flag) {
		if ( !user_rseed ) {
		    random_seed = (u_int) times(&ptimes) + process_id;
		}
		set_rseed (random_seed);
	    }
	    pass_time = times (&ptimes);	/* Start the pass timer	*/
#if defined(DEC)
	    if (table_flag) {
		status = table(TBL_SYSINFO,0,(char *)&p_table,1,sizeof(struct tbl_sysinfo));
		if (status == FAILURE) report_error ("table", FALSE);
	    }
#endif /* defined(DEC) */
	    if (output_file) {			/* Write/read the file.	*/
		bool do_read_pass;
		dip = active_dinfo = output_dinfo;
		dtf = dip->di_funcs;
		dip->di_mode = WRITE_MODE;
		(void) (*dtf->tf_start_test)(dip);
		(void) (*dtf->tf_write_file)(dip);
		(void) (*dtf->tf_flush_data)(dip);
		(void) (*dtf->tf_end_test)(dip);
		if (error_count >= error_limit) break;
		do_read_pass = (dip->di_dbytes_written != (large_t) 0);
		/*
		 * Now verify (read and data compare) the data just written.
		 */
		if ( verify_flag && do_read_pass && 
		     (!raw_flag || (raw_flag && reread_flag)) ) { /* Verify data written. */
		    int open_mode = (ropen_mode | open_flags);
		    
		    if (raw_flag) {
			report_pass (dip, RAW_STATS);	/* Report read-after-write. */
		    } else {
			report_pass (dip, WRITE_STATS);	/* Report write stats.	*/
		    }
		    /*
		     * For  multiple files, reset the pattern/IOT seed for read pass!
		     */
		    if (dip->di_file_limit) {
			if (!user_pattern) {
			    pattern = data_patterns[dip->di_pattern_index % npatterns];
			    if (pattern_buffer) copy_pattern (pattern, pattern_buffer);
			} else if (dip->di_iot_pattern) {
			    dip->di_iot_seed_per_pass = dip->di_iot_seed;
			    if (dip->di_unique_pattern) {
				dip->di_iot_seed_per_pass *= (pass_count + 1);
			    }
			}
		    }
		    dip->di_mode = READ_MODE;
		    if (multi_flag && media_changed) {
			status = RequestFirstVolume(dip, open_flags);
		    } else {
			status = (*dtf->tf_reopen_file)(dip, open_mode);
		    }
		    if (status != SUCCESS) break;
		    /*
		     * Reset the random seed, so reads mimic what we wrote!
		     */
		    if ( (io_type == RANDOM_IO) || variable_flag) {
			set_rseed (random_seed);
		    }
		    pass_time = times (&ptimes);	/* Time just the read.	*/
#if defined(DEC)
		    if (table_flag) {
			status = table(TBL_SYSINFO,0,(char *)&p_table,1,sizeof(struct tbl_sysinfo));
			if (status == FAILURE) report_error ("table", FALSE);
		    }
#endif /* defined(DEC) */
		    /*rotate_offset = 0;*/
		    if (pattern_buffer) {
			pattern_bufptr = pattern_buffer;
		    }
		    (void) (*dtf->tf_start_test)(dip);
		    (void) (*dtf->tf_read_file)(dip);
		    (void) (*dtf->tf_end_test)(dip);
		    pass_count++;			/* End read/write pass. */
		    report_pass (dip, READ_STATS);	/* Report read stats.	*/
		} else {
		    pass_count++;		        /* End of write pass.	*/
		    if ( (pass_limit > 1) || runtime) {
			/* Report write stats. */
			if (raw_flag) {
			    report_pass (dip, RAW_STATS);
			} else {
			    report_pass (dip, WRITE_STATS);
			}
		    }
		}
		if ( delete_per_pass			    &&
		     (total_errors < error_limit)           &&
		     ((pass_count < pass_limit) || runtime) &&
		     ((dip->di_dtype->dt_dtype == DT_MMAP) ||
		      (dip->di_dtype->dt_dtype == DT_REGULAR)) ) {
		    if (dip->di_fd != NoFd) {
			status = (*dip->di_funcs->tf_close)(dip);
		    }
		    status = delete_files(dip);
		    if (status == FAILURE) break;
		    if (dirpath) {
			dip->di_mode = WRITE_MODE;
			status = setup_directory_info(dip, dirpath);
			if (status == FAILURE) break;
		    }
		}
                /*
                 * Don't reopen if we've reached the error limit or the
                 * pass count, since we'll be terminating shortly.
                 */
		if ( (total_errors < error_limit) &&
                     ((pass_count < pass_limit) || runtime) ) {
		    int open_mode;
		    SetupBufferingMode(dip, &open_flags);
		    if (skip_count || raw_flag) {
#if defined(WIN32)
			open_mode = (GENERIC_READ | GENERIC_WRITE);
#else /* !defined(WIN32) */
			open_mode = (O_RDWR | wopen_flags | open_flags);
#endif /* defined(WIN32) */
		    } else {
			open_mode = (wopen_mode | wopen_flags | open_flags);
		    }
		    dip->di_mode = WRITE_MODE;
		    if (delete_per_pass) {
#if !defined(WIN32)
			open_mode |= O_CREAT;
#endif
			dip->di_file_number = 0;
			dip->di_subdir_number = 0;
			dip->di_subdir_depth = 0;
			status = (*dtf->tf_open)(dip, open_mode);
		    } else {
			status = (*dtf->tf_reopen_file)(dip, open_mode);
		    }
		    if (status != SUCCESS) break;
		}
	    } else { /* Reading only. */
		dip = active_dinfo = input_dinfo;
		dtf = dip->di_funcs;
		dip->di_mode = READ_MODE;
		/*
		 * Note: User must set random seed to repeat previous write sequence!
		 */
		if (user_rseed && ((io_type == RANDOM_IO) || variable_flag)) {
		    set_rseed (random_seed);
		}
		(void) (*dtf->tf_start_test)(dip);
		(void) (*dtf->tf_read_file)(dip);
		(void) (*dtf->tf_end_test)(dip);
		pass_count++;			/* End of read pass.	*/
		/*
		 * Prevent pass unless looping, since terminate reports
		 * the total statistics when called (prevents dup stats).
		 */
		if ( (pass_limit > 1) || runtime) {
		    report_pass (dip, READ_STATS);	/* Report read stats.	*/
		}
		if ( (total_errors < error_limit) &&
		     ((pass_count < pass_limit) || runtime) ) {
		    int open_mode;
		    SetupBufferingMode(dip, &open_flags);
		    open_mode = (ropen_mode | open_flags);
		    status = (*dtf->tf_reopen_file)(dip, open_mode);
		    if (status != SUCCESS) break;
		}
	    } /* End of a pass! */
#if defined(MUNSA)
	    if (munsa_flag) {
		if (debug_flag) {
		    Printf ("converting to dlm NL-> %d \n", DLM_NLMODE);
		}
		l_stat = dlm_cvt(&lkid, DLM_NLMODE, NULL, 0, 0, 0, NULL, 0);
		if (l_stat !=  DLM_SUCCESS) {
		    Fprintf ("dlm_cvt failed\n");
		    dlm_error(&lkid, l_stat);	/* exit with FATAL ERROR */ 
		}
	    } /* end if(munsa_flag)... */
#endif /* defined(MUNSA) */
	    if (stop_on_file) {
		if (stat (stop_on_file, &sb) == SUCCESS) {
		    if (verbose_flag) {
			Printf("Detected stop on file '%s', so stopping test...\n",  stop_on_file);
		    }
		    break;
		}
	    }
	}
    } /* end 'if ( (io_mode == COPY_MODE) || (io_mode == VERIFY_MODE) ) ' */

#if defined(MUNSA)
    if (munsa_flag) {
	if ((l_stat = dlm_unlock(&lkid, NULL, 0)) != DLM_SUCCESS)
	    dlm_error(&lkid, l_stat);  /* exit with FATAL ERROR  */

	if (debug_flag) {
	    Printf("\n   %s:  unlocked...\n", resnam);
	}
    }  /*  end if(munsa_flag)...  */
#endif /* defined(MUNSA) */

    terminate (exit_status);		/* Terminate with exit status.	*/
    /*NOTREACHED*/
    return (exit_status);		/* Quiet certain compilers!	*/
}

#if defined(MUNSA)
/************************************************************************
 *									*
 * dlm_error() - Print all MUNSA dlm related error msg.			*
 *									*
 * Inputs:	dlm_lkid_t       *lk                            	*
 *		dlm_status_t      stat			                *
 *									*
 * 	If errors are detected, we simply exit with a fatal error.	*
 *									*
 ************************************************************************/
void
dlm_error(dlm_lkid_t *lk, dlm_status_t err_stat)
{
	dlm_rsbinfo_t rsb;		/* used by dlm_sperrno() */

	LogMsg (efp, logLevelCrit, 0,
		"lock error %s on lkid 0x%lx\n",
			dlm_sperrno(err_stat), *lk);
	exit (FATAL_ERROR);
}
#endif /* defined(MUNSA) */

/************************************************************************
 *									*
 * parse_args() - Parse 'dt' Program Arguments.				*
 *									*
 * Inputs:	argc = The number of arguments.				*
 *		argv = Array pointer to arguments.			*
 *									*
 * 	If errors are detected, w
 *									*
 ************************************************************************/
void
parse_args (int argc, char **argv)
{
	int i;

	if (argc == 1) dtusage();
	for (i = 1; i < argc; i++) {
	    string = argv[i];
	    if (match ("aios=")) {
		aio_bufs = (int)number(ANY_RADIX);
		if (aio_bufs) aio_flag = TRUE;
		continue;
	    }
            if (match ("alarm=")) {
                alarmtime = time_value();
                continue;
            }
	    if (match ("bufmodes=")) {
            bufmode_loop:
                if (match(","))
                    goto bufmode_loop;
                if (*string == '\0')
                    continue;
		if (bufmode_count == NUM_BUFMODES) {
		    LogMsg (efp, logLevelError, 0,
			    "Too many buffering modes specified, max is %d\n", NUM_BUFMODES);
		    exit (FATAL_ERROR);
		}
		if (match("buffered")) {
		    buffer_modes[bufmode_count++] = BUFFERED_IO;
		} else if (match("unbuffered")) {
		    buffer_modes[bufmode_count++] = UNBUFFERED_IO;
		} else if (match("cachereads")) {
		    buffer_modes[bufmode_count++] = CACHE_READS;
		} else if (match("cachewrites")) {
		    buffer_modes[bufmode_count++] = CACHE_WRITES;
		} else {
		    LogMsg (efp, logLevelError, 0,
			    "Invalid bufmode keyword: %s\n", string);
		    exit (FATAL_ERROR);
		}
		goto bufmode_loop;
	    }
	    if (match ("boff=")) {
		if (match("dec")) {
		    boff_format = DEC_FMT;
		} else if (match ("hex")) {
		    boff_format = HEX_FMT;
		} else {
		    LogMsg (efp, logLevelError, 0,
			    "Valid buffer offset formats are: dec or hex\n");
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    if (match ("dfmt=")) {
		if (match("byte")) {
		    data_format = BYTE_FMT;
		} else if (match ("word")) {
		    data_format = WORD_FMT;
		} else {
		    LogMsg (efp, logLevelError, 0,
			    "Valid data formats are: byte or word\n");
		    exit (FATAL_ERROR);
		}
		continue;
	    }
            if (match ("keepalive=")) {
                keepalive = string;
                user_keepalive = TRUE;
                continue;
            }
            if (match ("keepalivet=")) {
                keepalive_time = time_value();
		if (keepalive_time) {
		    if (!alarmtime) alarmtime = keepalive_time;
		}
		continue;
	    }
            if (match ("pkeepalive=")) {
                pkeepalive = string;
                user_pkeepalive = TRUE;
                continue;
            }
            if (match ("tkeepalive=")) {
                tkeepalive = string;
                user_tkeepalive = TRUE;
                continue;
            }
            if (match ("noprogt=")) {
                noprogtime = time_value();
                if (noprogtime) {
		    noprog_flag = TRUE;
		    if (!alarmtime) alarmtime = noprogtime;
                }
                continue;
            }
            if (match ("noprogtt=")) {
                noprogttime = time_value();
		if (noprogttime && !noprogtime) {
		    noprog_flag = TRUE;
		    noprogtime = noprogttime;
		    if (!alarmtime) alarmtime = noprogttime;
		}
                continue;
            }
	    if (match ("notime=")) {
		int i;
		optiming_t *optp;
		/*
		 * May loop through table more than once, as we parse
		 * multiple options: notime=fsync,open,close...
		 */
		do {
		    optp = &optiming_table[OPEN_OP];
		    for ( i = OPEN_OP; (i < NUM_OPS); optp++, i++ ) {
			if ( match(",") ) break;
			if ( match(optp->opt_name) ) {
			    optp->opt_timing_flag = FALSE;
			    break;
			}
		    }
		    if (i == NUM_OPS) break;	/* optype not found! */
		    if (*string == '\0') break; /* done parsing */
		} while (TRUE);
		if ( i == NUM_OPS ) {
		    Fprintf("%s is not a valid operation type to disable!\n", string);
		    optp = &optiming_table[OPEN_OP];
		    Fprintf("Valid operation types are: ");
		    for ( i = OPEN_OP; (i < NUM_OPS); optp++, i++ ) {
			Fprint("%s ", optp->opt_name);
		    }
		    Fprint("\n");
		    exit (FATAL_ERROR);
		}
                continue;
            }
            if (match ("totals=")) {
                totals_msg = string;
                continue;
            }
	    if (match ("align=")) {
		if (match ("rotate")) {
		    rotate_flag = TRUE;
		    continue;
		}
		align_offset = (u_int) number(ANY_RADIX);
		continue;
	    }
	    if (match ("bs=")) {
		if (match("random")) {
		    if (user_min == FALSE) {
			min_size = MIN_RANDOM_SIZE;
		    }
		    if (user_max == FALSE) {
			max_size = MAX_RANDOM_SIZE;
		    }
		    variable_flag = TRUE;
		    continue;
		}
		block_size = number(ANY_RADIX);
		if ((ssize_t)block_size <= (ssize_t) 0) {
		    LogMsg (efp, logLevelError, 0,
			    "block size must be positive and non-zero.\n");
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    if (match ("capacity=")) {
		if (match ("max")) {
		    max_capacity = TRUE;
		} else {
		    user_capacity = large_number(ANY_RADIX);
		}
		continue;
	    }
	    if (match ("dsize=")) {
		device_size = number(ANY_RADIX);
		/*
		 * Adjust the dump limit to the device size (as appropriate).
		 */
		if ( (dump_limit == DEFAULT_DUMP_LIMIT) &&
		     (device_size > DEFAULT_DUMP_LIMIT) ) {
		    dump_limit = device_size;
		}
		continue;
	    }
	    if (match ("ffreq=")) {
		fsync_frequency = number(ANY_RADIX);
		continue;
	    }
	    if (match ("hdsize=")) {
		history_data_size = (int)number(ANY_RADIX);
		continue;
	    }
	    if (match ("history=")) {
		history_size = (int)number(ANY_RADIX);
		continue;
	    }
	    if (match ("lba=")) {
		lbdata_flag = TRUE;
		lbdata_addr = number(ANY_RADIX);
		user_lbdata = TRUE;
		continue;
	    }
	    if (match ("lbs=")) {
		lbdata_flag = TRUE;
		lbdata_size = number(ANY_RADIX);
		user_lbsize = TRUE;
		if ((ssize_t)lbdata_size <= (ssize_t) 0) {
		    LogMsg (efp, logLevelError, 0,
			    "lbdata size must be positive and non-zero.\n");
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    if (match ("maxbad=")) {
		max_bad_blocks = (u_int)number(ANY_RADIX);
		continue;
	    }
	    if ( (match ("count=")) || (match ("records=")) ) {
		if (fsincr_flag) {
		    LogMsg (efp, logLevelError, 0,
			    "Cannot use record count with file size increment option!\n");
		    exit (FATAL_ERROR);
		}
		record_limit = large_number(ANY_RADIX);
		continue;
	    }
	    if (match ("cancel_delay=")) {
		cancel_delay = (u_int) number(ANY_RADIX);
		continue;
	    }
	    if (match ("kill_delay=")) {
		kill_delay = (u_int) number(ANY_RADIX);
		continue;
	    }
	    if (match ("cdelay=")) {
		if (match ("random")) {
		    cdelay_count = RANDOM_DELAY_VALUE;
		} else {
		    cdelay_count = (u_int) number(ANY_RADIX);
		}
		continue;
	    }
	    if (match ("edelay=")) {
		if (match ("random")) {
		    edelay_count = RANDOM_DELAY_VALUE;
		} else {
		    edelay_count = (u_int) number(ANY_RADIX);
		}
		continue;
	    }
	    if (match ("rdelay=")) {
		if (match ("random")) {
		    rdelay_count = RANDOM_DELAY_VALUE;
		} else {
		    rdelay_count = (u_int) number(ANY_RADIX);
		}
		continue;
	    }
	    if (match ("sdelay=")) {
		if (match ("random")) {
		    sdelay_count = RANDOM_DELAY_VALUE;
		} else {
		    sdelay_count = (u_int) number(ANY_RADIX);
		}
		continue;
	    }
	    if (match ("tdelay=")) {
		tdelay_count = (u_int) number(ANY_RADIX);
		continue;
	    }
	    if (match ("wdelay=")) {
		if (match ("random")) {
		    wdelay_count = RANDOM_DELAY_VALUE;
		} else {
		    wdelay_count = (u_int) number(ANY_RADIX);
		}
		continue;
	    }
	    if (match ("retry_delay=")) {
		retry_delay = (u_int) number(ANY_RADIX);
		continue;
	    }
	    if (match ("term_retries=")) {
		term_wait_retries = (u_int) number(ANY_RADIX);
		continue;
	    }
	    if (match ("term_timeout=")) {
		term_wait_timeout = (u_int) number(ANY_RADIX);
		continue;
	    }
	    if (match ("errors=")) {
		error_limit = number(ANY_RADIX);
		continue;
	    }
	    if (match ("hz=")) {
		hz = (u_int) number(ANY_RADIX);
		continue;
	    }
	    if (match ("incr=")) {
		user_incr = TRUE;
		if (match ("var")) {
		    variable_flag = TRUE;
		} else {
		    incr_count = number(ANY_RADIX);
		}
		continue;
	    }
	    if (match ("dlimit=")) {
		dump_limit = number(ANY_RADIX);
		continue;
	    }
	    if (match ("limit=")) {
		data_limit = large_number(ANY_RADIX);
		if (!record_limit) {
		    record_limit = INFINITY;
		}
		continue;
	    }
	    if (match ("maxdata=")) {
		max_data = large_number(ANY_RADIX);
		continue;
	    }
	    if (match ("ralign=")) {
		io_type = RANDOM_IO;
                user_ralign = TRUE;
		random_align = number(ANY_RADIX);
		continue;
	    }
	    if (match ("rlimit=")) {
		io_type = RANDOM_IO;
		rdata_limit = large_number(ANY_RADIX);
		continue;
	    }
	    if (match ("max=")) {
		user_max = TRUE;
		max_size = (size_t) number(ANY_RADIX);
		continue;
	    }
	    if (match ("min=")) {
		user_min = TRUE;
		min_size = (size_t) number(ANY_RADIX);
		continue;
	    }
	    if (match ("enable=")) {
	    eloop:
		if (match(","))
		    goto eloop;
		if (*string == '\0')
		    continue;
		if (match("aio")) {
		    aio_flag = TRUE;
		    goto eloop;
		}
		if (match("bypass")) {
		    bypass_flag = TRUE;
		    goto eloop;
		}
		if (match("cerrors")) {
		    cerrors_flag = TRUE;
		    goto eloop;
		}
		if (match("child")) {
		    child_flag = TRUE;
		    logpid_flag = TRUE;
		    unique_file = TRUE;
		    goto eloop;
		}
		if (match("compare")) {
		    compare_flag = TRUE;
		    goto eloop;
		}
		if (match("coredump")) {
		    core_dump = TRUE;
		    goto eloop;
		}
		if (match("deleteperpass")) {
		    delete_per_pass = TRUE;
		    goto eloop;
		}
		if (match("debug")) {
		    debug_flag = TRUE;
		    goto eloop;
		}
		if (match("Debug")) {
		    Debug_flag = debug_flag = TRUE;
		    goto eloop;
		}
		if (match("edebug")) {
		    eDebugFlag = TRUE;
		    goto eloop;
		}
		if (match("fdebug")) {
		    fDebugFlag = TRUE;
		    goto eloop;
		}
		if (match("pdebug")) {
		    pDebugFlag = TRUE;
		    goto eloop;
		}
		if (match("rdebug")) {
		    rDebugFlag = TRUE;
		    goto eloop;
		}
		if (match("tdebug")) {
		    tDebugFlag = TRUE;
		    goto eloop;
		}
		if (match("diag")) {
		    logdiag_flag = TRUE;
		    goto eloop;
		}
		if (match("dumpall")) {
		    dumpall_flag = TRUE;
		    goto eloop;
		}
		if (match("dump")) {
		    dump_flag = TRUE;
		    goto eloop;
		}
		if (match("eof")) {
		    eof_status = TRUE;
		    goto eloop;
		}
#if defined(EEI)
		if (match("eei")) {
		    eei_flag = TRUE;
		    goto eloop;
		}
		if (match("resets")) {
		    eei_resets = TRUE;
		    goto eloop;
		}
#endif /* defined(EEI) */
		if (match("flush")) {
		    flush_flag = TRUE;
		    goto eloop;
		}
		if (match("fsincr")) {
		    /* Note: limit= sets the record limit too! */
		    if (record_limit && (record_limit != INFINITY)) {
			LogMsg (efp, logLevelError, 0,
				"Cannot use file size increment option with a record limit!\n");
			exit (FATAL_ERROR);
		    }
		    fsincr_flag = TRUE;
		    goto eloop;
		}
		if (match("fsync")) {
		    fsync_flag = TRUE;
		    goto eloop;
		}
		if (match("fsalign")) {
		    fsalign_flag = TRUE;
		    goto eloop;
		}
		if (match("funique")) {
		    unique_file = TRUE;
		    goto eloop;
		}
		if (match("hazard")) {
		    efp = ofp;			/* Write errors to stdout. */
		    hazard_flag = TRUE;
		    goto eloop;
		}
		if (match("header")) {
		    header_flag = TRUE;
		    goto eloop;
		}
		if (match("hdump")) {
		    history_dump = TRUE;
		    goto eloop;
		}
		if (match("htiming")) {
		    history_timing = TRUE;
		    goto eloop;
		}
		if (match("lbdata")) {
		    lbdata_flag = TRUE;
		    goto eloop;
		}
		if (match("logpid")) {
		    logpid_flag = TRUE;
		    goto eloop;
		}
		if (match("loopback")) {
		    loopback = TRUE;
		    verify_flag = FALSE;
		    goto eloop;
		}
		if (match("looponerror")) {
		    loop_on_error = TRUE;
		    goto eloop;
		}
		if (match("microdelay")) {
		    micro_flag = TRUE;
		    goto eloop;
		}
#if defined(MMAP)
		if (match("mmap")) {
		    mmap_flag = TRUE;
		    wopen_mode = O_RDWR;	/* MUST open read/write. */
		    goto eloop;
		}
#endif /* defined(MMAP) */
#if defined(TTY)
		if (match("modem")) {
		    open_flags = O_NONBLOCK;
		    modem_flag = TRUE;
		    goto eloop;
		}
#endif /* defined(TTY) */
		if (match("multi")) {
		    multi_flag = TRUE;
		    goto eloop;
		}
		if (match("noprog")) {
		    noprog_flag = TRUE;
		    goto eloop;
		}
		if (match("prefill")) {
		    prefill_flag = TRUE;
		    goto eloop;
		}
		if (match("pstats")) {
		    pstats_flag = TRUE;
		    goto eloop;
		}
		if (match("raw")) {
		    raw_flag = TRUE;
		    goto eloop;
		}
		if (match("reread")) {
		    reread_flag = TRUE;
		    goto eloop;
		}
		if (match("readcache")) {
		    read_cache_flag = TRUE;
		    goto eloop;
		}
		if (match("writecache")) {
		    write_cache_flag = TRUE;
		    goto eloop;
		}
		if (match("retryDC")) {
		    retryDC_flag = TRUE;
		    goto eloop;
		}
		if (match("retryIO")) {
		    retryIO_flag = TRUE;
		    goto eloop;
		}
		if (match("spad")) {
		    spad_check = TRUE;
		    goto eloop;
		}
		if (match("sighup")) {
		    sighup_flag = TRUE;
		    goto eloop;
		}
#if defined(WIN32)
		if (match("prealloc")) {
		    prealloc_flag = TRUE;
		    goto eloop;
		}
		if (match("sparse")) {
		    sparse_flag = TRUE;
		    goto eloop;
		}
#endif /* defined(WIN32) */
		if (match("stats")) {
		    stats_flag = TRUE;
		    goto eloop;
		}
#if defined(sun) && defined(TTY)
		if (match("softcar")) {
		    open_flags = O_NDELAY;
		    softcar_opt = ON;
		    goto eloop;
		}
#endif /* defined(sun) && defined(TTY) */
		if (match("syslog")) {
		    syslog_flag = TRUE;
		    goto eloop;
		}
#if defined(DEC)
		if (match("table")) {
		    table_flag = TRUE;
		    goto eloop;
		}
#endif /* defined(DEC) */
#if defined(TIMESTAMP)
		if (match("timestamp")) {
		    timestamp_flag = TRUE;
		    goto eloop;
		}
#endif /* !defined(TIMESTAMP) */
		if (match("trigargs")) {
		    trigargs_flag = TRUE;
		    goto eloop;
		}
#if defined(TTY)
		if (match("ttyport")) {
		    ttyport_flag = TRUE;
		    goto eloop;
		}
#endif /* defined(TTY) */
		if (match("unique")) {
		    unique_pattern = TRUE;
		    goto eloop;
		}
		if (match("verbose")) {
		    verbose_flag = TRUE;
		    goto eloop;
		}
		if (match("verify")) {
		    verify_flag = TRUE;
		    goto eloop;
		}
		LogMsg (efp, logLevelError, 0,
			"Invalid enable keyword: %s\n", string);
		exit (FATAL_ERROR);
	    }
	    if (match ("disable=")) {
	    dloop:
		if (match(","))
		    goto dloop;
		if (*string == '\0')
		    continue;
		if (match("aio")) {
		    aio_flag = FALSE;
		    goto dloop;
		}
		if (match("bypass")) {
		    bypass_flag = FALSE;
		    goto dloop;
		}
		if (match("cerrors")) {
		    cerrors_flag = FALSE;
		    goto dloop;
		}
		if (match("compare")) {
		    compare_flag = FALSE;
		    goto dloop;
		}
		if (match("deleteperpass")) {
		    delete_per_pass = FALSE;
		    goto dloop;
		}
		if (match("debug")) {
		    debug_flag = FALSE;
		    goto dloop;
		}
		if (match("Debug")) {
		    Debug_flag = debug_flag = FALSE;
		    goto dloop;
		}
		if (match("edebug")) {
		    eDebugFlag = FALSE;
		    goto dloop;
		}
		if (match("fdebug")) {
		    fDebugFlag = FALSE;
		    goto dloop;
		}
		if (match("pdebug")) {
		    pDebugFlag = FALSE;
		    goto dloop;
		}
		if (match("rdebug")) {
		    rDebugFlag = FALSE;
		    goto dloop;
		}
		if (match("tdebug")) {
		    tDebugFlag = FALSE;
		    goto dloop;
		}
		if (match("diag")) {
		    logdiag_flag = FALSE;
		    goto dloop;
		}
		if (match("dumpall")) {
		    dumpall_flag = FALSE;
		    goto dloop;
		}
		if (match("dump")) {
		    dump_flag = TRUE;
		    goto dloop;
		}
		if (match("eof")) {
		    eof_status = FALSE;
		    goto dloop;
		}
#if defined(EEI)
		if (match("eei")) {
		    eei_flag = FALSE;
		    goto dloop;
		}
		if (match("resets")) {
		    eei_resets = FALSE;
		    goto dloop;
		}
#endif /* defined(EEI) */
		if (match("flush")) {
		    flush_flag = FALSE;
		    goto dloop;
		}
		if (match("fsincr")) {
		    fsincr_flag = FALSE;
		    goto dloop;
		}
		if (match("fsync")) {
		    fsync_flag = FALSE;
		    goto dloop;
		}
		if (match("fsalign")) {
		    fsalign_flag = FALSE;
		    goto dloop;
		}
		if (match("funique")) {
		    unique_file = FALSE;
		    goto dloop;
		}
		if (match("hazard")) {
		    hazard_flag = FALSE;
		    goto dloop;
		}
		if (match("header")) {
		    header_flag = FALSE;
		    goto dloop;
		}
		if (match("hdump")) {
		    history_dump = FALSE;
		    goto dloop;
		}
		if (match("htiming")) {
		    history_timing = FALSE;
		    goto dloop;
		}
		if (match("lbdata")) {
		    lbdata_flag = FALSE;
		    user_lbdata = FALSE;
		    goto dloop;
		}
		if (match("logpid")) {
		    logpid_flag = FALSE;
		    goto dloop;
		}
		if (match("loopback")) {
		    loopback = FALSE;
		    goto dloop;
		}
		if (match("looponerror")) {
		    loop_on_error = FALSE;
		    goto dloop;
		}
		if (match("microdelay")) {
		    micro_flag = FALSE;
		    goto dloop;
		}
#if defined(MMAP)
		if (match("mmap")) {
		    mmap_flag = FALSE;
		    goto dloop;
		}
#endif /* defined(MMAP) */
#if defined(TTY)
		if (match("modem")) {
		    open_flags &= ~(O_NONBLOCK);
		    modem_flag = FALSE;
		    goto dloop;
		}
#endif /* defined(TTY) */
		if (match("noprog")) {
		    noprog_flag = FALSE;
		    goto dloop;
		}
		if (match("pad")) {
		    pad_check = FALSE;
		    goto dloop;
		}
		if (match("prefill")) {
		    prefill_flag = FALSE;
		    goto dloop;
		}
		if (match("pstats")) {
		    pstats_flag = FALSE;
		    goto dloop;
		}
		if (match("raw")) {
		    raw_flag = FALSE;
		    goto dloop;
		}
		if (match("reread")) {
		    reread_flag = FALSE;
		    goto dloop;
		}
		if (match("readcache")) {
		    read_cache_flag = FALSE;
		    goto dloop;
		}
		if (match("writecache")) {
		    write_cache_flag = FALSE;
		    goto dloop;
		}
		if (match("retryDC")) {
		    retryDC_flag = FALSE;
		    goto dloop;
		}
		if (match("retryIO")) {
		    retryIO_flag = FALSE;
		    goto dloop;
		}
		if (match("sighup")) {
		    sighup_flag = FALSE;
		    goto dloop;
		}
		if (match("spad")) {
		    spad_check = FALSE;
		    goto dloop;
		}
#if defined(WIN32)
		if (match("prealloc")) {
		    prealloc_flag = FALSE;
		    goto dloop;
		}
		if (match("sparse")) {
		    sparse_flag = FALSE;
		    goto dloop;
		}
#endif /* defined(WIN32) */
#if defined(TTY)
		if (match("softcar")) {
		    softcar_opt = OFF;
		    goto dloop;
		}
#endif /* defined(TTY) */
		if (match("stats")) {
		    stats_flag = FALSE;
                    stats_level = STATS_NONE;
		    goto dloop;
		}
		if (match("syslog")) {
		    syslog_flag = FALSE;
		    goto dloop;
		}
#if defined(DEC)
		if (match("table")) {
		    table_flag = FALSE;
		    goto dloop;
		}
#endif /* defined(DEC) */
		if (match("timestamp")) {
		    timestamp_flag = FALSE;
		    goto dloop;
		}
		if (match("trigargs")) {
		    trigargs_flag = FALSE;
		    goto dloop;
		}
		if (match("unique")) {
		    unique_pattern = FALSE;
		    goto dloop;
		}
		if (match("verbose")) {
		    verbose_flag = FALSE;
		    goto dloop;
		}
		if (match("verify")) {
		    verify_flag = FALSE;
		    goto dloop;
		}
		LogMsg (efp, logLevelError, 0,
			"Invalid disable keyword: %s\n", string);
		exit (FATAL_ERROR);
	    }
	    if (match ("dispose=")) {
		if (match("delete")) {
		    keep_existing = FALSE;
		    dispose_mode = DELETE_FILE;
		} else if (match("keeponerror")) { /* parse first! */
		    keep_existing = TRUE;
		    dispose_mode = KEEP_ON_ERROR;
		} else if (match("keep")) {
		    keep_existing = TRUE;
		    dispose_mode = KEEP_FILE;
		} else {
		    LogMsg (efp, logLevelError, 0,
			    "Dispose modes are 'delete', 'keep', or 'keeponerror'.\n", string);
		    exit (FATAL_ERROR);
		}
		continue;
	    }
#if defined(MUNSA)
	    if (match ("munsa=")) {
		munsa_flag = TRUE;
		flow_str = string;
		if (match("cr")) {
		    munsa_lock_type = DLM_CRMODE;  /* Concurrent Read */
		} else if (match("pr")) {
		    munsa_lock_type = DLM_PRMODE;  /* Protected Read */
		} else if (match("cw")) {
		    munsa_lock_type = DLM_CWMODE;  /* Concurrent Write */
		} else if (match("pw")) {
		    munsa_lock_type = DLM_PWMODE;  /* Protected Write */
		} else if (match("ex")) {
		    munsa_lock_type = DLM_EXMODE;  /* EXclusive mode */
		} else {
		    LogMsg (efp, logLevelError, 0,
		"Munsa lock types are 'cr', 'pr','cw', 'pw', or 'ex'.\n");
		    exit (FATAL_ERROR);
		}
		continue;
	    }
#endif /* defined(MUNSA) */
#if defined(TTY)
	    if (match ("flow=")) {
		flow_str = string;
		if (match("none")) {
		    flow_type = FLOW_NONE;	/* No flow control. */
		} else if (match("cts_rts")) {
		    flow_type = CTS_RTS;	/* CTS/RTS flow control. */
		} else if (match("xon_xoff")) {
		    flow_type = XON_XOFF;	/* XON/XOFF flow control. */
		} else {
		    LogMsg (efp, logLevelError, 0,
			    "Flow types are 'none', 'cts_rts', or 'xon_xoff'.\n");
		    exit (FATAL_ERROR);
		}
		continue;
	    }
#endif /* defined(TTY) */
	    if (match ("dir=")) {
		dirpath = string;
		continue;
	    }
	    if (match ("dirp=")) {
		dirprefix = string;
		continue;
	    }
#if not_implemented_yet
	    if (match ("dirs=")) {
		dir_limit = number(ANY_RADIX);
		continue;
	    }
#endif /* not_implemented_yet */
	    if (match ("files=")) {
		file_limit = number(ANY_RADIX);
		continue;
	    }
	    if (match ("maxfiles=")) {
		max_files = number(ANY_RADIX);
		continue;
	    }
	    if (match ("sdirs=")) {
		subdir_limit = number(ANY_RADIX);
		continue;
	    }
	    if (match ("depth=")) {
		subdir_depth = number(ANY_RADIX);
		continue;
	    }
	    if (match ("if=")) {
		input_file = string;
		continue;
	    }
	    if (match ("of=")) {
		output_file = string;
		continue;
	    }
	    if (match ("pf=")) {
		pattern_file = string;
		user_pattern = TRUE;
		continue;
	    }
	    if (match ("log=")) {
		log_file = string;
		continue;
	    }
	    if (match ("logt=")) {
		log_file = string;
		logappend_flag = FALSE;
		continue;
	    }
	    if (match ("logu=")) {
		log_file = string;
		unique_log = TRUE;
		continue;
	    }
	    if (match ("iodir=")) {
		if (match ("for")) {
		    io_dir = FORWARD;
		} else if (match ("rev")) {
		    io_dir = REVERSE;
		} else {
		    LogMsg (efp, logLevelError, 0,
			    "Valid I/O directions are: 'forward' or 'reverse'.\n");
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    if (match ("iomode=")) {
		if (match ("copy")) {
		    dispose_mode = KEEP_FILE;
		    io_mode = COPY_MODE;
		} else if (match ("test")) {
		    io_mode = TEST_MODE;
		} else if (match ("verify")) {
		    io_mode = VERIFY_MODE;
		    verify_only = TRUE;
		} else {
		    LogMsg (efp, logLevelError, 0,
			    "Valid I/O modes are: 'copy', 'test', or verify.\n");
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    if (match ("iotype=")) {
		if (match ("random")) {
		    io_type = RANDOM_IO;
		} else if (match ("sequential")) {
		    io_type = SEQUENTIAL_IO;
		} else {
		    LogMsg (efp, logLevelError, 0,
			    "Valid I/O types are: 'random' or 'sequential'.\n");
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    /*
	     * Special options to help seed IOT pattern with multiple passes.
	     */
	    if (match ("iotpass=")) {
		int iot_pass = number(ANY_RADIX);
		iot_seed *= iot_pass;
		iot_pattern = TRUE;
		continue;
	    }
	    if (match ("iotseed=")) {
		iot_seed = (u_int32)number(HEX_RADIX);
		iot_pattern = TRUE;
		continue;
	    }
	    if (match ("iotune=")) {
		iotune_file = string;
		continue;
	    }
	    /*
	     * Flags which apply to read and write of a file.
	     *
	     * NOTE: I'm not sure all of flags applying to write only!
	     */
	    if (match ("flags=")) {
	    floop:
		if (match(","))
		    goto floop;
		if (*string == '\0')
		    continue;
#if defined(O_EXCL)
		if (match("excl")) {
		    open_flags |= O_EXCL;	/* Exclusive open. */
		    goto floop;
		}
#endif /* defined(O_EXCL) */
#if defined(O_NDELAY)
		if (match("ndelay")) {
		    open_flags |= O_NDELAY;	/* Non-delay open. */
		    goto floop;
		}
#endif /* defined(O_NDELAY) */
#if defined(O_NONBLOCK)
		if (match("nonblock")) {
		    open_flags |= O_NONBLOCK;	/* Non-blocking open. */
		    goto floop;
		}
#endif /* defined(O_NONBLOCK) */
#if defined(O_CACHE)
		if (match("cache")) {		/* QNX specific. */
		    open_flags |= O_CACHE;	/* Keep data in cache. */
		    goto floop;
		}
#endif /* defined(O_CACHE) */
#if defined(O_DIRECT)
		if (match("direct")) {
		    open_flags |= O_DIRECT;	/* Direct disk access. */
		    goto floop;
		}
#endif /* defined(O_DIRECT) */
#if defined(SOLARIS) || defined(WIN32)
		if (match("direct")) {
		    dio_flag = TRUE;            /* No O_DIRECT on Solaris! */
		    goto floop;
		}
#endif /* defined(SOLARIS) || defined(WIN32) */
#if defined(O_FSYNC)
		if (match("fsync")) {		/* File integrity. */
		    open_flags |= O_FSYNC;	/* Syncronize file I/O. */
		    goto floop;
		}
#endif /* defined(O_FSYNC) */
#if defined(O_RSYNC)
		if (match("rsync")) {
		    open_flags |= O_RSYNC;	/* Read I/O integrity. */
		    goto floop;			/* Use with O_DSYNC or O_SYNC. */
		}
#endif /* defined(O_RSYNC) */
#if defined(O_SYNC)
		if (match("sync")) {
		    open_flags |= O_SYNC;	/* Synchronous all data access. */
		    goto floop;			/* Sync data & file attributes. */
		}
#endif /* defined(O_SYNC) */
#if defined(O_LARGEFILE)
		if (match("large")) {
		    open_flags |= O_LARGEFILE;	/* Enable large file support. */
		    goto floop;			/* Same as _FILE_OFFSET_BITS=64 */
		}
#endif /* defined(O_LARGEFILE) */
		LogMsg (efp, logLevelError, 0,
			"Invalid flag '%s' specified, please use 'help' for valid flags.\n", string); 
		exit (FATAL_ERROR);
	    } /* End if "flags=" option. */
	    if (match ("nice=")) {
		nice_value = number(ANY_RADIX);
		continue;
	    }
	    /*
	     * Flags which apply to opening a file for writes.
	     */
	    if (match ("oflags=")) {
	    oloop:
		if (match(","))
		    goto oloop;
		if (*string == '\0')
		    continue;
#if defined(WIN32)
		if (match("append")) {
		    wopen_mode = FILE_APPEND_DATA;
		    goto oloop;
		}
#else /* !defined(WIN32) */
# if defined(O_APPEND)
		if (match("append")) {
		    wopen_flags |= O_APPEND;	/* Append to file. */
		    goto oloop;
		}
# endif /* defined(O_APPEND) */
#endif /* defined(WIN32) */
#if defined(O_DEFER)
		if (match("defer")) {
		    wopen_flags |= O_DEFER;	/* Defer updates. */
		    goto oloop;
		}
#endif /* defined(O_DEFER) */
#if defined(O_DSYNC)
		if (match("dsync")) {		/* Write data integrity. */
		    wopen_flags |= O_DSYNC;	/* Synchronize data written. */
		    goto oloop;
		}
#endif /* defined(O_DSYNC) */
#if defined(O_SYNC)
		if (match("sync")) {
		    wopen_flags |= O_SYNC;	/* Synchronous all data access. */
		    goto oloop;			/* Sync data & file attributes. */
		}
#endif /* defined(O_SYNC) */
#if defined(O_TRUNC)
		if (match("trunc")) {
		    wopen_flags |= O_TRUNC;	/* Truncate output file. */
		    goto floop;
		}
#endif /* defined(O_TRUNC) */
#if defined(O_TEMP)
		if (match("temp")) {
		    wopen_flags |= O_TEMP;	/* Temporary file. */
		    goto oloop;
		}
#endif /* defined(O_TEMP) */
		LogMsg (efp, logLevelError, 0,
			"Invalid output flag '%s' specified, please use 'help' for valid flags.\n", string); 
		exit (FATAL_ERROR);
	    } /* End of "oflags=" option. */
#if defined(TTY)
	    if (match ("parity=")) {
		parity_str = string;
		if (match("even")) {
		    parity_code = PARENB;
		    data_bits_code = CS7;
		} else if (match("odd")) {
		    parity_code = (PARENB | PARODD);
		    data_bits_code = CS7;
		} else if (match("none")) {
		    parity_code = 0;
		    data_bits_code = CS8;
#if defined(_QNX_SOURCE)
		} else if (match("mark")) {
		    parity_code = (PARENB | PARODD | PARSTK);
		    data_bits_code = CS7;
		} else if (match("space")) {
		    parity_code = (PARENB | PARSTK);
		    data_bits_code = CS7;
#endif /* defined(_QNX_SOURCE) */
		} else {
#if defined(_QNX_SOURCE)
		   LogMsg (efp, logLevelError, 0,
			   "Valid parity settings are: even, odd, mark, space, or none.\n");
#else /* !defined(_QNX_SOURCE) */
		   LogMsg (efp, logLevelError, 0,
			   "Valid parity settings are: even, odd, or none.\n");
#endif /* defined(_QNX_SOURCE) */
		   exit (FATAL_ERROR);
		}
		if (data_bits_code == CS7) {
		    if (pattern == (u_int32)DEFAULT_PATTERN) {
			pattern = (u_int32)ASCII_PATTERN;
		    } else {
			pattern &= (u_int32)0x77777777;	/* Strip off parity bits. */
		    }
		}
		continue;
	    }
#endif /* defined(TTY) */
	    if (match ("oncerr=")) {
		if (match("abort")) {
		    oncerr_action = ABORT;
		} else if (match("continue")) {
		    oncerr_action = CONTINUE;
		} else {
		    LogMsg (efp, logLevelError, 0,
			    "On error actions are 'abort' or 'continue'.\n");
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    if (match ("passes=")) {
		pass_limit = number(ANY_RADIX);
		if (pass_limit == 0) {
		    LogMsg (efp, logLevelError, 0,
			    "Please specify a pass limit greater than zero!\n");
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    if (match ("pattern=")) {	/* TODO: Needs to be cleaned up. */
		int size = strlen(string);
		if (size == 0) {
		    LogMsg (efp, logLevelError, 0,
		"Please specify pattern of: { hex-pattern | incr | iot | string }\n");
		    exit (FATAL_ERROR);
		}
		iot_pattern = FALSE;
		user_pattern = TRUE;
		if (match ("incr")) {	/* Incrementing pattern. */
		    int v, size = 256;
		    u_char *buffer = (u_char *) myalloc (size, 0);
		    u_char *bp = buffer;
		    for (v = 0; v < size; v++) {
			*bp++ = v;
		    }
		    incr_pattern = TRUE;
		    setup_pattern (buffer, size);
		} else if ( (size == 3) && match("iot") ) {
		    iot_pattern = TRUE;
		    if (data_format == NONE_FMT) {
			data_format = WORD_FMT;
		    }
		    /* Allocate pattern buffer after parsing. */
		} else if ( IS_HexString(string) && (size <= 10) ) {
		    /* valid strings: XXXXXXXX or 0xXXXXXXXX */
		    pattern = (u_int32)number(HEX_RADIX);
		} else { /* Presume ASCII string for data pattern. */
		    u_char *buffer = (u_char *) myalloc (size, 0);
		    size = StrCopy (buffer, string, size);
		    pattern_size = size;
		    pattern_string = string;
		    setup_pattern (buffer, size);
		}
		if ( !iot_pattern ) unique_pattern = FALSE;
		continue;
	    }
	    if (match ("prefix=")) {
                if (prefix_string) { /* Free previous prefix (if any). */
                    free(prefix_string);
                    prefix_string = NULL;
                }
		prefix_size = strlen(string);
		if (prefix_size == 0) {
                    if (hazard_flag) {
                        continue;   /* Allow NULL override for Hazard. */
                    }
		    LogMsg (efp, logLevelError, 0,
			    "Please specify a non-empty prefix string!\n");
		    exit (FATAL_ERROR);
		} else if (prefix_size > BLOCK_SIZE) {
		    LogMsg (efp, logLevelError, 0,
			    "Please specify a prefix string < %d bytes!\n",
								prefix_size);
		    exit (FATAL_ERROR);
		}
		prefix_string = Malloc(++prefix_size); /* plus NULL! */
		(void)strcpy(prefix_string, string);
		continue;
	    }
	    if (match ("position=")) {
		file_position = (OFF_T)large_number(ANY_RADIX);
		user_position = TRUE;
		continue;
	    }
	    if (match ("procs=")) {
		num_procs = (u_short)number(ANY_RADIX);
		if (num_procs > MAX_PROCS) {
		    LogMsg (efp, logLevelError, 0,
			    "Please limit procs to <= %d!\n", MAX_PROCS);
		    exit (FATAL_ERROR);
		}
		continue;
	    }
#if defined(HP_UX)
            if (match ("qdepth=")) {
                qdepth = (u_int)number(ANY_RADIX);
                if (qdepth > SCSI_MAX_Q_DEPTH) {
                    LogMsg (efp, logLevelError, 0,
                            "Please specify a SCSI queue depth <= %d!\n", SCSI_MAX_Q_DEPTH);
                    exit (FATAL_ERROR);
                }
                continue;
            }
#endif /* defined(HP_UX) */
	    if (match ("rseed=")) {
		random_seed = (u_int)number(ANY_RADIX);
		user_rseed = TRUE;
		continue;
	    }
	    if (match ("runtime=")) {
		user_runtime = string;
		runtime = time_value();
		if (!record_limit) {
		    record_limit = INFINITY;
		}
		continue;
	    }
	    if (match ("seek=")) {
		seek_count = number(ANY_RADIX);
		continue;
	    }
	    if (match ("skip=")) {
		skip_count = number(ANY_RADIX);
		continue;
	    }
	    if (match ("slice=")) {
		slice_num = (u_short)number(ANY_RADIX);
		if ( (slice_num < 1) || (slice_num > MAX_SLICES) ) {
		    LogMsg (efp, logLevelError, 0,
			    "Please limit slice to (1 - %d)!\n", MAX_SLICES);
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    if (match ("slices=")) {
		num_slices = (u_short)number(ANY_RADIX);
		if (num_slices > MAX_SLICES) {
		    LogMsg (efp, logLevelError, 0,
			    "Please limit maximum slices to <= %d!\n", MAX_SLICES);
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    if (match ("step=")) {
		step_offset = (OFF_T)large_number(ANY_RADIX);
		continue;
	    }
	    if (match ("stats=")) {
                if (match ("brief")) {
                    stats_level = STATS_BRIEF;
		} else if (match ("full")) {
		    stats_level = STATS_FULL;
		} else if (match ("none")) {
                    pstats_flag = stats_flag = FALSE;
		    stats_level = STATS_NONE;
		    verbose_flag = FALSE; /* To overcome compatability check in report_pass()! */
		} else {
		    LogMsg (efp, logLevelError, 0,
			    "Valid stat levels are: 'brief', 'full', or 'none'\n");
		    exit (FATAL_ERROR);
		}
		continue;
	    }
#if defined(TTY)
	    if (match ("speed=")) {
		speed_str = string;
		baud_rate = (u_int32) number(ANY_RADIX);
		if (setup_baud_rate (baud_rate) == FAILURE) {
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    if (match ("timeout=")) {
		tty_timeout = (u_short)number(ANY_RADIX);
		continue;
	    }
	    if (match ("ttymin=")) {
		tty_minflag = TRUE;
		tty_minimum = (u_short)number(ANY_RADIX);
		continue;
	    }
#endif /* defined(TTY) */
	    if (match ("dtype=")) {
		struct dtype *dtp;
		if ((dtp = setup_device_type (string)) == NULL) {
		    exit (FATAL_ERROR);
		}
		input_dtype = output_dtype = dtp;
		continue;
	    }
	    if (match ("idtype=")) {
		if ((input_dtype = setup_device_type (string)) == NULL) {
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    if (match ("odtype=")) {
		if ((output_dtype = setup_device_type (string)) == NULL) {
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    if (match ("stopon=")) {
		stop_on_file = string;
		continue;
	    }
            if (match ("trigger=")) {
                if ((trigger = check_trigger_type (string)) == TRIGGER_INVALID) {
                    exit (FATAL_ERROR);
                }
                continue;
            }
	    if (match ("vrecords=")) {
	        volume_records = number(ANY_RADIX);
		continue;
	    }
	    if (match ("volumes=")) {
	        multi_flag = TRUE;
		volumes_flag = TRUE;
	        volume_limit = number(ANY_RADIX);
		continue;
	    }
#if defined(NETAPP)
	    if (match("ontap_cserver=")) {
	        ontap_cserver = string;
		continue;
	    }
	    if (match("ontap_nodes=")) {
	        ontap_nodes = string;
		continue;
	    }
	    if (match ("ontap_user=")) {
	        ontap_username = string;
		continue;
	    }
	    if (match ("ontap_password=")) {
	        ontap_password = string;
		continue;
	    }
	    if (match ("ontapi_path=")) {
	        ontapi_path = string;
		continue;
	    }
#endif /* defined(NETAPP) */
	    if (match ("help")) {
		dthelp();
	    }
	    if (match ("version")) {
		dtversion();
	    }
	    LogMsg (efp, logLevelError, 0,
		    "Invalid option '%s' specified, please use 'help' for valid options.\n",
									string);
	    exit (FATAL_ERROR);
	}
#if !defined(AIO)
	if (aio_flag) {
	    Printf ("Warning, POSIX AIO is NOT supported on this platform, disabling AIO!\n");
	    aio_flag = FALSE;
	}
#endif /* !defined(AIO) */
}

/************************************************************************
 *									*
 * match()	Match string against next input argument.		*
 *									*
 * Inputs:	s = Pointer to string to match against.			*
 *									*
 * Outputs:	Returns TRUE/FALSE = Match/No Match			*
 *		string = Updated input argument pointer.		*
 *									*
 ************************************************************************/
static int
match (char *s)
{
	char *cs;

	cs = string;
	while (*cs++ == *s) {
	    if (*s++ == '\0') {
		goto true;		/* I didn't write this junk. */
	    }
	}
	if (*s != '\0') {
		return (FALSE);
	}
true:
	cs--;
	string = cs;
	return (TRUE);
}

/************************************************************************
 *									*
 * number()	Converts ASCII string into numeric value.		*
 *									*
 * Inputs:	base = The base for numeric conversions.		*
 *									*
 * Outputs:	Returns converted number, exits on invalid numbers.	*
 *									*
 ************************************************************************/
static u_long
number (int base)
{
	char *eptr, *str = string;
	u_long value;

	value = CvtStrtoValue (str, &eptr, base);

	if (*eptr != '\0') {
	    LogMsg (efp, logLevelError, 0,
		    "Invalid character detected in number: '%c'\n", *eptr);
	    exit (FATAL_ERROR);
	}
	return (value);
}

static large_t
large_number(int base)
{
	char *eptr, *str = string;
	large_t value;

	value = CvtStrtoLarge (str, &eptr, base);

	if (*eptr != '\0') {
	    LogMsg (efp, logLevelError, 0,
		    "Invalid character detected in number: '%c'\n", *eptr);
	    exit (FATAL_ERROR);
	}
	return (value);
}

static time_t
time_value(void)
{
	char *eptr, *str = string;
	time_t value;

	value = CvtTimetoValue (str, &eptr);

	if (*eptr != '\0') {
	    LogMsg (efp, logLevelError, 0,
		    "Invalid character detected in time string: '%c'\n", *eptr);
	    exit (FATAL_ERROR);
	}
	return (value);
}

/************************************************************************
 *									*
 * report_error() - Report system error information.			*
 *									*
 * Inputs:	error_info = Additional error info for perror.		*
 *		record_flag = Controls reporting error/time info.	*
 *									*
 ************************************************************************/
void
report_error(
	char		*error_info,
	int		record_flag)
{
    struct dinfo *dip = active_dinfo;
    int saved_errno;
    char *emsg;

    saved_errno = errno;
#if defined(WIN32)
    errno = GetLastError();
#endif /* defined(WIN32) */

    if (dip) dip->di_errno = errno;

    /*
     * History is dumped here to display history on ALL errors.
     */
    if (dip && dip->di_history_size) {
	dump_history_data(dip);
    }

#if defined(WIN32)
    emsg = error_msg();
#else /* !defined(WIN32) */
    emsg = strerror(errno);
#endif /* defined(WIN32) */

    if (dip && dip->di_file_limit) {
	LogMsg (efp, logLevelCrit, 0,
		"File %s: '%s', errno = %d - %s\n", dip->di_dname, error_info, errno, emsg);
    } else {
	LogMsg (efp, logLevelCrit, 0,
		"'%s', errno = %d - %s\n", error_info, errno, emsg);
    }

#if defined(LOG_DIAG_INFO)
    if (logdiag_flag) {
	if (dip && dip->di_file_limit) {
	    (void)sprintf(msg_buffer, "%s: File %s, '%s', errno = %d - %s\n",
			    cmdname, dip->di_dname, error_info, errno, emsg);
	} else {
	    (void)sprintf(msg_buffer, "%s: '%s', errno = %d - %s\n",
			    cmdname, error_info, errno, emsg);
	}
	LogDiagMsg(msg_buffer);
    }
#endif /* defined(LOG_DIAG_INFO) */
#if defined(SYSLOG)
    if (syslog_flag) {
	if (dip && dip->di_file_limit) {
	    (void)sprintf(msg_buffer, "(%d): File %s, '%s', errno = %d - %s\n",
			  process_id, dip->di_dname, error_info, errno, emsg);
	} else {
	    (void)sprintf(msg_buffer, "(%d): '%s', errno = %d - %s\n",
			  process_id, error_info, errno, emsg);
	}
	syslog(LOG_ERR, msg_buffer);
    }
#endif /* defined(SYSLOG) */

    exit_status = FAILURE;
    if (record_flag) {
#if defined(WIN32) || defined(__sun)
	error_time = time((time_t *) 0);
	Fprintf ("Error number %lu occured on %s", ++error_count, ctime(&error_time));
#else /* !defined(WIN32) */
	char fmt[SMALL_BUFFER_SIZE];
	error_time = time((time_t *) 0);
        /* Two flavors of ctime_r(), this is a problem! */
	if (ctime_r(&error_time, fmt) == NULL) {
            //perror("ctime_r()"); /* wrong ctime_r()? */
            Fprintf("Error number %lu occurred on %s",
                     ++error_count, ctime(&error_time));
        } else {
            /* ctime_r() adds a newline '\n' to time stamp. */
	    Fprintf("Error number %lu occurred on %s", ++error_count, fmt);
        }
#endif /* defined(WIN32) || defined(__sun) */
    }
    errno = saved_errno;
    return;
}

/*
 * report_record() - Report Record Information.
 *
 * Description:
 *	This function is used when Debug is enabled, and is also
 * used to dump a history data entry.
 *
 * Inputs:
 *	dip = The device information pointer.
 *	files = The file number (limited to tapes at present).
 *	records = The record number.
 *	lba = The logical block number (NO_LBA if none).
 *	pos = The file position.
 *	mode = The test mode (READ_MODE or WRITE_MODE).
 *	buffer = The data buffer.
 *	bytes = The data byte count.
 *
 * Return Value:
 *	void
 */
void
report_record(
	struct dinfo	*dip,
	u_long		files,
	u_long		records,
	large_t		lba,
        OFF_T           pos,
	enum test_mode	mode,
	void		*buffer,
	size_t		bytes )
{
    char msg[STRING_BUFFER_SIZE];
    char *bp = msg;
    u_int32 boff = 0, eoff = 0;
    u_int32 blocks = 0;
    large_t elba = 0;

    /*
     * Depending on data supplied, calculate ending block,
     * and block offsets (for file system testing).
     */
    if ( (lba != NO_LBA) && dip->di_dsize) {
        elba = ((lba + howmany(bytes, dip->di_dsize)) - 1);
        blocks = (u_int32)((elba - lba) + 1);
	boff = (u_int32)(pos % dip->di_dsize);
	eoff = (u_int32)((pos + bytes) % dip->di_dsize);
    }
    if ( multiple_files || (dip->di_dtype->dt_dtype == DT_TAPE) ) {
	bp += sprintf(bp, "File #%lu, ", files);
    }
    bp += sprintf(bp, "Record #%lu - ", records);
    /*
     * A buffer indicates runtime debug, else assume dumping history.
     */
    if (buffer) {
	bp += sprintf(bp, "%s %u byte%s ",
		      (mode == READ_MODE) ? "Reading" : "Writing",
		      (unsigned int)bytes, (bytes > 1) ? "s" : "");
    } else {
	bp += sprintf(bp, "%s %u byte%s ",
		      (mode == READ_MODE) ? "Read" : "Wrote",
		      (unsigned int)bytes, (bytes > 1) ? "s" : "");
    }
    if (lba != NO_LBA) {
        bp += sprintf(bp, "(%u block%s) ", blocks, (blocks > 1) ? "s" : "");
    }
    if (buffer) {
	bp += sprintf(bp, "%s buffer %#lx, ",
		 (mode == READ_MODE) ? "into" : "from",
		 (u_long)buffer);
    }
    if (lba != NO_LBA) {
        bp += sprintf(bp, "lba%s ", (elba > lba) ? "'s" : "");
        if (boff && eoff) {
            bp += sprintf(bp, LUF ".%u - " LUF ".%u", lba, boff, elba, eoff);
        } else if (boff) {
            bp += sprintf(bp, LUF ".%u - " LUF, lba, boff, elba);
        } else if (eoff) {
            bp += sprintf(bp, LUF " - " LUF ".%u", lba, elba, eoff);
        } else if (lba != elba) {
	    bp += sprintf(bp, LUF " - " LUF, lba, elba);
        } else {
            bp += sprintf(bp, LUF, lba);
        }
    }
    bp += sprintf(bp, " (pos " FUF ")", pos);
    bp += sprintf(bp, "\n");
    if (buffer) {
	Printf(msg);
    } else {
	Print(msg);
    }
}

/*
 * keepalive_alarm() - Format and Display the Keepalive Message.
 *
 * Description:
 *    This function serves two purposes:
 *    1) handle alarms for keepalives.
 *    2) determine if our runtime is up.
 *
 * Basically, alarm() which is implemented using setitimer() on most
 * Unixes, does not allow more than one alarm, so we do double duty.
 *
 * Inputs:
 *    sig = The signal number (SIGALRM).
 */
void
keepalive_alarm(int sig)
{
    struct dinfo *dip = active_dinfo;
    char buffer[STRING_BUFFER_SIZE];
    register time_t initiated_time;
    int trig_action;

    initiated_time = dip->di_initiated_time;
    dip->di_last_alarm_time = time((time_t *)0);

    if (tDebugFlag) {
	char *bp = buffer;
	bp += Sprintf(bp, "Timer expired: initiated time %d, last alarm time %d",
		      initiated_time, dip->di_last_alarm_time);
	if (initiated_time) {
	    bp += Sprintf(bp, " (diff %d secs)\n", (dip->di_last_alarm_time - initiated_time));
	} else {
	    bp += Sprintf(bp, "\n");
	}
	Printf("%s", buffer);
    }

    /*
     * Check for no progress being made?
     */
    if (noprog_flag && initiated_time &&
	((dip->di_last_alarm_time - initiated_time) > noprogtime) ) {
	char *bp = buffer;
	char *optmsg = optiming_table[dip->di_optype].opt_name;
	time_t secs = (dip->di_last_alarm_time - initiated_time);

	if ( (dip->di_optype == READ_OP)  ||
	     (dip->di_optype == WRITE_OP) ||
	     (dip->di_optype == AIOWAIT_OP) ) {
	    OFF_T offset = (OFF_T)GetStatsValue(dip, ST_OFFSET, FALSE, NULL);
	    u_long records = (u_long)GetStatsValue(dip, ST_RECORDS, TRUE, NULL);
	    u_int32 lba = (dip->di_dsize) ? (u_int32)(offset / dip->di_dsize) : 0;
	    records++; /* Current outstanding record. */
	    if (optmsg) {
		bp += Sprintf(bp, "No progress made for record %u (lba %u, offset " FUF ") during %s() on %s for %d seconds!",
			      records, lba, offset, optmsg, dip->di_dname, secs);
	    } else {
		bp += Sprintf(bp, "No progress made for record %u (lba %u, offset " FUF ") on %s for %d seconds!",
			      records, lba, offset, dip->di_dname, secs);
	    }
	} else {
	    if (optmsg) {
		bp += Sprintf(bp, "No progress made during %s() on %s for %d seconds!",
			      optmsg, dip->di_dname, secs);
	    } else {
		bp += Sprintf(bp, "No progress made on %s for %d seconds!",
			      dip->di_dname, secs);
	    }
	}

	if (secs > SECS_PER_MIN) {
	    bp += Sprintf(bp, " (");
	    bp = bformat_time(bp, (secs * hz));
	    bp += Sprintf(bp, ")\n");
	} else {
	    bp += Sprintf(bp, "\n");
	}
	Printf("%s", buffer);
	/*
	 * The user can specify a no-progress trigger time to control when
	 * this gets executed.  This way one can emit no-progress messages to
	 * monitor progress, then execute the trigger at a higher threshold.
	 */
	if (secs > noprogttime) {
	    trig_action = ExecuteTrigger(dip, "noprog");
	    switch ( trig_action ) {
		case TRIGACT_CONTINUE:
		    break;
		case TRIGACT_TERMINATE:
		    ++error_count;
		    Printf("Terminating...\n");
		    terminate(exit_status = FAILURE);
		    break;
		case TRIGACT_SLEEP:
		    Printf("Sleeping forever...\n");
		    while (1) {
			sleep(60);
		    }
		    break;
		case TRIGACT_ABORT:
		    ++error_count;
		    Printf("Aborting...\n");
		    core_dump = TRUE;
		    terminate(exit_status = FAILURE);
		    break;
		default:
		    Printf("Unknown trigger action %d, continuing...\n", trig_action);
		    /* Per request of SAN Automation, we now exit on unexpected status! */
		    terminate(exit_status = FAILURE);
		    break;
	    }
	}
    }

    /*
     * Allow empty keepalive to monitor I/O progress only.
     */
    if ( keepalive && strlen(keepalive) &&
	 ((dip->di_last_alarm_time - dip->di_last_keepalive) >= keepalive_time) ) {
	dip->di_last_keepalive = dip->di_last_alarm_time;
        if (FmtKeepAlive(dip, keepalive, buffer) == SUCCESS) {
	    LogMsg (ofp, logLevelLog, 0, "%s\n", buffer);
	}
    }

    /*
     * If runtime specified, see if we hit our end time.
     */
    if (runtime) {
	if (dip->di_last_alarm_time >= runtime_end) {
	    terminate(sig); /* End of our runtime, exit. */
	}
    }

    if (iotune_file) {
	process_iotune(dip, iotune_file);
    }
#if !defined(WIN32) && !defined(THREADS)
    /* Reenable us! */
    (void) signal (SIGALRM, keepalive_alarm);
    (void) alarm (alarmtime);
#endif /* !defined(WIN32) && !defined(THREADS) */
    return;
}

/*
 * SignalHandler() - Signal handler for all signals we care about.
 *
 * Inputs:
 *	signal_number = The signal number.
 *
 * Return Value:
 *	void
 */
void
SignalHandler(int signal_number)
{
    if (debug_flag || pDebugFlag) {
	Printf("Caught signal %d\n", signal_number);
    }
    /*
     * Note: We need to do signal handling here, so terminate() only has
     * to worry about cleanup activity. But, that's major cleanup for
     * another day! (gotta get back to my boring day job! :-()
     */
    terminate(signal_number);
}

/*
 * terminate() - Terminate program with specified exit code.
 *
 * Inputs:
 *	code = The exit code or signal number if kill done.
 */
void
terminate (int code)
{
    struct dinfo *dip = active_dinfo;
    int status;
    bool do_cleanup = TRUE;

    if (debug_flag || pDebugFlag) {
	Printf("Terminating with code %d...\n", code);
    }
#if defined(THREADS)
    if ((alarmtime || runtime) && ParentThreadId) {
        if (ParentThreadId == pthread_self()) {
	    /* Parent Thread Processing */
	    if (debug_flag || tDebugFlag) {
		Printf("Canceling the monitor thread...\n");
	    }
            if (pthread_cancel(MonitorThreadId) == FAILURE) {
                Perror("pthread_cancel() on monitor thread ID %d failed!\n", MonitorThreadId);
            }
        } else {
	    /* Monitor Thread Processing */
	    if (debug_flag || tDebugFlag) {
		Printf("Canceling the I/O thread...\n");
	    }
	    (void)SetThreadCancelType(PTHREAD_CANCEL_ASYNCHRONOUS);
            if (pthread_cancel(ParentThreadId) == FAILURE) {
                Perror("pthread_cancel() on parent thread ID %d failed!\n", ParentThreadId);
            }
	    if (cancel_delay) {
		sleep(cancel_delay);	/* Give the thread a chance to actually cancel... */
					/* Note: This delay will affect throughput stats! */
	    }
        }
    }
#else /* !defined(THREADS) */
    if (alarmtime) alarm(0);	/* Cancel alarm (if any). */
#endif /* defined(THREADS) */

    /*
     * If we enter here more than once, just exit to avoid
     * possible recursion.  kernel should do I/O rundown!
     */
    if (terminating_flag) {
	if (core_dump && (code != SUCCESS)) {
	    if (debug_flag || pDebugFlag) {
		Printf("Forcing core dump via abort()...\n");
	    }
	    abort();		/* Generate a core dump. */
	} else {
	    if (debug_flag || pDebugFlag) {
		Printf("Exiting with code %d, due to already terminating!\n");
	    }
	    exit (code);	/* get outta here now... */
	}
    } else {
	terminating_flag++;	/* Show we're terminating. */
    }

#if defined(WIN32)
{
    DWORD waitResult;
    waitResult = WaitForSingleObject(exitMutex, INFINITE);
    if (waitResult != WAIT_OBJECT_0) {
	Fprintf("WaitForSingleObject() returned %d\n", waitResult);
    }
    /*
     * Note: We may get here via the parent thread simply exiting,
     * or via the timer thread or a signal thread.  In any case,
     * for the latter, we must kill the parent since this function
     * closes the handle, which will cause false failures otherwise.
     */
    if (ParentThreadId != GetCurrentThreadId()) {
	DWORD status;
	DWORD timeoutMs;
	int wait_retries = 0;

	/*
	 * SIGALRM means we've completed our runtime, so wait longer
	 * than if we're terminating due to another signal or exiting
	 * due to the trigger script (which now uses FAILURE).
	 */
	if (term_wait_retries == 0) { /* May get set by option. */
	    if (code == SIGALRM) {
		term_wait_retries = TERM_WAIT_RETRIES;
	    } else {
		term_wait_retries = TERM_WAIT_ARETRIES;
	    }
	}
	if (term_wait_timeout == 0) { /* May get set by option. */
	    term_wait_timeout = TERM_WAIT_TIMEOUT;
	}
	timeoutMs = (1000 * TERM_WAIT_TIMEOUT);
	/*
	 * Give the parent thread time to exit, which may take a while
	 * if the file is being flushed or there's heavy I/O load.
	 * Note: The I/O loops are now checking for terminating!
	 */
	do {
	    wait_retries++;
	    if (debug_flag) {
		Printf("Waiting for parent thread to exit, retry %d...\n", wait_retries);
	    }
	    status = WaitForSingleObject(ParentThread, timeoutMs);
	    if (wait_retries >= term_wait_retries) break;
	} while (status == WAIT_TIMEOUT);
	/*
	 * If the parent didn't exit cleanly, don't do file ops!
	 */
	if (status != WAIT_OBJECT_0) {
	    Printf("Parent thread is still active, exiting anyways...\n");
	    do_cleanup = FALSE;
	    // TerminateParentThread(); /* Dangerous! */
	}
    }
}
#endif /* defined(WIN32) */

    /*
     * If we're here due to our timer expiring, and we were writing
     * a file, then flush the data (to regular file).
     */
    if ( do_cleanup && (code == SIGALRM) && dip && (dip->di_mode == WRITE_MODE) ) {
	/*
	 * We may get alarm while open() is in progress, so handle.
	 */
	if ( !dip->di_closing && !dip->di_flushing && (dip->di_fd != NoFd) ) {
	    (void)(*dip->di_funcs->tf_flush_data)(dip);
	}
    }

    /*
     * If an alarm timer is active, cancel it and calculate
     * the elapsed run time.
     */
    if (TimerActive) {
#if !defined(WIN32)
	elapsed_time = runtime - alarm(0);
#endif /* !defined(WIN32) */
	TimerActive = FALSE;
	/*
	 * Be careful not to overwrite other signal codes.
	 */
	if (code == SIGALRM) {
	    code = exit_status;	/* Set the exit status! */
	}
    }

    /*
     * We only come here for signals when executing multiple
     * processes, so abort active procs and continue waiting.
     */
    if (procs_active && child_pid) {
	if (kill_delay) {
	    /* 
	     * When running interactively, process group gets signal, so
	     * if we send another signal quickly, children exit immediately
	     * rather than doing their cleanup and reporting statistics.
	     */
	    sleep(kill_delay);
	}
	abort_procs();		/* Abort any active procs. */
#if defined(WIN32)
	/*
	 * For Unix, we must return to waiting for processes to reap their
	 * exit status to avoid zombie processes!
	 * For native Windows, thankfully this is not necessary.
	 */
	exit (exit_status);
#else /* !defined(WIN32) */
	if (debug_flag || pDebugFlag) {
	    Printf("Parent returning from signal to wait for children...\n");
	}
	return;
#endif /* defined(WIN32) */
    }
#if (!defined(__MSDOS__) && !defined(WIN32)) || defined(__NUTC__)
    if (debug_flag && (code == SIGCHLD)) {
	Printf ("Child process exited prematurely, parent exiting...\n");
    } else 
#endif

    /*
     * Report stats before waiting for the child process.
     */
#if (!defined(__MSDOS__) && !defined(WIN32)) || defined(__NUTC__)
    (void) signal (SIGCHLD, SIG_DFL); /* Don't deliver SIGCHLD now. */
#endif
    /*
     * Close file, which for AIO waits for outstanding I/O's,
     * before reporting statistics so they'll be correct.
     */
    if (dip && do_cleanup && (dip->di_fd != NoFd) ) {
	dip->di_proc_eei = FALSE;
	status = (*dip->di_funcs->tf_close)(dip);
	if (status != SUCCESS) code = status;
    }
    gather_stats(dip);		/* Gather the device statistics. */
    gather_totals(dip);		/* Update the total statistics.	*/
    report_stats(dip, TOTAL_STATS);

    /*
     * TODO: This needs cleaned up.  Alot of legacy code...
     * NOTE: This code is used for tty support in loopback.
     */
#if !defined(WIN32)
    if (child_pid) {		/* Always wait for child status. */
	pid_t wpid = (pid_t) 0;
	/*
	 * If exiting with error, ensure the child gets killed.
	 */
	if (code != SUCCESS) {
	    (void) kill (child_pid, SIGTERM);
	}
	if (debug_flag) {
	    Printf ("Waiting for child PID %d to exit...\n", child_pid);
	}
#if defined(_BSD)
	do {
	    if ((wpid = wait (&child_status)) == FAILURE) {
		report_error ("wait", TRUE);
		break;
	    }
	} while (wpid != child_pid);
	if ( (code == SUCCESS) && (wpid != FAILURE) ) {
	    code = WEXITSTATUS (child_status);
	}
#else /* !defined(_BSD) */
	if ((wpid = waitpid (child_pid, &child_status, 0)) == FAILURE) {
	    report_error ("waitpid", TRUE);
	} else if (code == SUCCESS) {
	    code = WEXITSTATUS (child_status);
	}
#endif /* defined(_BSD) */
	if (debug_flag && (wpid != FAILURE)) {
	    Printf("Child PID %d, exited with status = %d\n",
					wpid, WEXITSTATUS(child_status));
	}
    }
#endif /* !defined(WIN32) */

    /*
     * If keep on error, do the appropriate thing!
     */
    if (dispose_mode == KEEP_ON_ERROR) {
	/* Note: Signals cause files to be kept! */
	if ( (code != SUCCESS) && (code != END_OF_FILE) ) {
	    dispose_mode = KEEP_FILE;
	} else if (dip && !dip->di_existing_file) {
	    dispose_mode = DELETE_FILE;
	}
    }
    /*
     * Delete the output file, if requested to do so.
     */
    if (do_cleanup && output_file && (io_mode == TEST_MODE)             &&
	(dispose_mode == DELETE_FILE)			                &&
	(dip && dip->di_dtype && (dip->di_dtype->dt_dtype == DT_REGULAR)) ) {
	int status = delete_files(dip);
	if (status) code = status; /* Delete failed, that's a test failure! */
    }

#if defined(LOG_DIAG_INFO)
    if (logdiag_flag) {
	sprintf(msg_buffer, "Finished: %s", cmd_line);
	LogDiagMsg(msg_buffer);
    }
#endif /* defined(LOG_DIAG_INFO) */
#if defined(SYSLOG)
    if (syslog_flag) {
	syslog(LOG_INFO, "(%d) Finished: %s", process_id, cmd_line);
    }
#endif /* defined(SYSLOG) */

    if (dip && dip->di_history_size && history_dump) {
	dump_history_data(dip);	/* For debugging history mechanism */
    }
    if ( (eof_status == FALSE) && (code == END_OF_FILE)) {
	code = SUCCESS;		/* Map end-of-file status to Success! */
    }
    if (debug_flag || pDebugFlag) {
	Printf ("Exiting with status code %d...\n", code);
    }
    if (core_dump && (code != SUCCESS) && (code != END_OF_FILE)) {
	abort();			/* Generate a core dump. */
    }
    /*
     * Map signal numbers and/or other errno's to FAILURE. (cleanup)
     * ( easier for scripts to handle! )
     */
    if ( (code != FAILURE) && (code != SUCCESS) && (code != END_OF_FILE) ) {
	code = FAILURE;			/* Usually a signal number. */
    }
    if (forked_flag && tdelay_count && (child_pid == (pid_t) 0)) {
	(void)sleep (tdelay_count);	/* Allow parent to wait for us.	*/
    }
    exit (code);
}

int
nofunc (struct dinfo *dip)
{
	return (SUCCESS);
}

static char *multi_prompt = 
    "\nPlease insert volume #%d in drive %s, press ENTER when ready to proceed: \007";
static char *multi_nready =
    "The drive is NOT ready or encountered an error, Retry operation (Yes): \007";

int
HandleMultiVolume (struct dinfo *dip)
{
    int status;

    status = RequestMultiVolume (dip, FALSE, dip->di_oflags);
    if (status == FAILURE) return (status);

    if (dip->di_mode == READ_MODE) {
	dip->di_volume_bytes = (large_t)(dip->di_dbytes_read + total_bytes_read);
	if (verbose_flag) {
	  if ( multiple_files || (dip->di_dtype->dt_dtype == DT_TAPE) ) {
	    Print ("    [ Continuing in file #%lu, record #%lu, bytes read so far " LUF "... ]\n",
		(dip->di_files_read + 1), (dip->di_records_read + 1), dip->di_volume_bytes);
	  } else {
	    Print ("    [ Continuing at record #%lu, bytes read so far " LUF "... ]\n",
			(dip->di_records_read + 1), dip->di_volume_bytes);
	  }
	}
	dip->di_vbytes_read = (large_t) 0;
    } else {
	dip->di_volume_bytes = (large_t)(dip->di_dbytes_written + total_bytes_written);
	if (verbose_flag) {
	  if ( multiple_files || dip->di_dtype->dt_dtype == DT_TAPE) {
	    Print ("    [ Continuing in file #%lu, record #%lu, bytes written so far " LUF "... ]\n",
		(dip->di_files_written + 1), (dip->di_records_written + 1), dip->di_volume_bytes);
	  } else {
	    Print ("    [ Continuing at record #%lu, bytes written so far " LUF "... ]\n",
			(dip->di_records_written + 1), dip->di_volume_bytes);
	  }
	}
	dip->di_vbytes_written = (large_t) 0;
    }
    (void)fflush(ofp);
    media_changed = TRUE;
    dip->di_volume_records = 0;
    if (exit_status == END_OF_FILE) {
	exit_status = SUCCESS;		/* Ensure END_OF_FILE status is reset! */
    }
    return (status);
}

int
RequestFirstVolume (struct dinfo *dip, int oflags)
{
    int status;

    multi_volume = 0;

    status = RequestMultiVolume (dip, TRUE, oflags);

    dip->di_volume_bytes = (large_t) 0;
    dip->di_volume_records = 0;

    return (status);
}

int
RequestMultiVolume (struct dinfo *dip, bool reopen, int oflags)
{
    struct dtfuncs *dtf = dip->di_funcs;
    char buffer[256];
    char *bp = buffer;
    FILE *fp;
    int saved_exit_status;
    u_long saved_error_count;
    int status;

    if (terminating_flag) return (FAILURE);

    if ( (status = (*dtf->tf_close)(dip)) == FAILURE) {
	return (status);
    }

    if ( (fp = fopen("/dev/tty", "r+")) == NULL) {
	report_error("open of /dev/tty failed", FALSE);
	return (FAILURE);
    }
    multi_volume++;

    (void)sprintf(bp, multi_prompt, multi_volume, dip->di_dname);

    (void) fputs (bp, fp); fflush(fp);
    if (fgets (bp, sizeof(buffer), fp) == NULL) {
	Print ("\n");
	status = FAILURE;	/* eof or an error */
	return (status);
    }

    saved_error_count = error_count;
    saved_exit_status = exit_status;

    /*
     * This is an important step, so allow the user to retry on errors.
     */
    do {
	if (!reopen) {
	    status = (*dtf->tf_open)(dip, oflags);
	} else {
	    status = (*dtf->tf_reopen_file)(dip, oflags);
	}
	if (status == SUCCESS) {
#if defined(TAPE) && !defined(__NUTC__) && !defined(__QNXNTO__) && !defined(AIX) && !defined(WIN32)
	    if (dip->di_dtype->dt_dtype == DT_TAPE) {
		status = DoRewindTape (dip);
		if (status == FAILURE) {
		    (void)(*dtf->tf_close)(dip);
		}
	    }
#endif /* defined(TAPE) && !defined(__NUTC__) && !defined(__QNX_NTO__) && !defined(AIX) && !defined(WIN32) */
	}
	if (status == FAILURE) {
	    (void) fputs (multi_nready, fp); fflush(fp);
	    if (fgets (bp, sizeof(buffer), fp) == NULL) {
		Print ("\n");
		break;
	    }
	    if ( (bp[0] == 'N') || (bp[0] == 'n') ) {
		break;
	    }
	    error_count = saved_error_count;
	    exit_status = saved_exit_status;
	} else {
	    break;		/* device is ready! */
	}
    } while (status == FAILURE);

    (void)fclose(fp);
    return (status);
}

#if defined(WIN32)

/*
 * suspend_thread() - Suspend the Specifed Thread.
 *
 * Inputs:
 *	thread_handle = Handle to thread to suspend.
 *
 * Return Value:
 *	The resume count, or -1 on failure.
 */
DWORD
resume_thread(HANDLE thread_handle)
{
    DWORD retval;

    if (debug_flag) {
	Printf("Resuming thread %p...\n", thread_handle);
    }
    retval = ResumeThread(thread_handle);
    if (retval == -1) {
	DWORD error = GetLastError();
	/*
	 * Parent thread may be terminating due to signal.
	 */
	if ( (error != ERROR_ACCESS_DENIED) &&
	     (error != ERROR_INVALID_HANDLE) ) {
	    report_error("ResumeThread() failed", FALSE);
	}
    } else if (tDebugFlag) {
	if (retval == 1) {
	    Printf("Suspended thread was restarted!\n");
	} else {
	    Printf("Resume count is %d (still suspended!)\n", retval);
	}
    }
    return(retval);
}

/*
 * suspend_thread() - Suspend the Specifed Thread.
 *
 * Inputs:
 *	thread_handle = Handle to thread to suspend.
 *
 * Return Value:
 *	The previous suspend count, or -1 on failure.
 */
DWORD
suspend_thread(HANDLE thread_handle)
{
    DWORD retval;
    /*
     * Suspend/Resume parent to emulate Unix signaling.
     * Main reason is to keep parent from exiting on us!
     * With native Windows, we're called from a separate thread.
     */
    if (debug_flag) {
	Printf("Suspending thread %p...\n", thread_handle);
    }
    retval = SuspendThread(thread_handle);
    if (retval == -1) {
	DWORD error = GetLastError();
	/*
	 * Parent thread may be terminating due to signal.
	 */
	if ( (error != ERROR_ACCESS_DENIED) &&
	     (error != ERROR_INVALID_HANDLE) ) {
	    report_error("SuspendThread() failed", FALSE);
	}
    } else if (tDebugFlag) {
	Printf("Suspend count is %d\n", retval);
    }
    return(retval);
}

/*
 * TerminateParentThread() - Terminate the Parent Thread.
 *
 * NOTE: This *should* be a generic function!
 *
 * Implicit Inputs:
 *	ParentThread = The parent thread handle.
 *	ParentThreadId = The parent thread ID.
 *
 * Return Value:
 *	Void
 */
void
TerminateParentThread(void)
{
    HANDLE thread_handle = ParentThread;
    DWORD thread_id = ParentThreadId;

    ParentThreadId = 0;
    ParentThread = (HANDLE) 0;	/* Help avoid recursion until locks added. */
    if (debug_flag) {
	Printf("Terminating parent thread %d...\n", thread_id);
    }
    /*
     * BEWARE: I found these warnings here...
     * URL: http://msdn.microsoft.com/en-us/library/ms686717(VS.85).aspx
     *
     * TerminateThread is a dangerous function that should only be used in the
     * most extreme cases. You should call TerminateThread only if you know exactly
     * what the target thread is doing, and you control all of the code that the
     * target thread could possibly be running at the time of the termination.
     *
     * For example, TerminateThread can result in the following problems:
     *  - If the target thread owns a critical section, the critical section
     *    will not be released. 
     *  - If the target thread is allocating memory from the heap, the heap lock
     *    will not be released. 
     *  - If the target thread is executing certain kernel32 calls when it is
     *    terminated, the kernel32 state for the thread's process could be inconsistent. 
     *  - If the target thread is manipulating the global state of a shared DLL, the
     *    state of the DLL could be destroyed, affecting other users of the DLL. 
     *
     * A user added these additional warnings:
     * The warnings above are a gross understatement. TerminateThread is NOT a
     * solution for shutting down worker threads. You will break your app, and
     * it will fail randomly.
     *
     * Set a flag for the thread to watch for and then voluntarily shut itself down.
     * If your threads are hanging, fix the hang. Involuntarily killing threads only
     * makes things worse.
     *
     * Robin: In general, I'm finding this to be TRUE, so should be last resort!
     */
    if ( !TerminateThread(thread_handle, ERROR_SUCCESS) ) {
	/*
	 * During Ctrl/C aborting, sometimes the parent thread is gone,
	 * while we're running in the signal handler thread.
	 */
	if ( (GetLastError() != ERROR_INVALID_HANDLE) || debug_flag) {
	    report_error("TerminateThread() failed", TRUE);
	}
    } else {
	DWORD thread_status;
	/*
	 * The terminated thread can still be active, so wait for it!
	 * Not waiting, can lead to unexpected things (like hangs).
	 */
	if (debug_flag) {
	    Printf("Waiting for parent thread %d to exit...\n", thread_id);
	}
	do {
	    if ( !GetExitCodeThread(thread_handle, &thread_status) ) {
		report_error("GetExitCodeThread() failed", TRUE);
		break;
	    }
	} while (thread_status == STILL_ACTIVE);
	if (debug_flag) {
	    Printf("Parent thread exited with status %d\n", thread_status);
	    if (thread_status) abort();
	}
    }
    (void)CloseHandle(thread_handle);
    return;
}

/*
 * Alarm() - Windows Alarm Handler.
 *
 * Description:
 *	This function implements a monitoring thread for emulating the
 * Unix alarm() API.  dt uses alarm() for keepalive messages, runtime,
 * and no-progress monitoring.  The function will terminate dt when the
 * runtime is reached, after terminating the parent thread to avoid
 * errors when terminate() closes files.
 *
 * Inputs:
 *	aa = The alarm arguments.
 *
 * Return Value:
 *	Always SUCCESS
 */
DWORD WINAPI
Alarm(struct AlrmArgs *aa)
{
    /*
     * Basically wait for the alarm timeout, and allow the standard
     * keepalive() handling to process noprog, keepalive, & runtime.
     *
     * Note: This is a separate thread from the rest of dt!
     */
    while (WaitForSingleObject(aa->event, aa->timeout) == WAIT_TIMEOUT) {
	if (terminating_flag) break;
	(void)keepalive_alarm(SIGALRM);
	if (terminating_flag) break;
    }
    return (ERROR_SUCCESS);
}

#define _SECOND 10000000

typedef struct _MYDATA {
   HANDLE hTimer;
} MYDATA;

VOID CALLBACK TimerAPCProc(
   LPVOID lpArg,               // Data value
   DWORD dwTimerLowValue,      // Timer low value
   DWORD dwTimerHighValue )    // Timer high value
{
   MYDATA *pMyData = (MYDATA *)lpArg;

   if ( !CloseHandle( pMyData->hTimer ) ) {
       Fprintf("CloseHandle() failed with error %d\n", GetLastError());
   }
   free(pMyData);
   terminate(SIGALRM);
   return;
}

void
winAlarm(u_int sec)
{
#if defined(TIMER_METHOD)
   /*
    * This method creates a timer that runs in our current thread context.
    * The caveat is the thread must enter an alertable state. A thread can
    * only do this by calling one of the following functions with the
    * appropriate flags:
    *
    * SleepEx() WaitForSingleObjectEx() WaitForMultipleObjectsEx() 
    * SignalObjectAndWait() MsgWaitForMultipleObjectsEx()
    *
    * Until dt is updated to use alertable API's, we can't enable this.
    */
    HANDLE          hTimer;
    BOOL            bSuccess;
    __int64         qwDueTime;
    LARGE_INTEGER   liDueTime;
    MYDATA          *MyData;

    MyData = Malloc(sizeof(*MyData));

    hTimer = CreateWaitableTimer(
				NULL,			// Default security attributes
				FALSE,			// Create auto-reset timer
				"MyTimer");		// Name of waitable timer
    if (hTimer == NULL) {
	Fprintf("CreateWaitableTimer() failed with error %d\n", GetLastError());
	return;
    } else {
	MyData->hTimer = hTimer;
	// Create an integer that will be used to signal the timer.
	qwDueTime = (-(int)sec * _SECOND);

	// Copy the relative time into a LARGE_INTEGER.
	liDueTime.LowPart  = (DWORD) ( qwDueTime & 0xFFFFFFFF );
	liDueTime.HighPart = (LONG)  ( qwDueTime >> 32 );

	bSuccess = SetWaitableTimer(
				   hTimer,	     // Handle to the timer object
				   &liDueTime,	     // When timer will become signaled
				   0,		     // Periodic timer, execute once
				   TimerAPCProc,     // Completion routine
				   MyData,	     // Argument to the completion routine
				   FALSE );	     // Do not restore a suspended system

	if ( !bSuccess ) {
	    Fprintf("SetWaitableTimer() failed with error %d\n", GetLastError());
	}
    } 
#else /* !defined(TIMER_METHOD) */
    static struct AlrmArgs aArgs = {1, INVALID_HANDLE_VALUE, INVALID_HANDLE_VALUE};

    if (aArgs.event == INVALID_HANDLE_VALUE) {
        char line[64];
        sprintf(line, "AlarmHndl%d", process_id);
        if ((aArgs.event = CreateEvent(NULL, FALSE, FALSE, line)) == 0) {
	    Fprintf("winAlarm: CreateEvent() failed\n");
	    aArgs.event = INVALID_HANDLE_VALUE;
	    return;
	}
    }
    if (sec > 0) {
	HANDLE th;
	DWORD tid;
	aArgs.timeout = (sec * 1000);
	aArgs.handle = ParentThread;
	//if ((th = CreateThread(NULL, 0, Alarm, &aArgs, 0, &tid)) == 0) {
	/*
	 * If you are going to call C run-time routines from a program built
	 * with Libcmt.lib, you must start your threads with the _beginthread
	 * or _beginthreadex function. Do not use the Win32 functions ExitThread
	 * and CreateThread. Using SuspendThread can lead to a deadlock when more
	 * than one thread is blocked waiting for the suspended thread to complete
	 * its access to a C run-time data structure.
	 *
	 * URL: http://msdn.microsoft.com/en-us/library/7t9ha0zh(VS.80).aspx
	 */
	if ((th = (HANDLE)_beginthreadex(NULL, 0, Alarm, &aArgs, 0, &tid)) == 0) {
	    report_error("_beginthreadex() failed", FALSE);
	}
    } else {
	if (SetEvent(aArgs.event) == 0) {
	    Fprintf("winAlarm: SetEvent() failed\n");
	}
    }
#endif /* defined(TIMER_METHOD) */
    return;
}

DWORD WINAPI
Rdstdin(HANDLE h)
{
    char buf[4096];
    DWORD n;

    if (ReadFile(GetStdHandle(STD_INPUT_HANDLE), buf, sizeof(buf), &n, NULL) == 0) {
	Fprintf("Rdstdin: ReadFile failed %d error\n", GetLastError());
	if (TerminateThread(h, 0 ) == 0) {
	    Fprintf("TerminateThread failed %d error\n", GetLastError());
	}
	Printf("Exiting...\n"); 
	terminate(SUCCESS);
    }
    return (0);
}

void
ReadStdin()
{
    HANDLE h,hmain;
    DWORD tid;

    if (!DuplicateHandle(GetCurrentProcess(), GetCurrentThread(),
			 GetCurrentProcess(), &hmain, 0, TRUE, DUPLICATE_SAME_ACCESS)) {
	Fprintf("ReadStdin: DuplicateHandle failed\n");
    }
    if ((h = CreateThread(NULL, 0, Rdstdin, hmain, 0, &tid)) == 0) {
	Fprintf("ReadStdin: CreateThread failed\n");
    } else {
	CloseHandle(h);
    }
    return;
}
    
#elif defined(THREADS)

/*
 * MonitorThread() - Unix Monitor Thread.
 *
 * Description:
 *      This thread is used to implement SIGALRM usually done via alarm()
 * but is necessary to avoid the issue where signals are blocked on some
 * OS's while doing I/O (Linux, HP-UX, Solaris, etc).
 *
 * Inputs:
 *	arg = Pointer to monitoring interval.
 *
 * Return Value:
 *	NULL pointer.
 */
void *
MonitorThread(void *arg)
{
    int *interval = (int *)arg;
    /*
     * Basically wait for the monitor interval, and allow the standard
     * keepalive() handling to process noprog, keepalive, & runtime.
     *
     * Note: This is a separate thread from the rest of dt!
     */
    while (1) {
        (void)sleep(*interval);
	if (terminating_flag) break;
	(void)keepalive_alarm(SIGALRM);
	if (terminating_flag) break;
    }
    pthread_exit(NULL);
    return(NULL);
}

/*
 * WTF! Linux stack size is 10MB by default! (will get cleaned up one day)
 */
#define THREAD_STACK_ENV "DT_THREAD_STACK_SIZE" /* Thread stack size.	*/
#if !defined(PTHREAD_STACK_MIN)
#  define PTHREAD_STACK_MIN 16384
#endif
/* Note: If the stack size is too small we seg fault, too large wastes address space! */
#define THREAD_STACK_SIZE	((PTHREAD_STACK_MIN + STRING_BUFFER_SIZE) * 4)
				/* Reduce TLS to avoid wasting swap/memory! */
#define MAX_THREADS 3		/* Maximum concurrent threads, for now! */

int
StartMonitorThread(int interval)
{
    static int monitorInterval;
    size_t currentStackSize = 0;
    size_t desiredStackSize = THREAD_STACK_SIZE;
#if 0
    int max_threads = MAX_THREADS;
#endif
    int status;
    char *p;

    if (p = getenv(THREAD_STACK_ENV)) {
	string = p;
	desiredStackSize = number(ANY_RADIX);
    }
    monitorInterval = interval;
    ParentThreadId = pthread_self();
    /* Remember: pthread API's return 0 for success; otherwise, an error number to indicate the error! */
    status = pthread_attr_init(tattrp);
    if (status != SUCCESS) {
	errno = status;
	Perror("pthread_attr_init() failed");
	return (status);
    }
    status = pthread_attr_setscope(tattrp, PTHREAD_SCOPE_SYSTEM);
    /* Note: This is NOT supported with Cygwin! */
    if ( (status != SUCCESS) && (status != ENOTSUP) ) {
	errno = status;
	Perror("pthread_attr_setscope() failed setting PTHREAD_SCOPE_SYSTEM");
	//return (status); /* This is considered non-fatal! */
    }

#if 0
    /*
     * Verify the thread stack size (TLS) is NOT too large! (Linux issue). 
     * Yea, only one thread right now, but I'm preparing for the future. :-)
     */
    status = pthread_attr_getstacksize(tattrp, &currentStackSize);
    if (status == SUCCESS) {
	if (debug_flag || tDebugFlag) {
	    Printf("Current thread stack size is %u\n", currentStackSize);
	}
    } else {
	if (debug_flag || tDebugFlag) {
	    errno = status;
	    Perror("pthread_attr_getstacksize() failed!");
	}
    }

    /*
     * The default stack size on Linux is 10M, which limits the threads created! 
     */
    if (currentStackSize && desiredStackSize && (currentStackSize > desiredStackSize) ) {
	/* Too small and we seg fault, too large limits our threads. */
	status = pthread_attr_setstacksize(tattrp, desiredStackSize);
	if (status == SUCCESS) {
	    if (debug_flag || tDebugFlag) {
		Printf("Thread stack size set to %u bytes\n", desiredStackSize);
	    }
	} else {
	    errno = status;
	    Perror("pthread_attr_setstacksize() failed setting stack size %u",
		   desiredStackSize);
	}
    }
#endif    

#if 0
    /* NOTE: Do NOT do this! Linux AIO is via threads at present!   */
    /* FWIW: Regardless, Linux threaded AIO is still very slow!     */
    /* TODO: Write functions that use native Linux AIO (not POSIX). */
    /*
     * Set the concurrency level before we create any threads.
     * Note: This is a hint, and actual behaviour is unspecified! (nice)
     *
     * Documentation:
     *     Unbound threads in a process may or may not be required to be
     * simultaneously active. By default, the threads implementation ensures
     * that a sufficient number of threads are active so that the process can
     * continue to make progress. While this conserves system resources, it
     * may not produce the most effective level of concurrency.
     */
    status = pthread_setconcurrency(max_threads);
    if (status != SUCCESS) {
	errno = status;
	Perror("pthread_setconcurrency() failed setting max_threads to %d", max_threads);
	return (status);
    }
#endif

    /*
     * Now, create detached threads so we don't need to do a pthread_join()
     */
    status = pthread_attr_setdetachstate(tattrp, PTHREAD_CREATE_DETACHED);
    if (status != SUCCESS) {
	errno = status;
	Perror("pthread_attr_setdetachstate() failed setting PTHREAD_CREATE_DETACHED");
	return (status);
    }
    status = pthread_create( &MonitorThreadId, tattrp, MonitorThread, &monitorInterval );
    if (status != SUCCESS) {
	errno = status;
        Perror ("pthread_create() failed");
        return (status);
    }
    /*
     * The main thread, will continue on as the I/O thread, but if/when the monitor* 
     * thread wants us to terminate, we want the I/O thread cancel to be immediate,
     * thus we need to enable async cancelation, otherwise we may have false errors.
     */ 
    /* Note: Linux AIO threads may screw this, not sure! */
    //(void)SetThreadCancelType(PTHREAD_CANCEL_ASYNCHRONOUS);

    return (status);
}

char *
decodeCancelType(int cancel_type)
{
    switch (cancel_type) {
	case PTHREAD_CANCEL_ASYNCHRONOUS:
	    return ("PTHREAD_CANCEL_ASYNCHRONOUS");
	    break;
	case PTHREAD_CANCEL_DEFERRED:
	    return ("PTHREAD_CANCEL_DEFERRED");
	    break;
	default:
	    return ("cancel type unknown");
	    break;
    }
}

int
SetThreadCancelType(int cancel_type)
{
    int status;
    int old_cancel_type = 0;

    if (tDebugFlag) {
	Printf("Setting the thread cancel type to %s...\n", decodeCancelType(cancel_type));
    }
    if ( (status = pthread_setcanceltype(cancel_type, &old_cancel_type)) == SUCCESS) {
	if (tDebugFlag) {
	    Printf("Previous cancel type is %d (%s)\n",
		  old_cancel_type, decodeCancelType(old_cancel_type));
	}
    } else {
	errno = status;
	Perror("pthread_setcanceltype() failed");
    }
    return (status);
}

#endif /* defined(THREADS) */

/*
 * init_vars() - Initialize The Device Structure Variables.
 *
 * Descriptions:
 *      Copy global variables to device specific structure.
 *
 * Inputs:
 *      dip = Device information pointer.
 *
 * Return Value:
 *      void
 */
void
init_vars(struct dinfo *dip)
{
    dip->di_file_limit = file_limit;
    dip->di_max_data = max_data;
    dip->di_max_files = max_files;
    /* I/O Delays */
    dip->di_close_delay = cdelay_count;
    dip->di_end_delay = edelay_count;
    dip->di_read_delay = rdelay_count;
    dip->di_start_delay = sdelay_count;
    dip->di_write_delay = wdelay_count;
    /* Data Pattern Information */
    dip->di_lbdata_flag = lbdata_flag;
    dip->di_unique_pattern = unique_pattern;
    dip->di_lbdata_addr = lbdata_addr;
    dip->di_lbdata_size = lbdata_size;
    dip->di_iot_pattern = iot_pattern;
    dip->di_iot_seed = iot_seed;
    dip->di_iot_seed_per_pass = iot_seed_per_pass;
#if 0
    dip->di_pattern = pattern;
    dip->di_patbuf_size = patbuf_size;
    dip->di_pattern_buffer = pattern_buffer;
    dip->di_pattern_bufptr = pattern_bufptr;
    dip->di_pattern_bufend = pattern_bufend;
#endif
#if defined(NETAPP)
    dip->di_ontap_cserver = ontap_cserver;
    dip->di_ontap_nodes = ontap_nodes;
    dip->di_ontap_username = ontap_username;
    dip->di_ontap_password = ontap_password;
    dip->di_ontapi_path = ontapi_path;
#endif /* defined(NETAPP) */
    /* Many more to come... */
    return;
}
